/* -----------------------------------------------------------------------

The MIT License (MIT)

Copyright (c) 2014 Huayan Wang

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

If you use this software, please cite [1]. If you use it with cycles turned on (default),
please cite [1] and [2].

[1] Huayan Wang and Daphne Koller: Subproblem-Tree Calibration: A Unified Approach to Max-Product Message Passing, ICML 2013
[2] Huayan Wang and Daphne Koller: A Fast and Exact Energy Minimization Algorithm for Cycle MRFs, ICML 2013

----------------------------------------------------------------------- */

  // Utility functions

  ///////////////////////////////////////////////////////////////
  // Fast min-sum
  // P. Felzenszwalb, J. McAuley
  // Fast Inference with Min-Sum Matrix Product
  // IEEE Transactions on Pattern Analysis and Machine Intelligence, Vol. 33, No. 12, December 2011
  // Code from the second author's website
  ///////////////////////////////////////////////////////////////

  /** Find the row minima of A and the column minima of B */
  void minrowcol(double* A, double* B, int a, int b, int c, int d,
		 double* minrowA, double* mincolB) {
    int i, j, k, idx;
    for (i = 0; i < a; i++){
      minrowA[i]=1e30;
    }
    for( j = 0; j < d; j++){
      mincolB[j]=1e30;
    }

    idx=0;
    for (j = 0; j < b; j++){
      for (i = 0; i < a; i++){
	minrowA[i] = MIN(minrowA[i], A[idx]);
	idx++;
      }
    }

    idx=0;
    for (k = 0; k < d; k++) {
      for (j = 0; j < c; j++){
	mincolB[k] = MIN(mincolB[k], B[idx]);
	idx++;
      }
    }

  }

  /** Normalize A and B by subtracting their row and column minima, respectively */
  void normalize(double* A, double* B, int a, int b, int c, double* minrowA,
		 double* mincolB) {

    int i,j,k,idx;
    idx=0;
    for( j = 0; j < b; j++)
      for( i = 0; i < a; i++){
	A[idx] -= minrowA[i];  // idx = i+j*a
	idx++;
      }
    idx=0;
    for( k = 0; k < c; k++)
      for( j = 0; j < b; j++){
	B[idx] -= mincolB[k];  // idx = j + k*b
	idx++;
      }
  }

  /** Undo the normalization applied above */
  void unnormalize(double* A, double* B, double* C, int a, int b, int c,
		   double* minrowA, double* mincolB) {
    int i,j,k,idx;
    idx=0;
    for( k = 0; k < c; k++)
      for( i = 0; i < a; i++){
	C[idx] += minrowA[i] + mincolB[k];
	idx++;
      }
    idx=0;
    for( j = 0; j < b; j++)
      for( i = 0; i < a; i++){
	A[idx] += minrowA[i];
	idx++;
      }
    idx=0;
    for( k = 0; k < c; k++)
      for( j = 0; j < b; j++){
	B[idx] += mincolB[k];
	idx++;
      }
  }



void pff_min_sum(int K0, int K1, int K2,  double* A, double* B, int* argmin_ik, double* C){

    int a=K0;
    int b=K1;
    int c=K2;

    //    double*  C = (double*)malloc(sizeof(double)*a*c);
    int i,j,k,pA,pB;
    for( i=0; i<a*c; ++i)
      C[i] = 0;


    int* AL = (int*)malloc(sizeof(int)*a);
    int* BL = (int*)malloc(sizeof(int)*c);

    int* Rowdone = (int*)malloc(sizeof(int)*a);
    int* Coldone = (int*)malloc(sizeof(int)*c);

    double* minrowA = (double*)malloc(sizeof(double)*a); // Min for each row of A (etc.)
    double* minrowB = (double*)malloc(sizeof(double)*b);
    double* mincolA = (double*)malloc(sizeof(double)*b);
    double* mincolB = (double*)malloc(sizeof(double)*c);



    for ( i = 0; i < a; i++)
      Rowdone[i] = 0;
    for ( k = 0; k < c; k++)
      Coldone[k] = 0;

    double infinity = 1e30;

    // Rescale the matrices.

    minrowcol(A, B, a, b, b, c, minrowA, mincolB);
    normalize(A, B, a, b, c, minrowA, mincolB);
    minrowcol(B, A, b, c, a, b, minrowB, mincolA);

    static double T_static  = 0.001;
    double Cmax;

    double T;
    int e0,e1,e2;

    /// The main part of the algorithm.
    while (1) {

      // Set all active entries to infinity.
      for ( i = 0; i < a; i++)
	for( k = 0; k < c; k++)
	  if (!Rowdone[i] && !Coldone[k])
	    C[i+k*a] = infinity;

      T = T_static;

      for ( j = 0; j < b; j++) {
	// build lists
	int numAL = 0;
	double mrB = minrowB[j];
	double mcA = mincolA[j];
	e0 = j*a;
	for ( i = 0; i < a; i++) {
	  if (!Rowdone[i] && A[e0] + mrB <= T)
	    AL[numAL++] = i;
	  e0++;
	}
	int numBL = 0;
	e1 = j;
	for ( k = 0; k < c; k++) {
	  if (!Coldone[k] && B[e1] + mcA <= T)
	    BL[numBL++] = k;
	  e1+=b; // e1 = j + k * b
	}

	for ( pA = 0; pA < numAL; pA++) {
	  int i = AL[pA];
	  for ( pB = 0; pB < numBL; pB++) {
	    int k = BL[pB];
	    e2 = i+k*a;
	    double val = A[i+j*a] + B[j+k*b];
	    if (val < C[e2]){
	      C[e2] = val;
	      argmin_ik[e2] = j;
	    }
	  }
	}
      }

      // check which rows/columns are done
      Cmax = -infinity;
      int done = 1;

      for ( i = 0; i < a; i++) {
	Rowdone[i] = 1;
	e2 = i;
	for ( k = 0; k < c; k++) {
	  if (C[e2] > T) {
	    Rowdone[i] = 0;
	    done = 0;
	  }
	  if (C[e2] > Cmax)
	    Cmax = C[e2];
	  e2+=a;
	}
      }
      e2=0;
      for ( k = 0; k < c; k++) {
	Coldone[k] = 1;
	for ( i = 0; i < a; i++) {
	  if (C[e2] > T) {
	    Coldone[k] = 0;
	    done = 0;
	  }
	  e2++;
	}
      }

      if (done)
	break;

      T_static *= 1.5;
    }

    // T grows exponentially, but we'll use its current value to help initialization next time
    if (Cmax)
      T_static = 0.75 * Cmax;
    if (T_static <= 0)
      T_static = 0.001;

    unnormalize(A, B, C, a, b, c, minrowA, mincolB);

    // Cleanup.
    free(AL);
    free(BL);
    free(Rowdone);
    free(Coldone);

    free(minrowA);
    free(minrowB);
    free(mincolA);
    free(mincolB);
  }


void print_vector(int* v, int n)
{
  printf("(");
  for(int i=0; i<n-1; i++)
    printf("%d, ", v[i]);
  printf("%d)\n", v[n-1]);
}

void print_vector(double* v, int n)
{
  printf("(");
  for(int i=0; i<n-1; i++)
    printf("%.2f, ", v[i]);
  printf("%.2f)\n", v[n-1]);
}

void print_matrix(double* v, int* n)
{
  printf("(\n");
  for(int k=0; k<n[0]; ++k){
    printf("   ");
    for(int l=0; l<n[1]; ++l){
      printf("%.2f, ", v[k+l*n[0]]);
    }
    printf("\n");
  }
  printf(")\n");  
}


void copyDblVec(const double *src, double* tar, int n)
{
  int i;
  for(i=0; i<n; ++i)
    tar[i] = src[i];
}


void setDblVal(double* vec, double val, int n)
{
  int i;
  for(i=0; i<n; ++i)
    vec[i]=val;
}


double minimize(double* vec, int n, int* argmin)
{
  int i;
  double minval = vec[0];
  *argmin = 0;
  for(i=1; i<n; ++i)
    if(vec[i] < minval){
      minval = vec[i];
      *argmin = i;
    }
  return minval;
}

double maximize(double* vec, int n, int* argmax)
{
  int i;
  double maxval = vec[0];
  *argmax = 0;
  for(i=1; i<n; ++i)
    if(vec[i] > maxval){
      maxval = vec[i];
      *argmax = i;
    }
  return maxval;
}


void rotate_cycle_assignment(int* a, int offset, int* result, int N)
  {
    int i;
    for(i=0; i<N; i++)
      result[i] = a[(i+offset+N)%N];
    result[N] = result[0];
  }

  void naive_min_sum(int K0, int K1, int K2,  double* A, double* B, int* argmin_ik, double* C){

    int i,j,k;
    for(i=0; i<K0*K2; ++i)
      C[i] = 1e30;
    
    for(i=0; i<K0; i++)
      for( j=0; j<K1; j++)
	for( k=0; k<K2; k++){
	  if(C[i+k*K0] > A[i+j*K0] + B[j+k*K1]){
	    C[i+k*K0] = A[i+j*K0] + B[j+k*K1];
	    argmin_ik[i+k*K0] = j;
	  }
	}
  }



void msg_i_ij_j(int K0, int K1, double* m0, double* M, double* m1)
{

  int i,j;
  int idx=0; // i+j*K0
  for( j=0; j<K1; j++){
    double minval = M[idx++] + m0[0];
    for(i=1; i<K0; i++){
      if(M[idx++] + m0[i] < minval){
	minval = M[idx-1] + m0[i];
      }
    }
    m1[j] = minval;
  }
}
 
void msg_j_ij_i(int K0, int K1, double* m1, double* M, double* m0)
{
  int i,j;
  int idx;
  for(i=0; i<K0; i++){
    double minval = M[i] + m1[0];
    idx = i + K0;  // i+j*K0
    for( j=1; j<K1; j++){
      if(M[idx] + m1[j] < minval){
	minval = M[idx] + m1[j];
      }
      idx += K0;
    }
    m0[i] = minval;
  }
}

void project_ij2i(int K0, int K1, double* M, int k1, double* m)
{
  for(int i=0; i<K0; i++)
    m[i] += M[i+k1*K0];
}

void project_ij2j(int K0, int K1, double* M, int k0, double* m)
{
   for(int i=0; i<K1; i++)
    m[i] += M[k0+i*K0];
}

// M(X0, X1) is a (K0 x K1) matrix, return \min_X0 M(X0, X1)
void reduce_i(int K0, int K1, double* M, double* result)
{
  int i,j;
  int idx=0; // i+j*K0
  for( j=0; j<K1; j++){
    double minval = M[idx++];
    for(i=1; i<K0; i++){
      if(M[idx++] < minval){
	minval = M[idx-1];
      }
    }
    result[j] = minval;
  }
}

// M(X0, X1) is a (K0 x K1) matrix, return \min_X1 M(X0, X1)
void reduce_j(int K0, int K1, double* M, double* result)
{
  int i,j;
  int idx;
  for(i=0; i<K0; i++){
    double minval = M[i];
    idx = i + K0;  // i+j*K0
    for( j=1; j<K1; j++){
      if(M[idx] < minval){
	minval = M[idx];
      }
      idx += K0;
    }
    result[i] = minval;
  }
}

// tar += s*M' (tar is K0xK1, M is K1xK0)
  void add_matrixT(int K0, int K1, double* tar, 
		   double* M, double s){
    int l,k;
    int idx=0;
    int idxT;
    for( l=0; l<K1; l++){
      idxT = l;
      for( k=0; k<K0; k++){
	tar[idx] += s*M[idxT]; // k+l*K0, l+k*K1
	idx++;
	idxT += K1;
      }
    }
  }

  // add v(X0) to M(X0,X1) (a K0XK1 matrix)
  void add_i2ij(int K0, int K1, double* v, 
		double* M, double factor) 
  {
    int i,j;
    int idx=0;
    for(j=0; j<K1; j++)
      for(i=0; i<K0; i++)
	M[idx++] += v[i]*factor; // M[i+j*K0]
  }

  void add_j2ij(int K0, int K1, double* v, 
		double* M, double factor) 
  {
    int i,j;
    int idx=0;
    for(j=0; j<K1; j++)
      for(i=0; i<K0; i++)
	M[idx++] += v[j]*factor; // M[i+j*K0]
  }


//  class Cycle
//  {
//  public:

typedef struct Cycle{



    int N; // length of cycle
    int Nmax;
    int* card;   // has length size+1

    // the solvers implemented DO support different cardinalities of variables
    // but you need to construct/read-in your own problem (as in randomCycle())

    double obj; // min energy
    int* assignment; // has length size+1, with the two ends agree

  //    int* unary_msg;
    double** ts;
    double** tp; // theta_p, pairwise energy terms
    double** cache;

    // min-marginals
    double** mmarg;
    // mmarg[0~N-1]: node min-marginals of variable [0~N-1]
    // mmarg[N]:   edge min-marginal (var_id[N-1], var_id[0])
    // mmarg[N+1]: edge min-marginal (var_id[0], var_id[1])
    // ...
    // mmarg[2N-1]: edge min-marginal (var_id[N-2], var_id[N-1])

  //  private:
    double* msg;
    double *TA_static;
    double *TB_static;
    int L;
    int Kmax;
    int **backtracer; 
    int* AL;
    int* BL;
    double** mincolA;
    int* OldnumAL;
    int* OldnumBL;

  int* rot_assignment;
  double** forward_msg;
  int** forward_argmin;

    // generate a random cycle
  //    void randomCycle(int, int, double);
    void runFullMinSum(int d, bool fast);
    void runFastSolver(int d);
  void writeCycle(const char* filename);
    // you get Felzenswalb's method using fast = true in FullMinSum
    // FastSolver does not give exact min-marginals
    // running both solvers writes to 'obj' and 'assignment'
    // 'd' specifies direction of message passing (or how to triangulate cycle)
    // it can be 0, 1, ..., N-1, where N is cycle size
    // it will determine which min-marginal(s) we will get as a by-product
    // d==0 : get min-marginal on (var_id[N-1], var_id[0])
    // d==1 : get min-marginal on (var_id[0], var_id[1])
    // ......
 

  void freeMemory();
  void allocateMemory(int,int);

    void reset_thresholds(int* Aidx, 
			  int* Bidx);
  void addin_singleton(int i);
    void pass_unary_message(int i);
    void reparameterize_singleton(int i);
    void reparameterize_singleton_reverse(int i);
    void fast_solver_preprocessing(int last_factor_id, 
				      int* Aidx, 
				      int* Bidx,
				      double* factor_offset, 
				      double* global_const, 
				      double* rlb,
				      double** drlb,
				      double** mincolA);

  } Cycle ;


  void Cycle::runFullMinSum(int d, bool fast) 
  {
    //    clock_t t = clock();
    int i;
    //    assert(d>=0 && d<N);

    if(ts){

      /*      for(i=0; i<N; ++i)
	      addin_singleton(i);*/

       for(i=0; i<N; ++i){
	add_i2ij(card[i], card[i+1], ts[i], tp[i], 1.0);
	setDblVal(ts[i], 0.0, card[i]);
       }

    }

    // TODO remove
    int tmp;
    int tmp1 = card[(N-1+d)%N] * card[d];
    minimize(tp[(N-1+d)%N], tmp1, &tmp);
    minimize(mmarg[N+d], tmp1, &tmp);
    ///////////

    copyDblVec(tp[(N-1+d)%N], mmarg[N+d], card[(N-1+d)%N] * card[d]);
    forward_msg[0] = tp[d];

    // forward pass
    for( i=1; i < N-1; i++){

      if(fast){
	//	printf("pff_min_sum ..."); fflush(stdout);
	pff_min_sum(card[d], card[(d+i)%N], card[(d+i+1)%N],
		    forward_msg[i-1], 
		    tp[(i+d)%N], 
		    forward_argmin[i],
		    forward_msg[i]);
	//	printf("done\n");

      }else{

	// (d, d+i+1)                     (d, d+i)        (d+i, d+i+1)
	naive_min_sum(card[d], card[(d+i)%N], card[(d+i+1)%N],
		      forward_msg[i-1], 
		      tp[(i+d)%N], 
		      forward_argmin[i],
		      forward_msg[i]);
      }
    }

    // (forward_msg[N-2])' + tp[N-1] -> (1,N)
    add_matrixT(card[(N-1+d)%N], card[d], mmarg[N+d], forward_msg[N-2], 1.0);


    // back trace for assignments
    int l;
    obj = minimize(mmarg[N+d], card[(N-1+d)%N]*card[d], &l); // l = a[N-1] + a[0]*card[N-1]
    //
    for( i=0; i<N+1; ++i)
      rot_assignment[i] = -1;
    rot_assignment[N-1]=l%card[(d+N-1)%N];
    rot_assignment[0] = (l-rot_assignment[N-1])/card[(d+N-1)%N];
    rot_assignment[N] = rot_assignment[0];
    for( i=N-2; i>=1; i--)
      rot_assignment[i] = forward_argmin[i][rot_assignment[0] + rot_assignment[i+1]*card[d]];
    rotate_cycle_assignment(rot_assignment, -d, assignment, N);
    reduce_j(card[(N-1+d)%N], card[d], mmarg[N+d], mmarg[(N-1+d)%N]);
    reduce_i(card[(N-1+d)%N], card[d], mmarg[N+d], mmarg[d]);

    /*
    int tmp;
    assert(minimize(mmarg[(N-1+d)%N], card[(N-1+d)%N], &tmp) == obj);
    assert(minimize(mmarg[d], card[d], &tmp) == obj);
    assert(minimize(mmarg[N+d], card[d]*card[(N-1+d)%N], &tmp) == obj);


    ////////////////////////////////////////////////
    double energy=0;
    for(i=0; i<N; ++i){
      energy += ts[i][assignment[i]];
      energy += tp[i][assignment[i]+assignment[i+1]*card[i]];
    }
    assert(fabs(energy-obj)<1e-6);
    */
    //////////////////////////////////////////////// 
   //  solver_time  += ((double)clock()-t)/CLOCKS_PER_SEC;
    //  solver_cnt++;


  }


void Cycle::addin_singleton(int i){

    int K0 = card[(N+i-1)%N];
    int K1 = card[i];
    int K2 = card[(i+1)%N];
    int k,l;

    //////////////////////////////////////////////////////////////
    for( l=0; l<K1; l++){ // for each assignment of node i
      double u = 0.5*ts[i][l];
      ts[i][l] = 0;
      int e0=l;
      int e1=l*K0;
      for( k=0; k<K2; k++){
	tp[i][e0] += u;
	e0+=K1;
      }
      for( k=0; k<K0; k++){
	tp[(i-1+N)%N][e1] += u;
	e1++;
      }
    }
    /////////////////////////////////////////////////////////////////
}

void Cycle::pass_unary_message(int i){

  // TODO 
    int K0 = card[(N+i-1)%N];
    int K1 = card[i];
    int K2 = card[(i+1)%N];
    int k,l;

    for( l=0; l<K1; l++){ // for each assignment of node i
      double min1=1e30;
      //      double min2=1e30;
      int e0 = l;
      int e1 = l*K0;
      for( k=0; k<K2; k++){
	min1 = MIN(min1, tp[i][e0]); // e0 = l+k*K1
	e0+=K1;
      }
      /*
      for( k=0; k<K0; k++){
	min2 = MIN(min2, tp[(i-1+N)%N][e1]); // e1 = k+l*K0
	e1++;
      }
      */
      //      double u = 0.5*(min2-min1);
      e0=l;
      e1=l*K0;
      for( k=0; k<K2; k++){
	tp[i][e0] -= min1;
	e0+=K1;
      }
      for( k=0; k<K0; k++){
	tp[(i-1+N)%N][e1] += min1;
	e1++;
      }
    }
}

  // the Omega+ operator defined in the paper
  void Cycle::reparameterize_singleton(int i) 
  {
    int K0 = card[(N+i-1)%N];
    int K1 = card[i];
    int K2 = card[(i+1)%N];
    int k,l;

    for( l=0; l<K1; l++){ // for each assignment of node i
      double min1=1e30;
      double min2=1e30;
      int e0 = l;
      int e1 = l*K0;
      for( k=0; k<K2; k++){
	min1 = MIN(min1, tp[i][e0]); // e0 = l+k*K1
	e0+=K1;
      }
      for( k=0; k<K0; k++){
	min2 = MIN(min2, tp[(i-1+N)%N][e1]); // e1 = k+l*K0
	e1++;
      }
      double u = 0.5*(min2-min1);
      e0=l;
      e1=l*K0;
      for( k=0; k<K2; k++){
	tp[i][e0] += u;
	e0+=K1;
      }
      for( k=0; k<K0; k++){
	tp[(i-1+N)%N][e1] -=u;
	e1++;
      }
    }
  }

  // the Omega- operator defined in the paper
  void Cycle::reparameterize_singleton_reverse(int i) 
  {
    int K0 = card[(N+i-1)%N];
    int K1 = card[i];
    int K2 = card[(i+1)%N];
    int k,l;
    for( l=0; l<K1; l++){ // for each assignment of node i
      double min1=1e30;
      double min2=1e30;
      int e0 = l;
      int e1 = l*K0;
      for( k=0; k<K2; k++){
	min1 = MIN(min1, tp[i][e0]); // e0 = l+k*K1
	e0+=K1;
      }
      for( k=0; k<K0; k++){
	min2 = MIN(min2, tp[(i-1+N)%N][e1]); // e1 = k+l*K0
	e1++;
      }
      double u = 0.5*(min2-min1);
      e0=l;
      e1=l*K0;

      for( k=0; k<K2; k++){
	tp[i][e0] -= u;
	e0+=K1;
      }
      for( k=0; k<K0; k++){
	tp[(i-1+N)%N][e1] += u;
	e1++;
      }
    }
  }

  void Cycle::freeMemory()
  {
    int i;
    N = Nmax;
    L = N-2;
    if(L>1){
      for(i=0; i<L-1; ++i)
	free(cache[i]);
      free(cache);
    }
    free(tp);
    for(i=0; i<2*N; ++i)
      free(mmarg[i]);
    free(mmarg);
    free(assignment);
    for (i = 0; i < L; i++){
      free(backtracer[i]);
      free(mincolA[i]);
    }
    free(backtracer);
    free(AL);
    free(BL);
    free(OldnumAL);
    free(OldnumBL);
    free(TA_static);
    free(TB_static);
    free(mincolA);
    free(msg);
    for( i=1; i<N-1; i++){
      free(forward_argmin[i]);
      free(forward_msg[i]);
    }
    free(forward_msg);
    free(forward_argmin);
    free(rot_assignment);
  }

void Cycle::allocateMemory(int Nmax_, int Kmax_)
{
    int i;
    Nmax = Nmax_;
    Kmax=Kmax_;
    N = Nmax;
    L = N-2;
    rot_assignment = (int*)malloc(sizeof(int)*(N+1));
    msg = (double*)malloc(sizeof(double)*2*Kmax);
    tp = (double**)malloc(sizeof(double*) * (N+L-1));
    if(L>1){
      cache = (double**)malloc(sizeof(double*) * (L-1));
      for(i=0; i<L-1; ++i)
	cache[i] = (double*)malloc(sizeof(double)*Kmax*Kmax);
    }
    assignment =  (int*)malloc(sizeof(int)*(N+1));
    mmarg = (double**)malloc(sizeof(double*) * 2*N);
    for(i=0; i<2*N; ++i)
      mmarg[i] = (double*)malloc(sizeof(double)*Kmax*Kmax);
    backtracer = (int**)malloc(sizeof(int*)*L);
    AL = (int*)malloc(sizeof(int)*Kmax);
    BL = (int*)malloc(sizeof(int)*Kmax);
    OldnumAL = (int*)malloc(sizeof(int)*Kmax);
    OldnumBL = (int*)malloc(sizeof(int)*Kmax);
    TA_static = (double*)malloc(sizeof(double)*L);
    TB_static = (double*)malloc(sizeof(double)*L);
    mincolA = (double**)malloc(sizeof(double*)*L);
    for (i = 0; i < L; i++){
      backtracer[i] = (int*)malloc(sizeof(int)*Kmax*Kmax);
      TA_static[i] = -1;
      TB_static[i] = -1;
      mincolA[i] = (double*)malloc(sizeof(double)*Kmax);
    }


    forward_msg = (double**)malloc(sizeof(double*)*(N-1));
    forward_argmin = (int**)malloc(sizeof(int*)*(N-1));
    for( i=1; i<N-1; i++){
      forward_argmin[i] = (int*)malloc(sizeof(int)*Kmax*Kmax);
      forward_msg[i] = (double*)malloc(sizeof(double)*Kmax*Kmax);
    }
  }

  // set TA_static and TB_static
  // note: this private function is not optimized for human readability
  void Cycle::reset_thresholds(int* Aidx, 
			       int* Bidx)
  {
    int i, j, s, idx;
    int K0,K1;
    double factor_offset;
    double* meanVal = (double*)malloc(sizeof(double)*L);
    for( i=0; i<L; ++i)
      meanVal[i] = 0;
    for( s=L-1; s>0; s--){
      idx=0;
      factor_offset = 1e30;
      K0 = card[Bidx[s]];
      K1 = card[(Bidx[s]+1)%N];
      for( i=0; i<K1; i++){
	for( j=0; j<K0; j++){
	  meanVal[s] += tp[Bidx[s]][idx];
	  factor_offset = MIN(factor_offset, tp[Bidx[s]][idx]);
	  idx++;
	}
      }
      meanVal[s] = meanVal[s]/(K0*K1) - factor_offset;
    }
    double meanValA0=0;
    double meanValB0=0;

    idx=0;
    K0 = card[Aidx[0]];
    K1 = card[(Aidx[0]+1)%N];
    factor_offset = 1e30;  
    for( i=0; i<K1; i++){
      for( j=0; j<K0; j++){
	meanValA0 += tp[Aidx[0]][idx];
	factor_offset = MIN(factor_offset, tp[Aidx[0]][idx]); // idx = j+i*K0
	idx++;
      }
    }
    meanValA0 = meanValA0/(K0*K1) - factor_offset;
    idx=0;
    K0 = card[Bidx[0]];
    K1 = card[(Bidx[0]+1)%N];
    factor_offset = 1e30;
    for( i=0; i<K1; i++){
      for( j=0; j<K0; j++){
	meanValB0 += tp[Bidx[0]][idx];
	factor_offset = MIN(factor_offset, tp[Bidx[0]][idx]);
	idx++;
      }
    }
    meanValB0 = meanValB0/(K0*K1) - factor_offset;
    meanVal[0] = (meanValA0+meanValB0)/2.0;

    double batch_size = 0.1;
    TB_static[0] = meanValB0 * batch_size;
    TA_static[0] = meanValA0 * batch_size;
    for( s=1; s<L; s++){
      TB_static[s] = meanVal[s] * batch_size;
      TA_static[s] = TA_static[s-1]+TB_static[s-1];
    }
    for( s=0; s<L; s++){
      TA_static[s] = MAX(TA_static[s], 1e-6);
      TB_static[s] = MAX(TB_static[s], 1e-6);
    }
    free(meanVal);
  }

  // compute the messages in cycle edges (small delta in the paper)
  // and other stuff
  // note: this private function is not optimized for human readability
  void Cycle::fast_solver_preprocessing(int last_factor_id, 
					   int* Aidx, 
					   int* Bidx,
					   double* factor_offset, 
					   double* global_const, 
					   double* rlb,
					   double** drlb,
					   double** mincolA)
  {

    int i,j, s;
    for( i=0; i<2*Kmax; i++)
      msg[i] = 1e30;
    int idx=0;
    int K0,K1;
    K0 = card[last_factor_id];
    K1 = card[(last_factor_id+1)%N];
    for( i=0; i<K1; i++)
      for( j=0; j<K0; j++){
	msg[j] = MIN(msg[j], tp[last_factor_id][idx]); // j+i*K0
	idx++;
      }
    for( i=0; i<K0; i++){
      rlb[L] = MIN(rlb[L], msg[i]);
      drlb[L][i] = msg[i];
    }
    for( s=L-1; s>0; s--){
      idx=0;
      factor_offset[Bidx[s]] = 1e30;
      K0 = card[Bidx[s]];
      K1 = card[(Bidx[s]+1)%N];
      for( i=0; i<K1; i++){
	for( j=0; j<K0; j++){
	  msg[Kmax+j] = MIN(msg[Kmax+j], tp[Bidx[s]][idx] + msg[i]); 
	  factor_offset[Bidx[s]] = MIN(factor_offset[Bidx[s]], tp[Bidx[s]][idx]);
	  idx++;
	}
      }
      for( j=0; j<K0; j++){
	msg[j] = msg[Kmax+j] - factor_offset[Bidx[s]];
	rlb[s] = MIN(rlb[s], msg[j]);
	drlb[s][j] = msg[j];
	msg[Kmax+j] = 1e30;
      }
      (*global_const) += factor_offset[Bidx[s]];
    }

    idx=0;
    K0 = card[Aidx[0]];
    K1 = card[(Aidx[0]+1)%N];
    factor_offset[Aidx[0]] = 1e30;  
    for( i=0; i<K1; i++){
      for( j=0; j<K0; j++){
	mincolA[0][i] = MIN(mincolA[0][i], tp[Aidx[0]][idx]);
	factor_offset[Aidx[0]] = MIN(factor_offset[Aidx[0]], tp[Aidx[0]][idx]);
	idx++;
      }
    }
    for( i=0; i<K1; i++){
      mincolA[0][i] -= factor_offset[Aidx[0]];
    }
    idx=0;
    K0 = card[Bidx[0]];
    K1 = card[(Bidx[0]+1)%N];
    factor_offset[Bidx[0]] = 1e30;
    for( i=0; i<K1; i++){
      for( j=0; j<K0; j++){
	msg[Kmax+j] = MIN(msg[Kmax+j], tp[Bidx[0]][idx] + msg[i]);
	factor_offset[Bidx[0]] = MIN(factor_offset[Bidx[0]], tp[Bidx[0]][idx]);
	idx++;
      }
    }
    for( j=0; j<K0; j++){
      drlb[0][j] = msg[Kmax+j] - factor_offset[Bidx[0]];
      rlb[0] = MIN(rlb[0], drlb[0][j]);
    }
    (*global_const) += factor_offset[Aidx[0]];
    (*global_const) += factor_offset[Bidx[0]];
  }

  void Cycle::writeCycle(const char* filename)
  {
    FILE* fp = fopen(filename, "w");
    fprintf(fp, "%d\n", N);
    for(int i=0; i<N; ++i){
      fprintf(fp, "%d ", card[i]);
      for(int j=0; j<card[i]; ++j)
	fprintf(fp, "%f ", ts[i][j]);
      fprintf(fp, "\n");
    }
    fprintf(fp, "\n");
    for(int i=0; i<N; ++i){
      fprintf(fp, "%d ", card[i] * card[(i+1)%N]);
      for(int j=0; j<card[i] * card[(i+1)%N]; ++j)
	fprintf(fp, "%f ", tp[i][j]);
      fprintf(fp, "\n");
    }
    fclose(fp);
  }

  // note: this function is not optimized for human readability,
  // in most cases you can use it as a black box as shown in cycle.cpp
  // however, if you insist on parsing it, it should help to
  // first try to understand the fast min-sum code in thirdparty.h,
  // which is considerably simpler, and shares many common variable names
  // and structures with this one
  void Cycle::runFastSolver(int d)
  {
    int i,j,s,k,pA,pB;
    double *TA = (double*)malloc(sizeof(double)*L);
    double *TB = (double*)malloc(sizeof(double)*L);
    int last_factor_id = (N-1+d)%N;
    long final_assignment = -1;
    double* factor_offset = (double*)malloc(sizeof(double)*(N+L-1));
    for(i=0; i<N+L-1; ++i)
      factor_offset[i] = 0;

    int* Aidx = (int*)malloc(sizeof(int)*L);
    int* Bidx = (int*)malloc(sizeof(int)*L);
    for(i=0; i<L; ++i){
      Aidx[i] = -1;
      Bidx[i] = -1;
    }

    Aidx[0] = d;
    for (i = 0; i < L; i++){
      Bidx[i] = (i+d+1)%N;
      for( j=0; j<Kmax; j++){
	mincolA[i][j] = 1e30;
      }
      if(i>0){
	Aidx[i] = N+(i-1);
	for( j=0; j<Kmax*Kmax; ++j)
	  tp[Aidx[i]][j] = 1e30;
      }
    }
    int idx;

    double* rlb = (double*)malloc(sizeof(double)*(L+1));
    double** drlb = (double**)malloc(sizeof(double*)*(L+1));
    for(i=0; i<L+1; ++i){
      rlb[i] = 1e30;
      drlb[i] = (double*)malloc(sizeof(double)*Kmax);
      for( j=0; j<Kmax; ++j)
	drlb[i][j] = 1e30;
    }

    double global_const = 0.0;

    if(ts)
      for(i=0; i<N; ++i)
	addin_singleton(i);

    for(i=(d+N-1); i>=d; i--)
      reparameterize_singleton(i%N);


    if(TA_static[0] < 0)
      reset_thresholds(Aidx, Bidx);
    fast_solver_preprocessing(last_factor_id, Aidx, Bidx, 
			      factor_offset, &global_const, rlb, drlb, mincolA);
    double ub = 1e30;
    int K0, K1, K2;
    K0 = card[(N+d-1)%N];
    K1 = card[d];
    /*
    for(j=0; j<K0*K1; ++j)
      mmarg[N+d][j] = 1e30;
    */
    bool done;
    for( s=0; s<L; s++){
      TA[s] = TA_static[s]/1.5;
      TB[s] = TB_static[s]/1.5;
    }
    int*  sdone = (int*)malloc(sizeof(int)*L);
    int*  updated = (int*)malloc(sizeof(int)*L);
    for(i=0; i<L; ++i){
      sdone[i] = 0;
      updated[i] = 0;
    }
    // for faster index computing
    int e0; // i+j*K
    int e1; // j+k*K
    int e2; // i+k*K
    int e3; // k+i+K
    do{
      done = 1;
      for( s=0; s<L; s++){
	if (sdone[s])
	  continue;
	done = 0;
	do{
	  updated[s] = 0;
	  if(TB[s] < ub-rlb[s+1] || TA[s] < ub-rlb[s]){
	    TB[s] = MIN(TB[s]*1.5, MAX(1e-6, ub-rlb[s+1]));
	    TA[s] = MIN(TA[s]*1.5, MAX(1e-6, ub-rlb[s]));
	  }else{
	    if(s == 0){
	      sdone[s] = 1;
	      break;
	    }else if(!updated[s-1]){
	      sdone[s] = sdone[s-1];
	      break;
	    }
	  }
	  K0 = card[d];
	  K1 = card[Bidx[s]];
	  K2 = card[(Bidx[s]+1)%N];
	  for( j=0; j<K1; j++){
	    int numAL = 0;
	    int numBL = 0;
	    double mcA = mincolA[s][j];

	    // extract from entries from A
	    e0 = j*K0;
	    for(i=0; i<K0; i++){

	      if(tp[Aidx[s]][e0] - factor_offset[Aidx[s]] <= 
		 MIN(TA[s], ub-drlb[s][j])){
		AL[numAL++] = i;
	      }
	      e0++;
	    }
	    // extract from entries from B
	    e1 = j;
	    for( k=0; k<K2; k++){
	      if(tp[Bidx[s]][e1] - factor_offset[Bidx[s]] + mcA <= 
		 MIN(TB[s], ub-drlb[s+1][k])) {
		BL[numBL++] = k;
	      }
	      e1+=K1;
	    }

	    // perform partial min-sum
	    for( pA=0; pA<numAL; pA++){
	      i = AL[pA];
	      for( pB=0; pB<numBL; pB++){
		k = BL[pB];
		e0 = i+j*K0;
		e1 = j+k*K1;
		e2 = i+k*K0;
		double val = tp[Aidx[s]][e0] + tp[Bidx[s]][e1] - 
		  factor_offset[Aidx[s]] - factor_offset[Bidx[s]];
		if(s < L-1){ // mid of the clique chain
		  if(val < tp[Aidx[s+1]][e2] && val < ub - drlb[s+1][k])
		    {
		      updated[s] = 1;
		      backtracer[s][e2] = j;
		      tp[Aidx[s+1]][e2] = val;
		      mincolA[s+1][k] = MIN(val, mincolA[s+1][k]);
		    }
		}else{ // end of the clique chain
		  e3 = k+i*K2;
		  double lf = tp[last_factor_id][e3] - 
		    factor_offset[last_factor_id];
		  //		  mmarg[N+d][e3] = MIN(mmarg[N+d][e3], val+lf);
		  if(val+lf < ub){
		    updated[s] = 1;
		    backtracer[s][e2] = j;
		    final_assignment = e2;
		    ub = val+lf;
		  }
		}
	      }
	    }
	  }
	  if(updated[s])
	    break;
	  else if(TB[s] >= ub-rlb[s+1] && TA[s] >= ub-rlb[s]){
	    sdone[s] = s==0?1:sdone[s-1];
	    break;
	  }
	}while( 1 ); // you should hope it will end :)
      }
    }while(!done);

    // memorize these value for solving next cycle faster
    // in a dual decomposition environment
    for(i=0; i<L; i++){
      TA_static[i] = TA[i]/1.5;
      TB_static[i] = TB[i]/1.5;
    } 

    idx=0;
    K0 = card[(N+d-1)%N];
    K1 = card[d];
    obj = ub + global_const; // minimum energy

    //  trace back to get assignment
    int* rot_assignment = (int*)malloc(sizeof(int)*(N+1));
    for(i=0; i<N+1; ++i)
      rot_assignment[i] = -1;
    rot_assignment[0] = final_assignment%card[d];
    rot_assignment[N-1] = (final_assignment-rot_assignment[0])/card[d];
    for(i=N-2; i>=1; i--){
      long idx = rot_assignment[0] + rot_assignment[i+1]*card[d];
      rot_assignment[i] = backtracer[i-1][idx];
    }
    rot_assignment[N] = rot_assignment[0];
    rotate_cycle_assignment(rot_assignment, -d, assignment, N);

    free(rot_assignment);
    for(i=0; i<L+1; ++i)
      free(drlb[i]);
    free(Aidx);
    free(Bidx);
    free(rlb);
    free(drlb);
    free(sdone);
    free(updated);
    free(factor_offset);
    free(TA);
    free(TB);
  }


