/* -----------------------------------------------------------------------

The MIT License (MIT)

Copyright (c) 2014 Huayan Wang

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

If you use this software, please cite [1]. If you use it with cycles turned on (default),
please cite [1] and [2].

[1] Huayan Wang and Daphne Koller: Subproblem-Tree Calibration: A Unified Approach to Max-Product Message Passing, ICML 2013
[2] Huayan Wang and Daphne Koller: A Fast and Exact Energy Minimization Algorithm for Cycle MRFs, ICML 2013

----------------------------------------------------------------------- */

#pragma once
#ifndef DUALBCA
#define DUALBCA

#define MAX(A,B) ((A)>(B)?(A):(B))
#define MIN(A,B) ((A)<(B)?(A):(B))

#include <string.h>
#include <limits.h>
#include <time.h>
#include <stdlib.h>
#include <vector>
#include "cycle.h"

#define INFVAL 1e8
#define DEBUG_THRESHOLD 1e-6

enum{
  PRIMAL_INDEPENDENT,
  PRIMAL_JIGSAW_FIXED,
  PRIMAL_JIGSAW_RAND,
  PRIMAL_CONDITIONAL_FIXED,
  PRIMAL_CONDITIONAL_RAND,
  PRIMAL_ROUNDOFF,
};

enum{
  GRAPH_BETHE,
  GRAPH_DC,
};

enum{
  TREE_PAIR,      //0
  TREE_STAR,      //1
  TREE_STARP,     //2
  TREE_SP1,
  TREE_SP2,
  TREE_SP3,
  TREE_SP4,       //6
};

enum{
  ALLOCATION_UNIFORM,  // 0
  ALLOCATION_UNARY,    // 1
  ALLOCATION_DEGREE,   // 2
  ALLOCATION_ROOT,     //3
  ALLOCATION_LEAF,       //4
  ALLOCATION_NEXT_LEAF,   //5
};

enum{
  FUNC_TYPE_TABLE,   // 0
  FUNC_TYPE_POTTS,   // 1
  FUNC_TYPE_CYCLE,   // 2
  FUNC_TYPE_POTTS0,  // 3
  FUNC_TYPE_TL1,     // 4
  FUNC_TYPE_TL2,     // 5
  FUNC_TYPE_ISING,   // 6
  FUNC_TYPE_GRIDPAIR_LMAX_HARD, // 7
};

enum{
  ACTION_STC,
  ACTION_ADD_CYCLE,
  ACTION_SUBSPACE_SEARCH,
  ACTION_RENEW_SOLUTION
};

typedef struct Function {
  int type;
  int size_val;
  double* val;
} Function;

typedef struct SubProblem { // factors

  // basic info
  int type;
  int is_binary;
  int num_variable;
  int* var_id;
  int* card;
  int func_id;
  double* val;

  // cached solution
  double min_energy;
  int min_energy_valid;
  int* sol;

  // dual variables
  int num_msg;
  double** msg;
 
} SubProblem;

typedef struct SubProblemTree {

  // representation of the subproblem tree
  int max_num_edges;
  int start_idx;
  int** sp_id;     // (child, parent)
  int* phi_id;
  double *subtree_weight;

  // cache space used for message passing
  double *subtree_weight_dynamic;

  // cache space used for building subproblem tree only
  int nb_size;
  int* nb_sp;
  int* nb_phi;
  int* candidates;
  int* color;
  int* parent_sp;
  int* parent_phi;

} SubProblemTree;

struct MapParameters;

typedef struct MapProblem {

  // representation of the problem
  int num_variable;
  int* card;
  int num_subproblem;
  SubProblem** sp;
  int num_function;
  Function* func;
  double gb_constant;

  int num_nonunary_subproblem;
  int* unary_subproblem_id;
  double* is_unary_subproblem;

  // used only when cycles are added
  int num_edge_factor;
  int max_cycle_length;
  int max_cardinality;
  int is_grid; // 0 == 'no', 4 == 'n-4 grid', 8 == 'n-8 grid'
  int grid_width;
  int grid_height;
  int* sp2phi;
  int* sp2cycle;
  int* grid_edge2sp;
  int* var_degree;
  int** var_nb;
  int** var_nb_edge;
  int cycle_dict_size;
  int* cycle_size;
  double* cycle_score;
  int** cycle_nodes;
  int** cycle_edges;
  int used_cycles;
  Cycle* cy_solver;

  // representation of the subproblem multi-graph
  int num_phi; 
  int** scope_of_phi;
  int* num_sp_of_phi;
  int** sp_of_phi;
  int* num_phi_of_sp;
  int** phi_of_sp;
  SubProblemTree* sptree;
  SubProblemTree* next_sptree; 
  double* subproblem_degree;

  int initialized;
  int cycle_initialized;

  int* sol;
  double primal_obj;
  double dual_obj;
  double last_dual_gap;
  double dual_gap_decrease;
  double init_primal_obj;
  double init_dual_obj;
  int next_action;
  int last_action;
  int msg_cnt;

  MapParameters* param; // hack for convenience

} MapProblem ;

typedef struct MapParameters {

  int primal_method;

  int     stc_inner_iterations;
  double  dual_gap_threshold;
  double  add_cycle_batchsize;
  double  start_cycle_batchsize;
  int write_to_log_when_nocycle;

  int bipartite_graph;
  int tree_selection;
  int allocation_method;

  int use_fast_solver;
  int pair_msg_for_high_order_clique;

  void* buffer; // hack to avoid mallocating repeatedly

} MapParameters ;

void readParameters(MapParameters* parameters, int argc, char** argv)
{
  parameters->stc_inner_iterations = 100000 / 2;
  parameters->dual_gap_threshold = 0.01;

  parameters->bipartite_graph = GRAPH_DC; 
  parameters->tree_selection = TREE_STARP;
  parameters->allocation_method = ALLOCATION_UNIFORM;
  parameters->write_to_log_when_nocycle = 1;

  parameters->primal_method = PRIMAL_CONDITIONAL_RAND;

  parameters->add_cycle_batchsize =  1; // turned on

  parameters->use_fast_solver = 1;
  parameters->start_cycle_batchsize = 0;
  parameters->pair_msg_for_high_order_clique = 1;

  int i;
  for(i=0;(i<argc) && ((argv[i])[0] == '-');i++) {
    switch ((argv[i])[1]) 
      { 
      case 'a': i++; parameters->allocation_method = atoi(argv[i]); break;
      case 'i': i++; parameters->stc_inner_iterations = atoi(argv[i]); break;
      case 'f': i++; parameters->use_fast_solver = atoi(argv[i]); break;
      case 's': i++; parameters->bipartite_graph = atoi(argv[i]); break;
      case 't': i++; parameters->tree_selection = atoi(argv[i]); break;
      case 'g': i++; parameters->dual_gap_threshold = atof(argv[i]); break;
      case 'c': i++; parameters->add_cycle_batchsize = atof(argv[i]); break;
      case 'b': i++; parameters->start_cycle_batchsize = atof(argv[i]); break;
      case 'p': i++;  parameters->primal_method = atoi(argv[i]); break;
      case 'w': i++;  parameters->write_to_log_when_nocycle = atoi(argv[i]); break;
      default: printf("\nUnrecognized option %s!\n\n",argv[i]);
	assert(0);
      }
  }

  // hack
  parameters->add_cycle_batchsize = -(parameters->add_cycle_batchsize);
}

void promotePairwiseScopes(int* sp, int* phi, int n, int num_var)
{
  int tmp, s=0, e=n-1;
  while(s<e){
    if(phi[s] < num_var){
      tmp = sp[s];
      sp[s] = sp[e];
      sp[e] = tmp;
      tmp = phi[s];
      phi[s] = phi[e];
      phi[e] = tmp;
      --e;
    }else{
      ++s;
    }
  }
}

// permutate lists v0 and v1 (keep the correspondences)
void permutateLists(int* v0, int* v1, int n)
{
  int i,tmp;
  for(i=1; i<n; ++i){
    int j = rand()%(i+1); // a number in {0,1,...,i}
    if(j!=i){
      tmp = v0[i];
      v0[i] = v0[j];
      v0[j] = tmp;
      tmp = v1[i];
      v1[i] = v1[j];
      v1[j] = tmp;
    }
  }
}

// get a permutation of {0, 1, ... n-1}
void randPermutation(int* v, int n)
{
  int i;
  for(i=0; i<n; ++i)
    v[i] = i;

  for(i=1; i<n; ++i){
    int j = rand()%(i+1); // a number in {0,1,...,i}
    if(j!=i){
      int tmp = v[i];
      v[i] = v[j];
      v[j] = tmp;
    }
  }
}


const char *getFilenameExt(const char *filename) {
  const char *dot = strrchr(filename, '.');
  if(!dot || dot == filename) return "";
  return dot + 1;
}

void copyIntVec(int *src, int* tar, int n)
{
  int i;
  for(i=0; i<n; ++i)
    tar[i] = src[i];
}


void append2IntVec(int a, int** vec, int* n)
{
  (*n)++;
  *vec = (int*)realloc(*vec, sizeof(int)*(*n));
  (*vec)[(*n)-1] = a;
}

void append2TwoIntVec(int a, int b, int** vec0, int** vec1, int* n)
{
  (*n)++;
  *vec0 = (int*)realloc(*vec0, sizeof(int)*(*n));
  *vec1 = (int*)realloc(*vec1, sizeof(int)*(*n));
  (*vec0)[(*n)-1] = a;
  (*vec1)[(*n)-1] = b;
}

void setIntVal(int* vec, int val, int n)
{
  int i;
  for(i=0; i<n; ++i)
    vec[i]=val;
}

int findInVec(int target, int* vec, int n)
{
  int i=0;
  while(i<n && target!=vec[i]) i++;
  return i;
}

void addMat2Mat(int K0, int K1, const double* src, double* tar, int transpose)
{
  if(transpose==0){
    for(int i=0; i<K0*K1; ++i)
      tar[i] += src[i];
  }else{
    int e=0;
    for(int j=0; j<K1; ++j){
      for(int i=0; i<K0; ++i){
	tar[j+i*K1] += src[e];
	++e;
      }
    }
  }
}

void addVec2Vec(const double* src, double* tar, int n, double factor)
{
  int i;
  for(i=0; i<n; ++i)
    tar[i] += src[i] * factor;
}

void readFunction(Function *f, FILE* fp)
{
  int i;
  fscanf(fp, "%d %d", &(f->type), &(f->size_val));
  f->val = (double*)malloc(sizeof(double)*(f->size_val));
  // assert(f->size_val >= 2);
  for(i=0; i<f->size_val; ++i)
    fscanf(fp, "%lf", &(f->val[i]));
}

void readSubProblem(SubProblem *sp, FILE* fp)
{
  sp->num_msg = 0;
  sp->msg = NULL;
  sp->min_energy_valid = 0;
  int i;
  fscanf(fp, "%d %d", &(sp->func_id), &(sp->num_variable));
  sp->var_id = (int*)malloc(sizeof(int)*(sp->num_variable+1));
  sp->card = (int*)malloc(sizeof(int)*sp->num_variable);
  sp->sol = (int*)malloc(sizeof(int)*(sp->num_variable+1));
  for(i=0; i<sp->num_variable; ++i)
    fscanf(fp, "%d", &(sp->var_id[i]));
  sp->var_id[sp->num_variable] = sp->var_id[0];
}

double evaluateSubProblemEnergy(SubProblem* p, int* states, int initialized)
{
  double energy = 0;
  int i;
  int N = p->num_variable;
  if(p->msg){
    for(i=0; i<N; ++i)
      energy += p->msg[i][states[p->var_id[i]]];

    if(N == 2 && p->msg[2])
      energy += p->msg[2][states[p->var_id[0]] + 
			  states[p->var_id[1]] * p->card[0]];
    if(N > 2){
      if(p->type == FUNC_TYPE_CYCLE){
	for(i=0; i<N; ++i){
	  energy += p->msg[i+N][states[p->var_id[i]] + 
				states[p->var_id[i+1]] * p->card[i]];
	}
      }
    }
  }

  /* original func */
  if(p->type == FUNC_TYPE_POTTS){

    energy += p->val[states[p->var_id[0]] == states[p->var_id[1]]?0:1]; 

  }else if(p->type == FUNC_TYPE_ISING){

    if(states[p->var_id[0]] != states[p->var_id[1]])
      energy += p->val[0];

  }else if(p->type == FUNC_TYPE_TABLE){

    int N = p->num_variable;
    if(p->num_variable == 1 && initialized == 0){
      energy += p->val[states[p->var_id[0]]]; 
    }else if(p->num_variable == 2){
      energy += p->val[states[p->var_id[0]] + states[p->var_id[1]] * p->card[0]];
    }else if(p->num_variable > 2){ // high order table
      int assignment = 0;
      for(int i=p->num_variable-1; i>=0; --i){
	assignment += states[p->var_id[i]];
	if(i>0) assignment *= p->card[(i+N-1)%N];
      }
      energy += p->val[assignment];

      if(p->num_msg > N + 1 /*parameters->pair_msg_for_high_order_clique*/){
	assert(p->num_msg == N+ N*N);
	for(int i=0; i<N; ++i){
	  for(int j=i+1; j<N; ++j){
	    int msg_id = N + i*N + j;
	    energy += p->msg[msg_id][ states[p->var_id[i]] + states[p->var_id[j]] * (p->card[i]) ];
	  }
	}
      }
    }

  }else if(p->type == FUNC_TYPE_POTTS0){

    if(states[p->var_id[0]] != states[p->var_id[1]])
      energy += p->val[1];

  }else if(p->type == FUNC_TYPE_TL1){

    int diff = abs(states[p->var_id[0]] - states[p->var_id[1]]);
    if(initialized)
      energy += p->val[diff];
    else
      energy += MIN(p->val[0] * diff, p->val[1]);

  }else if(p->type == FUNC_TYPE_TL2){

    int K = p->card[0];
    int diff = abs(states[p->var_id[0]] - states[p->var_id[1]]);
    if(initialized)
      energy += p->val[K+K+diff]; 
    else
      energy += MIN(p->val[0] * (diff*diff), p->val[1]);

  }else if(p->type != FUNC_TYPE_CYCLE){ // no need to add anything for cycle
    printf("ERR: evaluateSubProblemEnergy to be implemented (new type %d)\n", p->type);
    assert(p->type == FUNC_TYPE_CYCLE);
    assert(0);
  }
  return energy;
}

double evaluateEnergy(MapProblem* problem, int* states)
{
  int i;
  double e=problem->gb_constant;
  /*
  for(i=0; i<problem->num_singleton; ++i){     
    int var_id = problem->singleton[i]->var_id[0];
    e += problem->singleton[i]->val[states[var_id]];
  }
  */
  for(i=0; i<problem->num_subproblem; ++i)
    // if(problem->sp[i])
    e += evaluateSubProblemEnergy(problem->sp[i], states, problem->initialized);
  return e;
}

void readProblemUAI(MapProblem** p, const char* filename)
{
  FILE *fp = fopen(filename, "r");
  if(!fp)
    {printf("ERR: file %s not found\n", filename); exit(1);}
  *p = (MapProblem*)malloc(sizeof(MapProblem));
  int i,j;
  char buf[20];
  fscanf(fp, "%s\n", buf);
  //  assert(strcmp(buf, "MARKOV") == 0);
  fscanf(fp, "%d", &((*p)->num_variable));
/*
  printf("%d variables\n", (*p)->num_variable);
*/
  (*p)->card = (int*)malloc(sizeof(int)*((*p)->num_variable));
  (*p)->sol = (int*)malloc(sizeof(int)*((*p)->num_variable));
  (*p)->num_phi = 0;
  (*p)->num_sp_of_phi = NULL;
  (*p)->scope_of_phi = NULL;
  (*p)->sp_of_phi = NULL;
  (*p)->initialized = 0;
  for(i=0; i<(*p)->num_variable; i++){
    fscanf(fp, "%d", &((*p)->card[i]));
  }
  fscanf(fp, "%d", &((*p)->num_subproblem));  
/*
  printf("%d factors\n", (*p)->num_subproblem);
*/
  (*p)->num_function = (*p)->num_subproblem;
  (*p)->func = (Function*)malloc(sizeof(Function)*((*p)->num_function));
  (*p)->sp = (SubProblem**)malloc(sizeof(SubProblem*)*((*p)->num_subproblem));

  // read factor info
  for(i=0; i<(*p)->num_subproblem; ++i){
    (*p)->sp[i] = (SubProblem*)malloc(sizeof(SubProblem));
    SubProblem * sp = (*p)->sp[i];
    sp->num_msg = 0;
    sp->msg = NULL;
    sp->func_id = i;
    int num_var;
    fscanf(fp, "%d", &(num_var));
    //    printf("factor %d with %d vars\n", i, num_var);
    sp->num_variable = num_var;
    sp->min_energy_valid = 0;
    sp->var_id = (int*)malloc(sizeof(int)*(sp->num_variable+1));
    sp->card = (int*)malloc(sizeof(int)*sp->num_variable);
    sp->sol = (int*)malloc(sizeof(int)*(sp->num_variable+1));
    for(j=0; j<sp->num_variable; ++j)
      fscanf(fp, "%d", &(sp->var_id[j]));
    sp->var_id[sp->num_variable] = sp->var_id[0];
    sp->is_binary = 1;
    for(j=0; j<(*p)->sp[i]->num_variable; ++j){
      (*p)->sp[i]->card[j] = (*p)->card[(*p)->sp[i]->var_id[j]];
      if(sp->card[j] != 2)
	sp->is_binary = 0;
    }
  }

  for(i=0; i<(*p)->num_subproblem; ++i){
    Function *f = &((*p)->func[i]);
    f->type = FUNC_TYPE_TABLE;
    fscanf(fp, "%d", &(f->size_val));
    f->val = (double*)malloc(sizeof(double)*(f->size_val));

    //    printf("factor %d with %d parameters\n", i, f->size_val);
    if( (*p)->sp[i]->num_variable == 1){
      for(j=0; j<f->size_val; ++j){
	fscanf(fp, "%lf", &(f->val[j]));
	if(f->val[j] > 1e-100)
	  f->val[j] = -log(f->val[j]);
	else{
	  f->val[j] = INFVAL;
	}
      }
    }else if((*p)->sp[i]->num_variable == 2){
      //      assert(f->size_val == ((*p)->sp[i]->card[0]) * 
      //	     ((*p)->sp[i]->card[1]));
      int K = (*p)->sp[i]->card[0];
      for(int k=0; k<(*p)->sp[i]->card[0]; ++k){
	for(int l=0; l<(*p)->sp[i]->card[1]; ++l){
	  fscanf(fp, "%lf", &(f->val[k+l*K]));
	  if(f->val[k+l*K] > 1e-100)
	    f->val[k+l*K] = -log(f->val[k+l*K]);
	  else{
	    f->val[k+l*K] = INFVAL;
	  }
	}
      }
    }else{
      
      int N = (*p)->sp[i]->num_variable;
      int* stride = (*p)->sp[i]->sol;
      stride[0] = 1;
      for(j=1; j<N; ++j)
	stride[j] = stride[j-1] * ((*p)->sp[i]->card[j-1]);
      //      assert(stride[N-1] * ((*p)->sp[i]->card[N-1]) == f->size_val);

      int index = 0;
      for(j=0; j<f->size_val; ++j){
	// assert(index >= 0 && index < f->size_val);
	fscanf(fp, "%lf", &(f->val[index]));
	if(f->val[index] > 1e-100)
	  f->val[index] = -log(f->val[index]);
	else{
	  f->val[index] = INFVAL;
	}
	// find next index
	for(int k=N-1; k>=0; --k){
	  if( /*assignment[k]*/ ((index / stride[k]) % ((*p)->sp[i]->card[k])) == (*p)->sp[i]->card[k]-1 ){
	    index -= ((*p)->sp[i]->card[k]-1) * stride[k];
	  }else{
	    index += stride[k];
	    break;
	  }
	}
      }
    }
  }

  for(i=0; i<(*p)->num_subproblem; ++i){
    //    (*p)->sp[i]->size_val = (*p)->func[(*p)->sp[i]->func_id].size_val;
    (*p)->sp[i]->val = (*p)->func[(*p)->sp[i]->func_id].val;
    (*p)->sp[i]->type = (*p)->func[(*p)->sp[i]->func_id].type;
  }

  fclose(fp);
  (*p)->gb_constant = 0.0;
  //  (*p)->num_singleton = 0;
  //  (*p)->singleton = NULL;
  setIntVal((*p)->sol,0,(*p)->num_variable);
  (*p)->init_primal_obj = evaluateEnergy((*p), (*p)->sol);
  (*p)->primal_obj = (*p)->init_primal_obj;
}


void readProblemTXT(MapProblem** p, const char* filename)
{
  FILE *fp = fopen(filename, "r");
  if(!fp)
    {printf("ERR: file %s not found\n", filename); exit(1);}

  *p = (MapProblem*)malloc(sizeof(MapProblem));
  int i,j;
  fscanf(fp, "%d %d %d %lf", 
	 &((*p)->num_variable), &((*p)->num_function), &((*p)->num_subproblem), &((*p)->gb_constant));
/*
  printf("numVar=%d, ", (*p)->num_variable);
  printf("numEdge=%d \n", ((*p)->num_subproblem) - ((*p)->num_variable));
*/

  (*p)->card = (int*)malloc(sizeof(int)*((*p)->num_variable));
  (*p)->sol = (int*)malloc(sizeof(int)*((*p)->num_variable));
  (*p)->num_phi = 0;
  (*p)->num_sp_of_phi = NULL;
  (*p)->scope_of_phi = NULL;
  (*p)->sp_of_phi = NULL;
  (*p)->initialized = 0;
  for(i=0; i<(*p)->num_variable; i++){
    fscanf(fp, "%d", &((*p)->card[i]));
  }

  (*p)->func = (Function*)malloc(sizeof(Function)*((*p)->num_function));
  (*p)->sp = (SubProblem**)malloc(sizeof(SubProblem*)*((*p)->num_subproblem));
  for(i=0; i<(*p)->num_function; ++i)
    readFunction(&((*p)->func[i]), fp);



  for(i=0; i<(*p)->num_subproblem; ++i){
    (*p)->sp[i] = (SubProblem*)malloc(sizeof(SubProblem));
    readSubProblem((*p)->sp[i], fp);
    (*p)->sp[i]->is_binary = 1;
    for(j=0; j<(*p)->sp[i]->num_variable; ++j){
      (*p)->sp[i]->card[j] = (*p)->card[(*p)->sp[i]->var_id[j]];
      if((*p)->sp[i]->card[j] != 2)
	(*p)->sp[i]->is_binary = 0;
    }
    (*p)->sp[i]->val = (*p)->func[(*p)->sp[i]->func_id].val;
    //    (*p)->sp[i]->size_val = (*p)->func[(*p)->sp[i]->func_id].size_val;
    (*p)->sp[i]->type = (*p)->func[(*p)->sp[i]->func_id].type;
  }
  fclose(fp);
  setIntVal((*p)->sol,0,(*p)->num_variable);
  (*p)->init_primal_obj = evaluateEnergy((*p), (*p)->sol);
  (*p)->primal_obj = (*p)->init_primal_obj;
}

void createRandomPottsGrid(MapProblem** p, int n, int K, double force)
{
  *p = (MapProblem*)malloc(sizeof(MapProblem));
  int N = n*n;
  (*p)->num_variable = N;
  (*p)->num_function = N + 1;
  (*p)->num_subproblem = N + 2*n*(n-1);
  (*p)->card = (int*)malloc(sizeof(int)*((*p)->num_variable));
  (*p)->sol = (int*)malloc(sizeof(int)*((*p)->num_variable));
  (*p)->num_phi = 0;
  (*p)->num_sp_of_phi = NULL;
  (*p)->scope_of_phi = NULL;
  (*p)->sp_of_phi = NULL;
  (*p)->initialized = 0;
  for(int i=0; i<(*p)->num_variable; i++)
    (*p)->card[i] = K;
  (*p)->func = (Function*)malloc(sizeof(Function)*((*p)->num_function));
  (*p)->sp = (SubProblem**)malloc(sizeof(SubProblem*)*((*p)->num_subproblem));
  for(int i=0; i<(*p)->num_function; ++i){
    Function* f = &((*p)->func[i]);
    if(i<N){
      f->type = FUNC_TYPE_TABLE;
      f->size_val = K;
      f->val = (double*)malloc(sizeof(double)*(f->size_val));
      for(int j=0; j<K; ++j)
	f->val[j] = (double)rand()/RAND_MAX;
    }else{
      f->type = FUNC_TYPE_POTTS;
      f->size_val = 2;
      f->val = (double*)malloc(sizeof(double)*(f->size_val));
      f->val[0] = 0;
      f->val[1] = force;
    }
  }
  for(int i=0; i<(*p)->num_subproblem; ++i){
    (*p)->sp[i] = (SubProblem*)malloc(sizeof(SubProblem));
    (*p)->sp[i]->num_msg = 0;
    (*p)->sp[i]->msg = NULL;
    (*p)->sp[i]->min_energy_valid = 0;
  }

  for(int i=0; i<N; ++i){
    (*p)->sp[i]->num_variable = 1;
    (*p)->sp[i]->var_id = (int*)malloc(sizeof(int));
    (*p)->sp[i]->card = (int*)malloc(sizeof(int));
    (*p)->sp[i]->sol = (int*)malloc(sizeof(int));
    (*p)->sp[i]->var_id[0] = i;
    (*p)->sp[i]->func_id = i;
  }
  int sp_id=N;
  for(int i=0; i<n-1; ++i){
    for(int j=0; j<n; ++j){
      (*p)->sp[sp_id]->num_variable = 2;
      (*p)->sp[sp_id]->var_id = (int*)malloc(sizeof(int)*3);
      (*p)->sp[sp_id]->card = (int*)malloc(sizeof(int)*2);
      (*p)->sp[sp_id]->sol = (int*)malloc(sizeof(int)*3);
      (*p)->sp[sp_id]->var_id[0] = i + j*n;
      (*p)->sp[sp_id]->var_id[1] = (i+1) + j*n;
      (*p)->sp[sp_id]->var_id[2] = i + j*n;
      (*p)->sp[sp_id]->func_id = N;
      sp_id++;
    }
  }

  for(int i=0; i<n; ++i){
    for(int j=0; j<n-1; ++j){
      (*p)->sp[sp_id]->num_variable = 2;
      (*p)->sp[sp_id]->var_id = (int*)malloc(sizeof(int)*3);
      (*p)->sp[sp_id]->card = (int*)malloc(sizeof(int)*3);
      (*p)->sp[sp_id]->sol = (int*)malloc(sizeof(int)*3);
      (*p)->sp[sp_id]->var_id[0] = i + j*n;
      (*p)->sp[sp_id]->var_id[1] = i + (j+1)*n;
      (*p)->sp[sp_id]->var_id[2] = i + j*n;
      (*p)->sp[sp_id]->func_id = N;
      sp_id++;
    }
  }
  assert(sp_id == (*p)->num_subproblem);
  for(int i=0; i<(*p)->num_subproblem; ++i){
    for(int j=0; j<(*p)->sp[i]->num_variable; ++j){
      (*p)->sp[i]->card[j] = (*p)->card[(*p)->sp[i]->var_id[j]];
      (*p)->sp[i]->val = (*p)->func[(*p)->sp[i]->func_id].val;
      //      (*p)->sp[i]->size_val = (*p)->func[(*p)->sp[i]->func_id].size_val;
      (*p)->sp[i]->type = (*p)->func[(*p)->sp[i]->func_id].type;
    }
  }
}


void writeFunction(Function *f, FILE* fp)
{
  int i;
  fprintf(fp, "%d %d\n", f->type, f->size_val);
  for(i=0; i<f->size_val; ++i)
    fprintf(fp, "%f ", f->val[i]);
  fprintf(fp, "\n");
}

void writeSubProblem(SubProblem *sp, FILE* fp)
{
  if(sp->num_msg != 0){
    fclose(fp);
    printf("ERR: writing problem after inference is not supported yet\n");
    assert(0);
  }
  int i;
  fprintf(fp, "%d %d ", sp->func_id, sp->num_variable);
  for(i=0; i<sp->num_variable; ++i)
    fprintf(fp, "%d ", sp->var_id[i]);
  fprintf(fp, "\n");
}

void writeProblem(MapProblem* p, const char* filename)
{
  FILE *fp = fopen(filename, "w");
  int i;
  fprintf(fp, "%d %d %d %lf\n", p->num_variable, p->num_function, p->num_subproblem, p->gb_constant);
  for(i=0; i<p->num_variable; i++)
    fprintf(fp, "%d ", p->card[i]);
  fprintf(fp, "\n");
  for(i=0; i<p->num_function; ++i)
    writeFunction(&(p->func[i]), fp);
  for(i=0; i<p->num_subproblem; ++i)
    writeSubProblem(p->sp[i], fp);
  fclose(fp);
}

void getRandomState(MapProblem* problem, int* states)
{
  int i;
  for(i=0; i<problem->num_variable; ++i)
    states[i] = rand()%(problem->card[i]);
}

void  setupCycleSolver(Cycle* cy, SubProblem* p)
{
  int i;
  cy->N = p->num_variable;
  cy->L = cy->N-2;
  cy->card = p->card;
  cy->ts = p->msg;
  //  cy->unary_msg = p->unary_msg;
  for(i=0; i<cy->N; ++i)
    cy->tp[i] = p->msg[cy->N+i];
  for(i=0; i<cy->L-1; ++i)
    cy->tp[cy->N+i] = cy->cache[i];

  for (i = 0; i < cy->L; i++){
    cy->TA_static[i] = -1;
    cy->TB_static[i] = -1;
  }

}

int getNextConstrainedAssignment(int N, int* card, int* known, int* assignment)
{
  for(int i=0; i<N; ++i){
    if(known[i] >= 0){
      assignment[i] = known[i]; 
      if(i==N-1) return 0;
    }else{
      if(assignment[i] == card[i]-1){
	assignment[i] = 0;
	if(i==N-1) return 0;
      }else{
	assignment[i]++;
	break;
      }
    }
  }
  return 1;
}


double evaluateHighOrderTable(SubProblem* p, int* assignment)
{
  int N = p->num_variable;
  double energy = 0.0;
  for(int i=0; i<p->num_variable; ++i)
    energy += p->msg[i][assignment[i]];
  int e = 0;
  for(int i=p->num_variable-1; i>=0; --i){
    e += assignment[i];
    if(i>0) e *= p->card[i-1];
  }
  energy += p->val[e];  

  if(p->num_msg > N+1){
    assert(p->num_msg == N+ N*N);
    for(int i=0; i<N; ++i){
      for(int j=i+1; j<N; ++j){
	int msg_id = N + i*N + j;
	energy += p->msg[msg_id][assignment[i] + assignment[j] * (p->card[i])];
      }
    }
  }
  return energy;
}


void getNextMsgVal(SubProblem *p, int* stride, int i, int N, double* msgval){
  for(int j=0; j<N; ++j){ // find the variable to modify next
    int current = i / stride[j] % (p->card[j]);
    int next = (current+1)%(p->card[j]);
    *msgval += (p->msg[j][next] - p->msg[j][current]);
    for(int j1=0; j1<j; ++j1){
      int msg_id = N+ j1*N+j;
      *msgval += (p->msg[msg_id][next*(p->card[j1])] 
		  -p->msg[msg_id][(current)*(p->card[j1])]);
    }
    for(int j1=j+1; j1<N; ++j1){
      int msg_id = N + j*N+j1;
      int a1 = i / stride[j1] % (p->card[j1]);
      *msgval += (p->msg[msg_id][next + a1 *(p->card[j])] 
		  -p->msg[msg_id][current + a1 *(p->card[j])]); 
    }
    if(next)
      break;
  }
}


// binary version
void getNextMsgVal_binary(SubProblem *p, int* stride, int i, int N, double* msgval){
  for(int j=0; j<N; ++j){ // find the variable to modify next
    int current = (i >> j) & 1; //(i & (1<<j)) >> j;
    int next = 1-current;
    *msgval += (p->msg[j][next] - p->msg[j][current]);
    if(p->num_msg > N+1){
      for(int j1=0; j1<j; ++j1){
	int msg_id = N+ j1*N+j;
	*msgval += (p->msg[msg_id][next<<1] 
		    -p->msg[msg_id][current<<1]);
      }
      for(int j1=j+1; j1<N; ++j1){
	int msg_id = N + j*N+j1;
	int a1 = ((i >> j1) & 1) << 1; //(i & (1<<j1)) >> (j1-1);
	*msgval += (p->msg[msg_id][next + a1]
		    -p->msg[msg_id][current + a1]);
      }
    }
    if(next)
      break;
  }
}
  
void DistanceTransformL2(int K, double* source, double* dest, int* argmin, 
			 int* parabolas, int* intersections, double* val)

{
  int i, j, k, p;
  int r = 0; // = number of parabolas minus 1
  // parabolas[p] will be base of parabola p (0<=p<=r)
  // intersections[p] will be intersection between parabolas p-1 and p (1<=p<=r)
  // intersections[0] will be always 0
  parabolas[0] = 0;
  intersections[0] = 0;

  for (i=1; i<K; i++)
    {
      while ( 1 )
	{
	  j = parabolas[r]; // base of previous rightmost visible parabola
	  // k is intersection of parabolas i and j
	  //	  	  k = (int) ( 1 + ( (i + j)  +  (source[j] - source[i]) / 
	  //	  			    (m_alpha * (j - i)) ) / 2);
	  double v = j>i?val[j-i]:(-val[i-j]);
	  k = (int) ( 1 + ( (i + j)  +  (source[j] - source[i]) * v )/2);

	  if (k >= K)
	    {
	      // i is not visible
	      break;
	    }
	  if (k < 0)
	    {
	      k = 0;
	    }

	  if (k > intersections[r])
	    {
	      // intersection is rightmost, add it to end
	      r ++;
	      parabolas[r] = i;
	      intersections[r] = k;
	      break;
	    }
	  // j is not visible
	  if (r == 0)
	    {
	      parabolas[0] = i;
	      break;
	    }
	  r --;
	}
    }

  intersections[r + 1] = K;

  i = 0;
  for (p=0; p<=r; p++)
    {
      j = parabolas[p];
      // i values in [intersections[p], intersections[p+1]) are assigned to j
      for ( ; i<intersections[p+1]; i++)
	{
	  dest[i] = source[j] + val[K+abs(i-j)]; //m_alpha*(i-j)*(i-j);
	  if(argmin) argmin[i] = j;
	}
    }
}

double getSubProblemMin(SubProblem* p, Cycle* cy, MapParameters *parameters){

  if(p->min_energy_valid) return p->min_energy;
  double min_e;

  if(p->num_variable==1){ // unary 
    int argmin = 0;
    min_e = p->msg[0][0];
    for(int k=1; k<p->card[0]; ++k){
      if(p->msg[0][k] < min_e){
	min_e = p->msg[0][k];
	argmin = k;
      }
    }
    p->min_energy_valid=1;
    p->min_energy = min_e;
    p->sol[0] = argmin;
    return min_e;
  }

  if(p->type == FUNC_TYPE_POTTS){
    if(p->msg[2]){ // has binary message, quadratic time
      int e=0, k,l,K = p->card[0],L = p->card[1];
      int argmin = 0;
      min_e = p->msg[0][0]+p->msg[1][0]+p->msg[2][0]+p->val[0];
      for(l=0; l<L; ++l){
	for(k=0; k<K; ++k){
	  double val = p->msg[2][e] +p->msg[0][k] +
	    p->msg[1][l] + (k==l?p->val[0]:p->val[1]);
	  if(val < min_e){
	    min_e = val;
	    argmin = e; // k+l*K
	  }
	  ++e;
	}
      }
      p->sol[0] = argmin%K;
      p->sol[1] = argmin/K;
      p->sol[2] = p->sol[0];

    }else{ // no binary message, linear time

      int k,l,K = p->card[0],L = p->card[1];
      int K0 = MIN(K,L);
      int kargmin = 0, largmin=0;
      min_e = p->msg[0][0]+p->msg[1][0];
      for(k=1; k<K0; ++k){
	if(p->msg[0][k]+p->msg[1][k] < min_e){
	  min_e = p->msg[0][k]+p->msg[1][k];
	  kargmin = k;
	}
      }
      min_e += p->val[0]; // min energy given 'agree'
      largmin = kargmin;

      // minimize independently
      double mk, ml;
      int mk_argmin = 0, ml_argmin = 0;
      mk = p->msg[0][0]; 
      for(k=1; k<K; ++k)
	if(p->msg[0][k] < mk){
	  mk = p->msg[0][k];
	  mk_argmin = k;
	}
      ml = p->msg[1][0]; 
      for(l=1;l<L; ++l)
	if(p->msg[1][l] < ml){
	  ml = p->msg[1][l];
	  ml_argmin = l;
	}
      double val = ml + mk + p->val[mk_argmin==ml_argmin?0:1];
      if (val < min_e){
	min_e = val;
	largmin = ml_argmin;
	kargmin = mk_argmin;
      }

      if(p->val[1] < p->val[0] && ml_argmin == mk_argmin){
	// minimize subject to 'disagree'
	double mk1, ml1;
	int k1argmin=0, l1argmin=0;
	mk1 = p->msg[0][(ml_argmin+1)%K]; 
	for(k=0; k<K; ++k)
	  if(k != ml_argmin && p->msg[0][k] < mk1){
	    mk1 = p->msg[0][k];
	    k1argmin = k;
	  }
	ml1 = p->msg[1][(mk_argmin+1)%L]; 
	for(l=0;l<L; ++l)
	  if(l != mk_argmin && p->msg[1][l] < ml1){
	    ml1 = p->msg[1][l];
	    l1argmin = l;
	  }
	val = ml1 + mk + p->val[1];
	if(val < min_e){
	  min_e = val; kargmin = mk_argmin; largmin = l1argmin;
	}
	val = ml + mk1 + p->val[1];
	if(val < min_e){
	  min_e = val; kargmin = k1argmin; largmin = ml_argmin;
	}
      }
      p->sol[0] = kargmin;
      p->sol[1] = largmin;
      p->sol[2] = p->sol[0];
    }

  }else if(p->type == FUNC_TYPE_TABLE){

    if(p->num_variable == 2){
      int argmin;
      int e=0,k,l,K = p->card[0],L = p->card[1];
      if(p->msg[2]){ 
	min_e = p->msg[0][0]+p->msg[1][0]+p->msg[2][0]+p->val[0];
	argmin=0;
	for(l=0; l<L; ++l)
	  for(k=0; k<K; ++k){
	    double val = p->val[e]+p->msg[2][e] +p->msg[0][k] +p->msg[1][l];
	    if(val < min_e){
	      min_e = val;
	      argmin = e;
	    }
	    ++e;
	  }
      }else{
	min_e = p->msg[0][0]+p->msg[1][0]+p->val[0];
	argmin = 0;
	for(l=0; l<L; ++l)
	  for(k=0; k<K; ++k){
	    double val =  p->val[e]+p->msg[0][k] +p->msg[1][l];
	    if(val < min_e){
	      min_e = val;
	      argmin = e;
	    }
	    ++e;
	  }
      }
      p->sol[0] = argmin%K;
      p->sol[1] = argmin/K;
      p->sol[2] = p->sol[0];

    }else{

      int N = p->num_variable;
      double msgval = p->msg[0][0];
      int* stride = (int*)malloc(sizeof(int) * N);
      //int* stride = p->sol;
      stride[0]=1;
      for(int i=1; i<N; ++i){
	stride[i] = stride[i-1] * (p->card[i-1]);
	msgval += p->msg[i][0];
      }

      if(p->num_msg > N + 1){
	for(int i=0; i<N; ++i){
	  for(int j=i+1; j<N; ++j){
	    int msg_id = N + i*N + j;
	    msgval += p->msg[msg_id][0];
	  }
	}
      }
      //      double msgval0 = msgval;

      int size_val = stride[N-1] * (p->card[N-1]);
      int argmin = 0;
      min_e = p->val[0] + msgval;
      if(p->is_binary){
	for(int i=0; i<size_val; ++i){
	  double value = p->val[i] + msgval;
	  if(value < min_e){
	    min_e = value;
	    argmin = i;
	  }
	  getNextMsgVal_binary(p, stride, i, N, &msgval);
	}
      }else{
	for(int i=0; i<size_val; ++i){
	  double value = p->val[i] + msgval;
	  if(value < min_e){
	    min_e = value;
	    argmin = i;
	  }
	  getNextMsgVal(p, stride, i, N, &msgval);
	}
      }
      //      assert(fabs(msgval0 - msgval) < 1e-4);
      for(int i=0; i<N; ++i){
	p->sol[i] = (argmin / stride[i]) % (p->card[i]);
      }
      free(stride);
    }
  }else if(p->type == FUNC_TYPE_CYCLE){

    setupCycleSolver(cy, p);

    if(parameters->use_fast_solver)
      cy->runFastSolver(0);
    else
      cy->runFullMinSum(0, 1);

    min_e = cy->obj;
    copyIntVec(cy->assignment, p->sol, p->num_variable+1);

  }else if(p->type == FUNC_TYPE_POTTS0){ 

    if(p->msg[2]){ // has binary message, quadratic time
      int e=0, k,l,K = p->card[0],L = p->card[1];
      int argmin = 0;
      min_e = p->msg[0][0]+p->msg[1][0]+p->msg[2][0]+p->val[0];
      for(l=0; l<L; ++l){
	for(k=0; k<K; ++k){
	  double val = p->msg[2][e] +p->msg[0][k] +
	    p->msg[1][l] + (k==l?p->val[0]:p->val[1]);
	  if(val < min_e){
	    min_e = val;
	    argmin = e; // k+l*K
	  }
	  ++e;
	}
      }
      p->sol[0] = argmin%K;
      p->sol[1] = argmin/K;
      p->sol[2] = p->sol[0];

    }else{ // no binary message, linear time 

      int k,l,K = p->card[0],L = p->card[1];
      int K0 = MIN(K,L);
      int kargmin = 0, largmin = 0, kargmin0=0;
      double kmin0 = p->msg[0][0];
      for(k=1; k<K; ++k){
	if(p->msg[0][k] < kmin0){
	  kmin0 = p->msg[0][k];
	  kargmin0 = k;
	}
      }
      min_e = p->msg[0][0] + p->msg[1][0];
      for(l=0; l<K0; ++l){
	double agree = p->msg[0][l] + p->msg[1][l];
	double disagree = kmin0 + p->msg[1][l] + p->val[1];
	if(agree < disagree){
	  if(agree < min_e){
	    min_e = agree;
	    largmin = l;
	    kargmin = l;
	  }
	}else{
	  if(disagree < min_e){
	    min_e = disagree;
	    largmin = l;
	    kargmin = kargmin0;
	  }
	}
      }
      for(l=K0; l<L; ++l){
	double disagree = kmin0 + p->msg[1][l] + p->val[1];
	if(disagree < min_e){
	  min_e = disagree;
	  largmin = l;
	  kargmin = kargmin0;
	}
      }
      p->sol[0] = kargmin;
      p->sol[1] = largmin;
      p->sol[2] = p->sol[0];
    }

  }else if(p->type == FUNC_TYPE_ISING){ 

    if(p->msg[2]){
      int e=0, k,l;
      int argmin = 0;
      min_e = p->msg[0][0]+p->msg[1][0]+p->msg[2][0];
      for(l=0; l<2; ++l){
	for(k=0; k<2; ++k){
	  double val = p->msg[2][e] +p->msg[0][k] +
	    p->msg[1][l] + (k==l?0:p->val[0]);
	  if(val < min_e){
	    min_e = val;
	    argmin = e;
	  }
	  ++e;
	}
      }
      p->sol[0] = argmin%2;
      p->sol[1] = argmin/2;
      p->sol[2] = p->sol[0];

    }else{
      double buf[2];
      int intbuf[2];
      if (p->msg[0][0] < p->msg[0][1] + p->val[0]){
	buf[0] = p->msg[0][0]; intbuf[0] = 0;
      }else{
	buf[0] = p->msg[0][1] + p->val[0]; intbuf[0] = 1;
      }
      if (p->msg[0][1] < p->msg[0][0] + p->val[0]){
	buf[1] = p->msg[0][1]; intbuf[1] = 1;
      }else{
	buf[1] = p->msg[0][0] + p->val[0]; intbuf[1] = 0;
      }
      if(buf[0] + p->msg[1][0] < buf[1] + p->msg[1][1]){
	min_e = buf[0] + p->msg[1][0]; p->sol[1] = 0, p->sol[0] = intbuf[0];
      }else{
	min_e = buf[1] + p->msg[1][1]; p->sol[1] = 1, p->sol[0] = intbuf[1];
      }
      p->sol[2] = p->sol[0];
    }

  }else if(p->type == FUNC_TYPE_TL1){ 

    if(p->msg[2]){ // has binary message, quadratic time
      int e=0, k,l,K = p->card[0],L = p->card[1];
      int argmin = 0;
      min_e = p->msg[0][0]+p->msg[1][0]+p->msg[2][0];
      for(l=0; l<L; ++l){
	for(k=0; k<K; ++k){
	  double val = p->msg[2][e] +p->msg[0][k] +
	    p->msg[1][l] + p->val[abs(k-l)];
	  if(val < min_e){
	    min_e = val;
	    argmin = e; // k+l*K
	  }
	  ++e;
	}
      }
      p->sol[0] = argmin%K;
      p->sol[1] = argmin/K;
      p->sol[2] = p->sol[0];

    }else{ // no binary message, linear time

      int l,K = p->card[0],L = p->card[1];
      assert(K==L);
      double alpha = p->val[1];
      double gamma = p->val[K-1];
    
      double* buf = (double*)(parameters->buffer);
      int* intbuf = (int*)(buf + K);

      buf[0] = p->msg[0][0];
      intbuf[0] = 0;
      double vMin = p->msg[0][0];
      int kargmin = 0, largmin = 0, kargmin0 = 0;
      for(l=1; l<L; ++l){
	buf[l] = p->msg[0][l];
	intbuf[l] = l;
	if(buf[l] > p->msg[0][l-1] + alpha){
	  buf[l] = p->msg[0][l-1] + alpha;
	  intbuf[l] = intbuf[l-1];
	}
	if(vMin > p->msg[0][l] /*buf[l]*/){
	  vMin = p->msg[0][l]; //buf[l];
	  kargmin0 = l;
	}
      }
      l--;
      if (buf[l] > vMin + gamma){
	buf[l] = vMin + gamma;
	intbuf[l] = kargmin0;
      }
      for(l--; l>=0; l--){
	if(buf[l] > buf[l+1] + alpha){
	  buf[l] = buf[l+1] + alpha;
	  intbuf[l] = intbuf[l+1];
	}
	if (buf[l] > vMin + gamma){
	  buf[l] = vMin + gamma;
	  intbuf[l] = kargmin0;
	}
      }
      min_e = buf[0] + p->msg[1][0];
      kargmin = intbuf[0];
      largmin = 0;
      for(l=1; l<L; ++l){
	if( buf[l] + p->msg[1][l] < min_e){
	  min_e = buf[l] + p->msg[1][l];
	  kargmin = intbuf[l];
	  largmin = l;
	}
      }
      p->sol[0] = kargmin;
      p->sol[1] = largmin;
      p->sol[2] = p->sol[0];
    }

  }else if(p->type == FUNC_TYPE_TL2){ 

    if(p->msg[2]){ // has binary message, quadratic time
      int e=0, k,l,K = p->card[0],L = p->card[1];
      int argmin = 0;
      min_e = p->msg[0][0]+p->msg[1][0]+p->msg[2][0];
      for(l=0; l<L; ++l){
	for(k=0; k<K; ++k){
	  double val = p->msg[2][e] +p->msg[0][k] +
	    p->msg[1][l] + p->val[K+K+abs(k-l)];
	  if(val < min_e){
	    min_e = val;
	    argmin = e; // k+l*K
	  }
	  ++e;
	}
      }
      p->sol[0] = argmin%K;
      p->sol[1] = argmin/K;
      p->sol[2] = p->sol[0];

    }else{ // no binary message, linear time

      int l,K = p->card[0],L = p->card[1];
      assert(K==L);
      double* buf = (double*)(parameters->buffer);
      int* intbuf = (int*)(buf + K);
      int* parabolas = intbuf + K;
      int* intersections = parabolas + K;
      double alpha = p->val[K+1];
      double gamma = p->val[0];
      double vMin = p->msg[0][0];
      int kargmin = 0, largmin = 0, kargmin0 = 0;
      for(l=1; l<L; ++l){
	if(vMin > p->msg[0][l] /*buf[l]*/){
	  vMin = p->msg[0][l]; //buf[l];
	  kargmin0 = l;
	}
      } 
      if(alpha == 0){ 
	kargmin = kargmin0;
	largmin = 0;    
	min_e = vMin + p->msg[1][0];
	for(l=1; l<L; ++l){
	  if(vMin + p->msg[1][l] < min_e){
	    min_e = vMin + p->msg[1][l];
	    largmin = l;
	  }
	}
      }else{
	DistanceTransformL2(K, p->msg[0], buf, intbuf, 
			    parabolas, intersections, p->val);
	for(l=0; l<L; ++l){
	  if (buf[l] > vMin + gamma){
	    buf[l] = vMin + gamma;
	    intbuf[l] = kargmin0;
	  }
	}
	min_e = buf[0] + p->msg[1][0];
	kargmin = intbuf[0];
	largmin = 0;    
	for(l=1; l<L; ++l){
	  if( buf[l] + p->msg[1][l] < min_e){
	    min_e = buf[l] + p->msg[1][l];
	    kargmin = intbuf[l];
	    largmin = l;
	  }
	}
      }
      p->sol[0] = kargmin;
      p->sol[1] = largmin;
      p->sol[2] = p->sol[0];
    }

  }else{
    printf("ERR: getSubProblemMin to be implemented (new type)\n");
    assert(p->type == FUNC_TYPE_CYCLE);
    assert(0);
  }
  p->min_energy_valid = 1;
  p->min_energy = min_e;
  return min_e;
}


double evaluateDual(MapProblem* problem, MapParameters* parameters)
{
  int i;
  double d=problem->gb_constant;
  for(i=0; i<problem->num_subproblem; ++i){
    d += getSubProblemMin(problem->sp[i], problem->cy_solver, parameters);
    assert(!isnan(d));
  }
  return d;
}


void renewDualObj(MapProblem* problem, MapParameters* parameters)
{
  problem->dual_obj = evaluateDual(problem, parameters);
}

void getHalfCycleMinMarginal(SubProblem* p, int* solution, 
			     int d, int Kmax, double* mmarginal)
{
  int N = p->num_variable;
  int dl=(d-1+N)%N, dr=(d+1)%N;
  while(solution[p->var_id[dl]] < 0) dl = (dl-1+N)%N;
  while(solution[p->var_id[dr]] < 0) dr = (dr+1)%N;
  int one_assigned_var = (dl == dr)?1:0;
  //  printf("d=%d, dl=%d, dr=%d\n", d, dl, dr);
  int al = solution[p->var_id[dl]];
  int ar = solution[p->var_id[dr]];
  double* message0 = (double*)malloc(sizeof(double)*Kmax);
  double* message1 = (double*)malloc(sizeof(double)*Kmax);
  setDblVal(message0, p->msg[dl][al], p->card[dl+1]); 
  //  printf("used msg[%d]\n", dl);
  project_ij2j(p->card[dl], p->card[dl+1], p->msg[N+dl], al, message0);
  //  printf("used msg[%d]\n", N+dl);
  dl = (dl+1)%N;
  while(dl != d){
    addVec2Vec(p->msg[dl], message0, p->card[dl], 1.0);
    //    printf("used msg[%d]\n", dl);
    msg_i_ij_j(p->card[dl], p->card[dl+1], message0, p->msg[N+dl], message1);
    //    printf("used msg[%d]\n", N+dl);
    double* tmp = message0; message0 = message1; message1 = tmp;
    dl = (dl+1)%N;
  }
  addVec2Vec(message0, mmarginal, p->card[d], 1.0);
  if(one_assigned_var)
    setDblVal(message0, 0.0, p->card[(dr-1+N)%N]);
  else{
    setDblVal(message0, p->msg[dr][ar], p->card[(dr-1+N)%N]);
    //    printf("used msg[%d]\n", dr);
  }
  project_ij2i(p->card[(dr-1+N)%N], p->card[dr], p->msg[N+(dr-1+N)%N], 
	       ar, message0);  
  //  printf("used msg[%d]\n", N+(dr-1+N)%N);
  dr = (dr-1+N)%N;
  while(dr != d){
    addVec2Vec(p->msg[dr], message0, p->card[dr], 1.0);
    //    printf("used msg[%d]\n", dr);
    msg_j_ij_i(p->card[(dr-1+N)%N], p->card[dr],message0, 
	       p->msg[N +(dr-1+N)%N], message1);
    //    printf("used msg[%d]\n", N +(dr-1+N)%N);
    double* tmp = message0; message0 = message1; message1 = tmp;
    dr = (dr-1+N)%N;
  }
  addVec2Vec(message0, mmarginal, p->card[d], 1.0);
  addVec2Vec(p->msg[d], mmarginal, p->card[d], 1.0);
  //  printf("used msg[%d]\n", d);
  free(message0);
  free(message1);
}


void getConditionalMinMarginalHighOrderTable(SubProblem* p, 
					     int var_sub_id,
					     int* solution, 
					     double* mmarginal)
{
  int N = p->num_variable;
  double* mm = (double*)malloc(sizeof(double)*(p->card[var_sub_id]));

  int* assignment = (int*)malloc(sizeof(int)*(p->num_variable));
  int* known = (int*)malloc(sizeof(int)*(p->num_variable));
  setIntVal(assignment, 0, p->num_variable);
  int first = -1; // first unknown sub idx
  for(int i=0; i<p->num_variable; ++i){
    known[i] = solution[p->var_id[i]];
    if(first==-1 && known[i] == -1)
      first = i;
    if(known[i] >= 0)
      assignment[i] = known[i];
  }
  for(int k=0; k<p->card[var_sub_id]; ++k){
    assignment[var_sub_id] = k;
    mm[k] = evaluateHighOrderTable(p, assignment);
  }
  assignment[var_sub_id] = 0;
  assignment[first] = -1;
    
  while(getNextConstrainedAssignment(N, p->card, known, assignment)){
    double val = evaluateHighOrderTable(p, assignment);
    if(val < mm[assignment[var_sub_id]])
      mm[assignment[var_sub_id]] = val;
  }
  free(assignment);
  free(known);
  for(int k=0; k<p->card[var_sub_id]; ++k){
    mmarginal[k] += mm[k];
  }
  free(mm);
}

void vecset(double* v, double val, int n)
{
  for(int i=0; i<n; ++i)
    v[i] = val;
}

void vecadd(double* tar, const double* src , double factor, int n)
{
  for(int i=0; i<n; ++i)
    tar[i] += factor * src[i];
}


int getPairMinMarginalHighOrderTable(SubProblem* p, 
				     int var_sub_id0,
				     int var_sub_id1,
				     double factor0,
				     double factor1,
				     double* mmarginal)
{
  int N = p->num_variable;
  int K = p->card[var_sub_id0];
  int L = p->card[var_sub_id1];
  double* mm = (double*)malloc(sizeof(double)*(K*L));
  vecset(mm, 1e8, K*L);

  double msgval = p->msg[0][0];
  int* stride = (int*)malloc(sizeof(int) * N);
  //  int* stride = p->stride; //(int*)(mm+(p->card[var_sub_id]));
  stride[0]=1;
  for(int i=1; i<N; ++i){ 
    stride[i] = stride[i-1] * (p->card[i-1]);
    msgval += p->msg[i][0];
  }
  for(int i=0; i<N; ++i){
    for(int j=i+1; j<N; ++j){
      int msg_id = N + i*N + j;
      msgval += p->msg[msg_id][0];
    }
  }

  int size_val = stride[N-1] * (p->card[N-1]);
  //  double msgval0 = msgval;

  if(p->is_binary){
    for(int i=0; i<size_val; ++i){
      double value = p->val[i] + msgval;
      int k = (i >> var_sub_id0) & 1; //(i / stride[var_sub_id0]) % K;
      int l = ((i >> var_sub_id1) & 1) << 1; //(i / stride[var_sub_id1]) % L;
      mm[k + l] = MIN(mm[k + l], value);
      getNextMsgVal_binary(p, stride, i, N, &msgval);
    }
  }else{
    for(int i=0; i<size_val; ++i){
      double value = p->val[i] + msgval;
      int k = (i / stride[var_sub_id0]) % K;
      int l = (i / stride[var_sub_id1]) % L;
      mm[k + l*K] = MIN(mm[k + l*K], value);
      getNextMsgVal(p, stride, i, N, &msgval);
    }
  }
  free(stride);
  //  assert(fabs(msgval0 - msgval) < 1e-4);

  int transpose = (var_sub_id0 > var_sub_id1);
  int msg_id = N + MIN(var_sub_id0, var_sub_id1) * N + MAX(var_sub_id0, var_sub_id1);

  /*
  p->min_energy_valid=0;
  double min_energy0 = getSubProblemMin(p,0,0);
  double min_energy1 = 1e8;
  for(int k=0; k< K*L; ++k){
    min_energy1 = MIN(min_energy1, mm[k]);
  }
  assert(fabs(min_energy0 -min_energy1)<DEBUG_THRESHOLD);
  */
  
  double minval = mm[0];
  if(transpose){
    int e=0;
    for(int l=0; l<L; ++l){
      for(int k=0; k<K; ++k){
	p->msg[msg_id][l + k*L] -= factor0 * mm[e]; // e = k+l*K
	mmarginal[e] += factor1 * mm[e];
	minval = MIN(minval, mm[e]);
	++e;
      }
    }
  }else{
    for(int k=0; k< K*L; ++k){
      p->msg[msg_id][k] -= factor0*mm[k];
      mmarginal[k] += factor1*mm[k];
      minval = MIN(minval, mm[k]);
    }
  }

  p->min_energy = (1.0-factor0)*minval;

  /*
  p->min_energy_valid=0;
  getSubProblemMin(p,0,0);
  int* states = (int*)malloc(sizeof(int)*50);
  for(int i=0; i<N; ++i)
    states[p->var_id[i]] = p->sol[i];
  assert(fabs(getSubProblemMin(p,0,0) - evaluateSubProblemEnergy(p, states, 1))<DEBUG_THRESHOLD);
  assert(fabs(getSubProblemMin(p, NULL, NULL))<DEBUG_THRESHOLD);
  free(states);
  */
  free(mm);

  return 0;  
}

int getMinMarginalHighOrderTable(SubProblem* p, 
				 int var_sub_id,
				 double factor0,
				 double factor1,
				 double* mmarginal)
{
  int N = p->num_variable;
  double* mm = (double*)malloc(sizeof(double)*(p->card[var_sub_id]));
  //double* mm = (double*)buffer;
  vecset(mm, 1e8, p->card[var_sub_id]);

  double msgval = p->msg[0][0];
  int* stride = (int*)malloc(sizeof(int) * N);
  //  int* stride = p->stride; //(int*)(mm+(p->card[var_sub_id]));
  stride[0]=1;
  for(int i=1; i<N; ++i){ 
    stride[i] = stride[i-1] * (p->card[i-1]);
    msgval += p->msg[i][0];
  }
  if(p->num_msg > N + 1){
    for(int i=0; i<N; ++i){
      for(int j=i+1; j<N; ++j){
	int msg_id = N + i*N + j;
	msgval += p->msg[msg_id][0];
      }
    }
  }
  int size_val = stride[N-1] * (p->card[N-1]);

  //  double msgval0 = msgval;
  if(p->is_binary){
    for(int i=0; i<size_val; ++i){
      double value = p->val[i] + msgval;
      int k = (i >> var_sub_id) & 1; //(i / stride[var_sub_id]) % (p->card[var_sub_id]);
      mm[k] = MIN(mm[k], value);
      getNextMsgVal_binary(p, stride, i, N, &msgval);
    }
  }else{
    for(int i=0; i<size_val; ++i){
      double value = p->val[i] + msgval;
      int k = (i / stride[var_sub_id]) % (p->card[var_sub_id]);
      mm[k] = MIN(mm[k], value);
      getNextMsgVal(p, stride, i, N, &msgval);
    }
  }
  free(stride);
  //  assert(fabs(msgval0 - msgval) < 1e-4);

  double minval = mm[0];
  for(int k=0; k<p->card[var_sub_id]; ++k){
    p->msg[var_sub_id][k] -= factor0*mm[k];
    mmarginal[k] += factor1*mm[k];
    minval = MIN(minval, mm[k]);
  }

  free(mm);
  p->min_energy = (1.0-factor0)*minval;


  /*
  p->min_energy_valid=0;
  assert(fabs(getSubProblemMin(p, NULL, NULL))<DEBUG_THRESHOLD);
  */
  return 0;
}

int extractMinMarginal(MapParameters* parameters, SubProblem* p, Cycle* cy, int dim,
		       int* scope, double factor0, double factor1,
		       double* target);


void renewMRFSolutionIndependent(MapProblem* problem, MapParameters* parameters)
{
  int i,j;
  int* solution = (int*)malloc(sizeof(int)*(problem->num_variable));
  double* mmarginal = (double*)malloc(sizeof(double)*(problem->max_cardinality));
  setIntVal(solution, -1, problem->num_variable);

  int start=0;
  for(int i0=0; i0<problem->num_variable; ++i0){
    i = (i0+start)%(problem->num_variable);
    if(problem->num_sp_of_phi[i]==0){
      SubProblem *p = problem->sp[problem->unary_subproblem_id[i]];
      if(p->min_energy_valid){
	solution[i] = p->sol[0];
      }else{
	p->min_energy = minimize(p->msg[0], p->card[0], &(p->sol[0]));
	p->min_energy_valid = 1;
	solution[i] = p->sol[0];
      }
      continue;
    }
    setDblVal(mmarginal, 0.0, problem->card[i]);
    
    // add all min-marginals 
    for(j=0; j<problem->num_sp_of_phi[i]; ++j){
      SubProblem* p = problem->sp[problem->sp_of_phi[i][j]];
      extractMinMarginal(parameters, p, problem->cy_solver, 1, 
			 problem->scope_of_phi[i], 0.0, 1.0, mmarginal);
    }
    minimize(mmarginal, problem->card[i], &(solution[i]));
  }
  double new_energy = evaluateEnergy(problem, solution);
  if(new_energy < problem->primal_obj){
    problem->primal_obj = new_energy;
    copyIntVec(solution, problem->sol, problem->num_variable);
  }
  free(solution);
}

void renewMRFSolutionRoundoff(MapProblem* problem, MapParameters* parameters)
{
  int i,j;
  int* solution = (int*)malloc(sizeof(int)*(problem->num_variable));
  double** votes = (double**)malloc(sizeof(double*)*(problem->num_variable));
  for(i=0; i<problem->num_variable; ++i){
    votes[i] = (double*)malloc(sizeof(double)*(problem->card[i]));
    setDblVal(votes[i], 0, problem->card[i]);
  }
  // for each subproblem, cast votes
  for(i=0; i<problem->num_subproblem; ++i){
    SubProblem *p = problem->sp[i];
    getSubProblemMin(p, problem->cy_solver, parameters);
    for(j=0; j<p->num_variable; ++j)
      votes[p->var_id[j]][p->sol[j]] ++;
  }
  // pick the label with most vote
  for(i=0; i<problem->num_variable; ++i)
    maximize(votes[i], problem->card[i], &(solution[i]));

  double new_energy = evaluateEnergy(problem, solution);
  if(new_energy < problem->primal_obj){
    problem->primal_obj = new_energy;
    copyIntVec(solution, problem->sol, problem->num_variable);
  }
  for(i=0; i<problem->num_variable; ++i)
    free(votes[i]);
  free(votes);
  free(solution);    
}

void renewMRFSolutionJigsaw(MapProblem* problem, MapParameters* parameters,
			    int fixed_ordering)
{
  int i,j;
  int* solution = (int*)malloc(sizeof(int)*(problem->num_variable));
  setIntVal(solution, -1, problem->num_variable);
  int cnt = 0;
  int start;
  if(fixed_ordering)
    start = 0;
  else
    start = rand()%(problem->num_subproblem);

  for(int i0=0; i0<problem->num_subproblem; ++i0){
    i = (i0+start)%(problem->num_subproblem);
    SubProblem *p = problem->sp[i];
    getSubProblemMin(p, problem->cy_solver, parameters);
    for(j=0; j<p->num_variable; ++j){
      if(solution[p->var_id[j]] == -1){
	solution[p->var_id[j]] = p->sol[j];
	cnt++;
      }
    }
  }
  assert(cnt==problem->num_variable);
  double new_energy = evaluateEnergy(problem, solution);
  if(new_energy < problem->primal_obj){
    problem->primal_obj = new_energy;
    copyIntVec(solution, problem->sol, problem->num_variable);
  }
  free(solution);  
}

int* getOrderOfVariables(MapProblem* problem, int dfs)
{
  //int F = problem->fip->num_f1_nodes[problem->category];
  int F= problem->num_variable;
  int root;
  int cnt=0;
  int* nodes = (int*)malloc(sizeof(int) * F);
  int* color  = (int*)malloc(sizeof(int) * F); //0:white, 1:grey, 2:black
  int* C = (int*)malloc(sizeof(int) * F);
  int current, c0, c1;

  setIntVal(color, 0, F);  

  while(cnt<F){
    do{
      root = rand()%F;
    }while(color[root] > 0);
    c0=0; 
    c1=0;
    C[c1++] = root;
    color[root] = 1;

    while(c0!=c1){
      if(dfs) current = C[--c1]; // use C as stack
      else current = C[c0++];    // use C as queue
      nodes[cnt++] = current;
      color[current] = 2;

      if(problem->num_sp_of_phi[current] == 0) continue;

      int start = rand() % (problem->num_sp_of_phi[current]);
      for(int i0=start; i0<start+problem->num_sp_of_phi[current]; ++i0){ // for each subproblem containing current f1 node
	int i = i0 % (problem->num_sp_of_phi[current]);
	SubProblem* p = problem->sp[problem->sp_of_phi[current][i]];
	//	if(p->type == FUNC_TYPE_GRIDPAIR_LINF_HARD){
	int start1 = rand() % (p->num_variable);
	for(int j0 = start1; j0<start1 + p->num_variable; j0++){
	  int j = j0 % (p->num_variable);
	  //	  int j = (p->var_id[0] == current)?1:0;
	  if(p->var_id[j] == current) continue;
	  int nb = p->var_id[j];
	  if(color[nb] == 0){
	    color[nb] = 1;
	    C[c1++] = nb;
	  }
	}
      }
    }
  }
  //  for(int i=0; i<F; ++i)
  //    assert(color[i] == 2);

  free(color);
  free(C);
  return nodes;
}


void renewMRFSolutionConditional(MapProblem* problem, MapParameters* parameters,
				 int fixed_ordering)
{
  //  clock_t tic = clock();
  int i,j;
  int* solution = (int*)malloc(sizeof(int)*(problem->num_variable));
  double* mmarginal = (double*)malloc(sizeof(double)*(problem->max_cardinality));

  setIntVal(solution, -1, problem->num_variable);

  int simple_random_order = rand()%2;
  int dfs = rand()%2;

  int* order;
  int start;
  if(simple_random_order)
    start = rand()%(problem->num_variable);
  else
    order = getOrderOfVariables(problem, dfs);

  for(int i0=0; i0<problem->num_variable; ++i0){

    if(simple_random_order)
      i = (i0+start)%(problem->num_variable);
    else
      i = order[i0]; 

    // if(problem->num_sp_of_phi[i]==0) continue;
    // assert(problem->num_sp_of_phi[i] > 0);
    if(problem->num_sp_of_phi[i]==0){

      solution[i] = problem->sol[i];
      /*
      if(problem->unary_subproblem_id[i] < 0){
	solution[i] = 0;
	continue;
      }
      SubProblem *p = problem->sp[problem->unary_subproblem_id[i]];
      // assert(p);
      // assert(p->num_variable == 1);
      if(p->min_energy_valid){
	solution[i] = p->sol[0];
      }else{
	p->min_energy = minimize(p->msg[0], p->card[0], &(p->sol[0]));
	p->min_energy_valid = 1;
	solution[i] = p->sol[0];
	}*/
      continue;
    }

    setDblVal(mmarginal, 0.0, problem->card[i]);
    
    // add all min-marginals (conditioned on assigned variables)
    // to mmarginal
    for(j=0; j<problem->num_sp_of_phi[i]; ++j){
      SubProblem* p = problem->sp[problem->sp_of_phi[i][j]];
      if(p->type == FUNC_TYPE_POTTS || p->type == FUNC_TYPE_POTTS0){
	int k,l,K = p->card[0],L = p->card[1];

	if(p->msg[2]){ // has binary message, quadratic time
	  if(i==p->var_id[0]){
	    int var1 = p->var_id[1];
	    if(solution[var1] < 0){
	      // get min marginal
	      assert(K == problem->card[i]);
	      for(k=0; k<K; ++k){
		int e = k;
		double m = p->msg[1][0] + p->msg[2][e] + p->val[k==0?0:1];
		e += K;
		for(l=1; l<L; ++l){
		  m = MIN(m, p->msg[1][l] + p->msg[2][e] + p->val[k==l?0:1]);
		  e += K;
		}
		m += p->msg[0][k];
		mmarginal[k] += m;
	      }

	    }else{
	      // get projection
	      assert(K == problem->card[i]);
	      for(k=0; k<K; ++k){
		l = solution[var1];
		double m = p->msg[0][k] + // p->msg[1][l] + 
		  p->msg[2][k+l*K] + p->val[k==l?0:1];
		mmarginal[k] += m;
	      }
	    }
	  }else{
	    int var1 = p->var_id[0];

	    if(solution[var1] < 0){
	      // get min marginal
	      int e=0;
	      assert(L == problem->card[i]);
	      for(l=0; l<L; ++l){
		double m = p->msg[0][0] + p->msg[2][e] + p->val[l==0?0:1];
		++e;
		for(k=1; k<K; ++k){
		  m = MIN(m, p->msg[0][k] + p->msg[2][e] + p->val[k==l?0:1]);
		  ++e;
		}
		m += p->msg[1][l];
		mmarginal[l] += m;
	      }

	    }else{
	      // get projection
	      assert(L == problem->card[i]);
	      for(l=0; l<L; ++l){
		k = solution[var1];
		double m =  p->msg[1][l] + //p->msg[0][k] +
		  p->msg[2][k+l*K] + p->val[l==k?0:1];
		mmarginal[l] += m;
	      }
	    }
	  }

	}else{// no binary msg, linear time 

	  int var1 = (i == p->var_id[0])? p->var_id[1] : p->var_id[0];
	  int id0 = i == p->var_id[0]?0:1;
	  int id1 = 1-id0;
	  int card0 = i == p->var_id[0]?K:L;
	  int card1 = i == p->var_id[0]?L:K;
	  if(solution[var1] < 0){
	    double min_L = p->msg[id1][0];
	    int min_L_argmin = 0;
	    for(l=1; l<card1; ++l){
	      if(p->msg[id1][l] < min_L){
		min_L = p->msg[id1][l];
		min_L_argmin = l;
	      }
	    }
	    
	    if(p->type == FUNC_TYPE_POTTS0){
	      assert(card0 == problem->card[i]);
	      for(k=0; k<card0; ++k){
		double m = min_L + (k==min_L_argmin?0:p->val[1]);
		if(k<card1)
		  m = MIN(m, p->msg[id1][k] + p->val[0]);
		m += p->msg[id0][k];
		mmarginal[k] += m;
	      }
	    }else if(p->type == FUNC_TYPE_POTTS){
	      assert(card0 == problem->card[i]);
	      for(k=0; k<card0; ++k){
		double m = min_L + p->val[k==min_L_argmin?0:1];
		if(k == min_L_argmin && p->val[1] < p->val[0]){ 
		  // might prefer k!=l
		  for(l=0; l<card1; ++l)
		    m = MIN(m, p->msg[id1][l] + p->val[k==l?0:1]);
		}else if(k != min_L_argmin && k<card1 && p->val[1] > p->val[0]){ 
		  // might prefer k==l
		  m = MIN(m, p->msg[id1][k] + p->val[0]);
		}
		m += p->msg[id0][k];
		mmarginal[k] += m;
	      }
	    }

	  }else{
	    // projection
	    assert(card0 == problem->card[i]);
	    for(k=0; k<card0; ++k){
	      l = solution[var1];
	      double m = p->msg[id0][k] + p->val[k==l?0:1]; //+ p->msg[id1][l];
	      mmarginal[k] += m;
	    }
	  }
	}

      }else if(p->type == FUNC_TYPE_TABLE){

	if(p->num_variable == 1){
	  for(int k=0; k<p->card[0]; ++k)
	    mmarginal[k] += p->msg[0][k];
	}else if(p->num_variable == 2){
	  int k,l,K = p->card[0],L = p->card[1];
	  if(i==p->var_id[0]){
	    if(solution[p->var_id[1]]<0){
	      assert(K == problem->card[i]);
	      for(k=0; k<K; ++k){
		int e = k;
		double m = p->msg[1][0] + (p->msg[2]?p->msg[2][e]:0) + 
		  p->val[e];
		e += K;
		for(l=1; l<L; ++l){
		  m = MIN(m, p->msg[1][l] + (p->msg[2]?p->msg[2][e]:0) + 
			  p->val[e]);
		  e += K;
		}
		m += p->msg[0][k];
		mmarginal[k] += m;
	      }
	    }else{ // projection
	      assert(K == problem->card[i]);
	      for(k=0; k<K; ++k){
		l = solution[p->var_id[1]];
		double m = p->msg[0][k] + // p->msg[1][0] + 
		  (p->msg[2]?p->msg[2][k+l*K]:0) + 
		  p->val[k+l*K];
		mmarginal[k] += m;
	      }
	    }

	  }else{

	    if(solution[p->var_id[0]] < 0){
	      int e=0;
	      assert(L == problem->card[i]);
	      for(l=0; l<L; ++l){
		double m = p->msg[0][0] + (p->msg[2]?p->msg[2][e]:0) + p->val[e];
		++e;
		for(k=1; k<K; ++k){
		  m = MIN(m, p->msg[0][k] + (p->msg[2]?p->msg[2][e]:0) + p->val[e]);
		  ++e;
		}
		m += p->msg[1][l];
		mmarginal[l] += m;
	      }
	    }else{ // projection
	      assert(L == problem->card[i]);
	      for(l=0; l<L; ++l){
		k = solution[p->var_id[0]];
		double m =  p->msg[1][l] + //p->msg[0][0] +
		  (p->msg[2]?p->msg[2][k+l*K]:0) + p->val[k+l*K];
		mmarginal[l] += m;
	      }
	    }
	  }
	}else{
	  int d=0, N = p->num_variable;
	  while(p->var_id[d] != i && d<N) d++;
	  //assert(d<N);
	  getConditionalMinMarginalHighOrderTable(p, d, solution, mmarginal); 
	}

      }else if(p->type == FUNC_TYPE_CYCLE){

	// current variable is 'i'
	// find 'i' in the cycle,
	Cycle* cy = problem->cy_solver;
	int l, d=0, N = p->num_variable;
	while(p->var_id[d] != i && d<N) d++;
	int d0=d;

	do{d=(d+1)%N;}while(solution[p->var_id[d]] < 0 && p->var_id[d]!=i);
	if(p->var_id[d]==i){
	  setupCycleSolver(cy, p);
	  cy->runFullMinSum(d,1);
	  assert(p->card[d] == problem->card[i]);
	  for(l=0; l<p->card[d]; ++l)
	    mmarginal[l] += cy->mmarg[d][l];
	}else{
	  getHalfCycleMinMarginal(p, solution, d0, problem->max_cardinality,
				  mmarginal);
	}

      }else if(p->type == FUNC_TYPE_ISING){ 
	int id0 = i == p->var_id[0]?0:1;
	int id1 = 1-id0;
	if(solution[p->var_id[id1]]<0){
	  double buf[2];
	  buf[0] = MIN(p->msg[id1][0] + (p->msg[2]?p->msg[2][0]:0), 
		       p->msg[id1][1] + p->val[0] + 
		       (p->msg[2]?p->msg[2][id0==0?2:1]:0));
	  buf[1] = MIN(p->msg[id1][1] + (p->msg[2]?p->msg[2][3]:0), 
		       p->msg[id1][0] + p->val[0] + 
		       (p->msg[2]?p->msg[2][id0==0?1:2]:0)) ;
	  mmarginal[0] += (buf[0] + p->msg[id0][0]);
	  mmarginal[1] += (buf[1] + p->msg[id0][1]);
	}else{
	  if(solution[p->var_id[id1]] == 0){
	    mmarginal[0] += p->msg[id0][0] + (p->msg[2]?p->msg[2][0]:0);
	    mmarginal[1] += p->msg[id0][1] + p->val[0] + 
	      (p->msg[2]?p->msg[2][id0==0?1:2]:0);
	  }else{
	    mmarginal[0] += p->msg[id0][0] + p->val[0] + 
	      (p->msg[2]?p->msg[2][id0==0?2:1]:0);
	    mmarginal[1] += p->msg[id0][1] + (p->msg[2]?p->msg[2][3]:0);
	  }
	}
      }else if(p->type == FUNC_TYPE_TL1){ 

	int k,l,K = p->card[0],L = p->card[1];

	if(p->msg[2]){ // has binary message, quadratic time
	  if(i==p->var_id[0]){
	    int var1 = p->var_id[1];
	    if(solution[var1] < 0){
	      // get min marginal
	      assert(K == problem->card[i]);
	      for(k=0; k<K; ++k){
		int e = k;
		double m = p->msg[1][0] + p->msg[2][e] + p->val[k];
		e += K;
		for(l=1; l<L; ++l){
		  m = MIN(m, p->msg[1][l] + p->msg[2][e] + p->val[abs(k-l)]);
		  e += K;
		}
		m += p->msg[0][k];
		mmarginal[k] += m;
	      }

	    }else{
	      // get projection
	      assert(K == problem->card[i]);
	      for(k=0; k<K; ++k){
		l = solution[var1];
		double m = p->msg[0][k] + // p->msg[1][l] + 
		  p->msg[2][k+l*K] + p->val[abs(k-l)];
		mmarginal[k] += m;
	      }
	    }
	  }else{
	    int var1 = p->var_id[0];

	    if(solution[var1] < 0){
	      // get min marginal
	      int e=0;
	      assert(L == problem->card[i]);
	      for(l=0; l<L; ++l){
		double m = p->msg[0][0] + p->msg[2][e] + p->val[l];
		++e;
		for(k=1; k<K; ++k){
		  m = MIN(m, p->msg[0][k] + p->msg[2][e] + p->val[abs(k-l)]);
		  ++e;
		}
		m += p->msg[1][l];
		mmarginal[l] += m;
	      }

	    }else{
	      // get projection
	      assert(L == problem->card[i]);
	      for(l=0; l<L; ++l){
		k = solution[var1];
		double m =  p->msg[1][l] + //p->msg[0][k] +
		  p->msg[2][k+l*K] + p->val[abs(k-l)];
		mmarginal[l] += m;
	      }
	    }
	  }
	}else{ // no msg[2], linear time

	  int k,l, K = p->card[0],L = p->card[1];
	  assert(K==L);
	  int id0 = i == p->var_id[0]?0:1;
	  int id1 = 1-id0;
	  if(solution[p->var_id[id1]]<0){

	    double alpha = p->val[1];
	    double gamma = p->val[K-1];
	    double* buf = (double*)(parameters->buffer);
	    buf[0] = p->msg[id1][0];
	    double vMin = p->msg[id1][0];
	    for(l=1; l<L; ++l){
	      buf[l] = p->msg[id1][l];
	      if(buf[l] > p->msg[id1][l-1] + alpha)
		buf[l] = p->msg[id1][l-1] + alpha;
	      if(vMin > p->msg[id1][l] /*buf[l]*/)
		vMin = p->msg[id1][l]; //buf[l];
	    }
	    l--;
	    if (buf[l] > vMin + gamma)
	      buf[l] = vMin + gamma;
	    for(l--; l>=0; l--){
	      if(buf[l] > buf[l+1] + alpha)
		buf[l] = buf[l+1] + alpha;
	      if (buf[l] > vMin + gamma)
		buf[l] = vMin + gamma;
	    }
	    assert(K == problem->card[i]);
	    for(k=0; k<K; ++k){
	      double m = buf[k] + p->msg[id0][k];
	      mmarginal[k] += m;
	    }

	  }else{ // projection
	    assert(K == problem->card[i]);
	    for(k=0; k<K; ++k){  // for each state of ``this variable''
	      l = solution[p->var_id[id1]];
	      double m =  p->msg[id0][k] + //p->msg[id1][l] +
		p->val[abs(k-l)];
	      mmarginal[k] += m;
	    }
	  }
	}

      }else if(p->type == FUNC_TYPE_TL2){

	int k,l,K = p->card[0],L = p->card[1];
	if(p->msg[2]){ // has binary message, quadratic time
	  if(i==p->var_id[0]){
	    int var1 = p->var_id[1];
	    if(solution[var1] < 0){
	      // get min marginal
	      assert(K == problem->card[i]);
	      for(k=0; k<K; ++k){
		int e = k;
		double m = p->msg[1][0] + p->msg[2][e] + p->val[K+K+k];
		e += K;
		for(l=1; l<L; ++l){
		  m = MIN(m, p->msg[1][l] + p->msg[2][e] + p->val[K+K+abs(k-l)]);
		  e += K;
		}
		m += p->msg[0][k];
		mmarginal[k] += m;
	      }

	    }else{
	      // get projection
	      assert(K == problem->card[i]);
	      for(k=0; k<K; ++k){
		l = solution[var1];
		double m = p->msg[0][k] + // p->msg[1][l] + 
		  p->msg[2][k+l*K] + p->val[K+K+abs(k-l)];
		mmarginal[k] += m;
	      }
	    }
	  }else{
	    int var1 = p->var_id[0];

	    if(solution[var1] < 0){
	      // get min marginal
	      int e=0;
	      assert(L == problem->card[i]);
	      for(l=0; l<L; ++l){
		double m = p->msg[0][0] + p->msg[2][e] + p->val[K+K+l];
		++e;
		for(k=1; k<K; ++k){
		  m = MIN(m, p->msg[0][k] + p->msg[2][e] + p->val[K+K+abs(k-l)]);
		  ++e;
		}
		m += p->msg[1][l];
		mmarginal[l] += m;
	      }

	    }else{
	      // get projection
	      assert(L == problem->card[i]);
	      for(l=0; l<L; ++l){
		k = solution[var1];
		double m =  p->msg[1][l] + //p->msg[0][k] +
		  p->msg[2][k+l*K] + p->val[K+K+abs(k-l)];
		mmarginal[l] += m;
	      }
	    }
	  }
	}else{ // no msg[2], linear time

	  int k,l, K = p->card[0],L = p->card[1];
	  assert(K==L);
	  int id0 = i == p->var_id[0]?0:1;
	  int id1 = 1-id0;
	  if(solution[p->var_id[id1]]<0){
	    double alpha = p->val[K+1];
	    double gamma = p->val[0];

	    double* buf = (double*)(parameters->buffer);
	    int* intbuf = (int*)(buf + K);
	    int* parabolas = intbuf + K;
	    int* intersections = parabolas + K;
	    double vMin = p->msg[id1][0];
	    for(l=1; l<L; ++l)
	      if(vMin > p->msg[id1][l] /*buf[l]*/)
		vMin = p->msg[id1][l]; //buf[l];
	    if(alpha == 0){ 
	      assert(K == problem->card[i]);
	      for(k=0; k<K; ++k){
		double m = vMin + p->msg[id0][k];
		mmarginal[k] += m;
	      }
	    }else{
	      DistanceTransformL2(K, p->msg[id1], buf, NULL,
				  parabolas, intersections, p->val);
	      for(l=0; l<L; ++l)
		if (buf[l] > vMin + gamma)
		  buf[l] = vMin + gamma;
	      assert(K == problem->card[i]);
	      for(k=0; k<K; ++k){
		double m = buf[k] + p->msg[id0][k];
		mmarginal[k] += m;
	      }
	    }
	  }else{ // projection
	    assert(K == problem->card[i]);
	    for(k=0; k<K; ++k){  // for each state of ``this variable''
	      l = solution[p->var_id[id1]];
	      double m = p->msg[id0][k] + //p->msg[id1][l] + 
		p->val[K+K+abs(k-l)];
	      mmarginal[k] += m;
	    }
	  }
	}

      }else{
	printf("ERR: renewMRFSolution to be implemented (new type)\n");
	assert(p->type == FUNC_TYPE_CYCLE);
	assert(0);
      }
    }
    minimize(mmarginal, problem->card[i], &(solution[i]));
  }

  //  assert(fabs((problem->primal_obj - 
  //	       evaluateEnergy(problem, problem->sol))/(problem->primal_obj))<1e-6);

  double new_energy = evaluateEnergy(problem, solution);
  if(new_energy < problem->primal_obj){
    problem->primal_obj = new_energy;
    copyIntVec(solution, problem->sol, problem->num_variable);
  }

  free(solution);
  free(mmarginal);
  if(!simple_random_order)
    free(order);
}

void renewMRFSolution(MapProblem* problem, MapParameters* parameters)
{
  if(parameters->primal_method == PRIMAL_INDEPENDENT){
    renewMRFSolutionIndependent(problem, parameters);
  }else if(parameters->primal_method == PRIMAL_JIGSAW_FIXED || 
	   parameters->primal_method == PRIMAL_JIGSAW_RAND){
    renewMRFSolutionJigsaw(problem, parameters, 
			   parameters->primal_method == 
			   PRIMAL_JIGSAW_FIXED);
  }else if(parameters->primal_method == PRIMAL_CONDITIONAL_FIXED || 
	   parameters->primal_method == PRIMAL_CONDITIONAL_RAND){
    renewMRFSolutionConditional(problem, parameters, 
				parameters->primal_method == 
				PRIMAL_CONDITIONAL_FIXED);
  }else if(parameters->primal_method == PRIMAL_ROUNDOFF){
    renewMRFSolutionRoundoff(problem, parameters);
  }
}

void freeSubProblem(SubProblem * p)
{
  int j;
  if(p->type != FUNC_TYPE_CYCLE)
    free(p->var_id);
  free(p->sol);
  free(p->card);
  if(p->num_msg>0){
    for(j=0; j<p->num_msg; ++j){
      if(p->msg[j])
	free(p->msg[j]);
    }
    free(p->msg);
  }
}

void allocateSubProblemTree(MapProblem* problem)
{
  int i;
  problem->sptree = (SubProblemTree*)malloc(sizeof(SubProblemTree));
  SubProblemTree* t = problem->sptree;
  t->max_num_edges = problem->num_subproblem-1;
  int N = problem->num_subproblem;
  int M = t->max_num_edges;
  t->sp_id = (int**)malloc(sizeof(int*)*M);
  t->phi_id = (int*)malloc(sizeof(int)*M);
  t->subtree_weight = (double*)malloc(sizeof(double)*N);
  t->subtree_weight_dynamic = (double*)malloc(sizeof(double)*N);
  t->nb_size = N*10;
  t->nb_sp = (int*)malloc(sizeof(int)*(t->nb_size));
  t->nb_phi = (int*)malloc(sizeof(int)*(t->nb_size));
  t->candidates = (int*)malloc(sizeof(int)*N);
  t->color = (int*)malloc(sizeof(int)*N);
  t->parent_sp = (int*)malloc(sizeof(int)*N);
  t->parent_phi = (int*)malloc(sizeof(int)*N);
  for(i=0; i<M; ++i)
    t->sp_id[i] = (int*)malloc(sizeof(int)*2);
}

void freeSubProblemTree(MapProblem* problem)
{
  SubProblemTree* t = problem->sptree;
  int i, M = t->max_num_edges;
  for(i=0; i<M; ++i)
    free(t->sp_id[i]);
  free(t->sp_id);
  free(t->phi_id);
  free(t->subtree_weight);
  free(t->subtree_weight_dynamic);
  free(t->nb_sp);
  free(t->nb_phi);
  free(t->candidates);
  free(t->color);
  free(t->parent_sp);
  free(t->parent_phi);
  free(problem->sptree);
}

void  convertPottsIntoTable(MapProblem* problem)
{
  printf("convertPottsIntoTable\n");
  int K = problem->card[0]; // ONLY WORKS FOR UNIFORM CARDINALITY
  for(int i=0; i<problem->num_function; ++i){
    Function* f = &(problem->func[i]);
    if(f->type == FUNC_TYPE_POTTS){
      f->type = FUNC_TYPE_TABLE;
      f->size_val = K*K;
      double v0 = f->val[0];
      double v1 = f->val[1];
      free(f->val);
      f->val = (double*)malloc(sizeof(double)*K*K);
      setDblVal(f->val, v1, K*K);
      for(int k=0; k<K; ++k)
	f->val[k+k*K] = v0;
    }
  }
  for(int i=0; i<problem->num_subproblem; ++i){
    SubProblem* p = problem->sp[i];
    p->type = problem->func[p->func_id].type;
    p->val =  problem->func[p->func_id].val;
    //    p->size_val =  problem->func[p->func_id].size_val;
  }
}

void checkConnectivity(MapProblem* problem)
{
  int total_visited = 1;
  int N = problem->num_variable;
  int* visited = (int*)malloc(sizeof(int)*N);
  setIntVal(visited, 0, N);
  visited[0] = 1;
  int changed = 1;
  while(changed){
    changed = 0;
    for(int i=0; i<problem->num_subproblem; ++i){
      int num_visited = 0;
      for(int j=0; j<problem->sp[i]->num_variable; ++j)
	num_visited += visited[problem->sp[i]->var_id[j]];
      if(num_visited > 0 && num_visited < problem->sp[i]->num_variable){
	changed = 1;
	for(int j=0; j<problem->sp[i]->num_variable; ++j)
	  visited[problem->sp[i]->var_id[j]] = 1;
	total_visited += problem->sp[i]->num_variable - num_visited;
      }
    }
  }
  if(total_visited == N) 
    printf("graph IS connected\n");
  else
    printf("graph IS NOT connected\n");
}

int extractMinMarginal(MapParameters* parameters, SubProblem* p, Cycle* cy, int dim,
		       int* scope, double factor0, double factor1,
		       double* target)
{
  if(p->min_energy_valid)
    p->min_energy *= (1.0-factor0);

  if(p->num_variable==1){ // unary
    for(int k=0; k<p->card[0]; ++k){
      double m = p->msg[0][k];
      p->msg[0][k] -= factor0*m;
      target[k] += factor1*m;
    }
    return 0;
  }

  
  if(p->type == FUNC_TYPE_POTTS || p->type == FUNC_TYPE_POTTS0){

    int k,l,K = p->card[0],L = p->card[1];

    if(p->msg[2]){ // has binary message, quadratic time

      if(dim==1){
	int argmin=0;
	if(scope[0] == p->var_id[0]){
	  for(k=0; k<K; ++k){
	    int e = k; // k+l*K
	    double m = p->msg[1][0] + p->msg[2][e] + p->val[k==0?0:1];
	    e += K;
	    for(l=1; l<L; ++l){
	      m = MIN(m, p->msg[1][l] + p->msg[2][e]  + p->val[k==l?0:1]);
	      e += K;
	    }
	    m += p->msg[0][k];

	    p->msg[0][k] -= factor0*m;
	    target[k] += factor1*m;
	  }
	  return argmin;
	}else{

	  int e=0; // k+l*K
	  for(l=0; l<L; ++l){
	    double m = p->msg[0][0] + p->msg[2][e] + p->val[l==0?0:1];
	    ++e;
	    for(k=1; k<K; ++k){
	      m = MIN(m, p->msg[0][k] + p->msg[2][e] + p->val[k==l?0:1]);
	      ++e;
	    }
	    m += p->msg[1][l];

	    p->msg[1][l] -= factor0*m;
	    target[l] += factor1*m;
	  }
	  return argmin;
	}
      }else{

	int e=0;
	int argmin=0;
	int transpose = (scope[0] == p->var_id[0])?0:1;
	for(l=0; l<L; ++l){
	  for(k=0; k<K; ++k){
	    double m = p->msg[0][k] + p->msg[1][l] + 
	      p->msg[2][e] + p->val[k==l?0:1];

	    p->msg[2][e] -= factor0*m;
	    target[transpose?(l+k*L):e] += factor1*m;
	    ++e; // k+l*K
	  }
	}
	return argmin;
      }
    }else{ // no binary message, linear time,
      if(dim==1){


	int argmin=0;
	int id0 = scope[0] == p->var_id[0]?0:1;
	int id1 = 1-id0;
	int card0 = scope[0] == p->var_id[0]?K:L;
	int card1 = scope[0] == p->var_id[0]?L:K;
	double min_L = p->msg[id1][0];
	int min_L_argmin = 0;
	for(l=1; l<card1; ++l){
	  if(p->msg[id1][l] < min_L){
	    min_L = p->msg[id1][l];
	    min_L_argmin = l;
	  }
	}

	if(p->type == FUNC_TYPE_POTTS0){

	  for(k=0; k<card0; ++k){
	    double m = min_L + (k==min_L_argmin?0:p->val[1]);
	    if(k<card1)
	      m = MIN(m, p->msg[id1][k] + p->val[0]);
	    m += p->msg[id0][k];

	    p->msg[id0][k] -= factor0*m;
	    target[k] += factor1*m;
	  }
	}else if(p->type == FUNC_TYPE_POTTS){
	  for(k=0; k<card0; ++k){
	    double m = min_L + p->val[k==min_L_argmin?0:1];
	    if(k == min_L_argmin && p->val[1] < p->val[0]){ 
	      // might prefer k!=l
	      for(l=0; l<card1; ++l)
		m = MIN(m, p->msg[id1][l] + p->val[k==l?0:1]);
	    }else if(k != min_L_argmin && k<card1 && p->val[1] > p->val[0]){ 
	      // might prefer k==l
	      m = MIN(m, p->msg[id1][k] + p->val[0]);
	    }
	    m += p->msg[id0][k];

	    p->msg[id0][k] -= factor0*m;
	    target[k] += factor1*m;
	  }

	}
	return argmin;

      }else{

	int e=0;

	int argmin = 0;
	if(scope[0] == p->var_id[0]){
	  for(l=0; l<L; ++l){
	    for(k=0; k<K; ++k){
	      double m = p->msg[0][k] + p->msg[1][l] + p->val[k==l?0:1];
	      target[e] += factor1*m;
	      ++e;
	    }
	  }
	}else{
	  for(k=0; k<K; ++k){
	    for(l=0; l<L; ++l){
	      double m = p->msg[0][k] + p->msg[1][l] + p->val[k==l?0:1];
	      target[e] += factor1*m; 
	      ++e;
	    }
	  }
	}
	return argmin;
      }
    }


  }else if(p->type == FUNC_TYPE_TABLE){

    if(p->num_variable == 2){
      int k,l,K = p->card[0],L = p->card[1];
      if(p->msg[2]){
	if(dim==1){

	  int argmin = 0;

	  if(scope[0] == p->var_id[0]){
	    for(k=0; k<K; ++k){
	      int e = k;
	      double m = p->msg[1][0] + p->msg[2][e] + p->val[e];
	      e += K;
	      for(l=1; l<L; ++l){
		m = MIN(m, p->msg[1][l] + p->msg[2][e] + p->val[e]);
		e += K;
	      }
	      m += p->msg[0][k];
	      p->msg[0][k] -= factor0*m;
	      target[k] += factor1*m;
	    }
	    return argmin;
	  }else{
	    int e=0;
	    for(l=0; l<L; ++l){
	      double m = p->msg[0][0] + p->msg[2][e] + p->val[e];
	      ++e;
	      for(k=1; k<K; ++k){
		m = MIN(m, p->msg[0][k] + p->msg[2][e] + p->val[e]);
		++e;
	      }
	      m += p->msg[1][l];
	      p->msg[1][l] -= factor0*m;
	      target[l] += factor1*m;
	    }
	    return argmin;
	  }
	}else{


	  int e=0;
	  int argmin = 0;
	  int transpose = (scope[0] == p->var_id[0])?0:1;
	  for(l=0; l<L; ++l){
	    for(k=0; k<K; ++k){
	      double m = p->msg[0][k] + p->msg[1][l] + p->msg[2][e] + p->val[e];
	      p->msg[2][e] -= factor0*m;
	      target[transpose?(l+k*L):e] += factor1*m;
	      ++e;
	    }
	  }
	  return argmin;
	}

      }else{
	if(dim==1){
	  int argmin = 0;
	  if(scope[0] == p->var_id[0]){
	    for(k=0; k<K; ++k){
	      int e = k;
	      double m = p->msg[1][0] + p->val[e];
	      e += K;
	      for(l=1; l<L; ++l){
		m = MIN(m, p->msg[1][l] + p->val[e]);
		e += K;
	      }
	      m += p->msg[0][k];

	      p->msg[0][k] -= factor0*m;
	      target[k] += factor1*m;
	    }
	    return argmin;
	  }else{
	    int e=0;
	    for(l=0; l<L; ++l){
	      double m = p->msg[0][0] + p->val[e];
	      ++e;
	      for(k=1; k<K; ++k){
		m = MIN(m, p->msg[0][k] + p->val[e]);
		++e;
	      }
	      m += p->msg[1][l];

	      p->msg[1][l] -= factor0*m;
	      target[l] += factor1*m;
	    }
	    return argmin;
	  }
	}else{
	  int e=0;

	  int argmin = 0;
	  if(scope[0] == p->var_id[0]){
	    for(l=0; l<L; ++l){
	      for(k=0; k<K; ++k){
		double m = p->msg[0][k] + p->msg[1][l] + p->val[e];

		target[e] += factor1*m;
		++e;
	      }
	    }
	    return argmin;
	  }else{
	    for(l=0; l<L; ++l){
	      for(k=0; k<K; ++k){
		double m = p->msg[0][k] + p->msg[1][l] + p->val[e];

		target[l+k*L] += factor1*m; 
		++e; // e = k+l*K
	      }
	    }
	    return argmin;
	  }
	}
      }
    }else{
      if(dim==1){
	int d=0,  N = p->num_variable;
	while(p->var_id[d] != scope[0] && d<N) d++;
	assert(d<N);
	return getMinMarginalHighOrderTable(p, d, factor0, factor1, target);
      }else{
	int i0=0, i1=0, N=p->num_variable;
	while(p->var_id[i0] != scope[0] && i0 < N) i0++;
	while(p->var_id[i1] != scope[1] && i1 < N) i1++;
	assert(i0<N);
	assert(i1<N);
	return getPairMinMarginalHighOrderTable(p, i0, i1, factor0, factor1, target);
      }
    }

  }else if(p->type == FUNC_TYPE_CYCLE){

    int idx, d=0;
    int transpose = 0;
    int k, l, K, L, N = p->num_variable;

    if(dim==1){
      while(p->var_id[d] != scope[0] && d<N) d++;

      assert(d<N);
      K = p->card[d];
      setupCycleSolver(cy, p);
      //      assert(d<N);
      //      assert(cy->mmarg[d+N]);
      cy->runFullMinSum(d, 1);

      for(k=0; k<p->card[d]; ++k){
	p->msg[d][k] -= factor0*(cy->mmarg[d][k]);
	target[k] += factor1*(cy->mmarg[d][k]);
      }
      p->min_energy = (1.0-factor0)*cy->obj;
      p->min_energy_valid=1;
      return cy->assignment[d];

    }else{ // dim==2

      while(p->var_id[d] != scope[0] && d<N) d++;

      if(p->var_id[d+1] == scope[1]){
	idx = N+d;
	K = p->card[d];  // scope[0]
	L = p->card[d+1]; // scope[1]
	d = (d+1)%N;
      }else{
	transpose = 1;

	if(d==0) idx = 2*N-1;
	else idx = N+d-1;
	K = p->card[(N+(d-1))%N]; // scope[1]
	L = p->card[d];           // scope[0]
      }
      assert(d<N);
      assert(cy->mmarg[d+N]);
      setupCycleSolver(cy, p);
      cy->runFullMinSum(d, 1);


      int e=0;
      for(l=0; l<L; ++l){
	for(k=0; k<K; ++k){
	  p->msg[idx][e] -= factor0*(cy->mmarg[N+d][e]);
	  target[transpose?(l+k*L):e] += factor1*(cy->mmarg[N+d][e]);
	  ++e;
	}
      }
    }
    p->min_energy = (1.0-factor0)*cy->obj;
    p->min_energy_valid=1;
    return transpose? 
      (cy->assignment[d] + cy->assignment[(N+(d-1))%N]*(cy->card[d])):
      (cy->assignment[(d+N-1)%N] + cy->assignment[d]*(cy->card[(d+N-1)%N]));

  }else if(p->type == FUNC_TYPE_ISING){

    if(dim == 1){
      int id0 = scope[0] == p->var_id[0]?0:1;
      int id1 = 1-id0;
      double buf[2];
      buf[0] = MIN(p->msg[id1][0] + (p->msg[2]?p->msg[2][0]:0), 
		   p->msg[id1][1] + p->val[0] + 
		   (p->msg[2]?p->msg[2][id0==0?2:1]:0));
      buf[1] = MIN(p->msg[id1][1] + (p->msg[2]?p->msg[2][3]:0), 
		   p->msg[id1][0] + p->val[0] + 
		   (p->msg[2]?p->msg[2][id0==0?1:2]:0)) ;
      double m0 = buf[0] + p->msg[id0][0];
      double m1 = buf[1] + p->msg[id0][1];
      p->msg[id0][1] -= factor0 * m1;
      target[1] += factor1 * m1;
      p->msg[id0][0] -= factor0 * m0;
      target[0] += factor1 * m0;
      return 0;

    }else{

      int transpose = (scope[0] == p->var_id[0])?0:1;
      int e=0;
      for(int l=0; l<2; ++l){
	for(int k=0; k<2; ++k){
	  double m = p->msg[0][k] + p->msg[1][l] +
	    (k==l?0:p->val[0]);
	  if(p->msg[2]){
	    m += p->msg[2][e];
	    p->msg[2][e] -= factor0*m;
	  }
	  target[transpose?(l+k*2):e] += factor1*m;
	  ++e;
	}
      }
      return 0;
    }

  }else if(p->type == FUNC_TYPE_TL1){

    int k,l,K = p->card[0],L = p->card[1];

    if(p->msg[2]){ // has binary message, quadratic time

      if(dim==1){

	int argmin=0;
	if(scope[0] == p->var_id[0]){
	  for(k=0; k<K; ++k){
	    int e = k; // k+l*K
	    double m = p->msg[1][0] + p->msg[2][e] + p->val[k];
	    e += K;
	    for(l=1; l<L; ++l){
	      m = MIN(m, p->msg[1][l] + p->msg[2][e]  + p->val[abs(k-l)]);
	      e += K;
	    }
	    m += p->msg[0][k];


	    p->msg[0][k] -= factor0*m;
	    target[k] += factor1*m;
	  }
	  return argmin;
	}else{

	  int e=0; // k+l*K
	  for(l=0; l<L; ++l){
	    double m = p->msg[0][0] + p->msg[2][e] + p->val[l];
	    ++e;
	    for(k=1; k<K; ++k){
	      m = MIN(m, p->msg[0][k] + p->msg[2][e] + p->val[abs(k-l)]);
	      ++e;
	    }
	    m += p->msg[1][l];

	    p->msg[1][l] -= factor0*m;
	    target[l] += factor1*m;
	  }
	  return argmin;
	}
      }else{
	int e=0;
	int argmin=0;
	int transpose = (scope[0] == p->var_id[0])?0:1;
	for(l=0; l<L; ++l){
	  for(k=0; k<K; ++k){
	    double m = p->msg[0][k] + p->msg[1][l] + 
	      p->msg[2][e] + p->val[abs(k-l)];


	    p->msg[2][e] -= factor0*m;
	    target[transpose?(l+k*L):e] += factor1*m;
	    ++e; // k+l*K
	  }
	}
	return argmin;
      }
    }else{ // no binary message, linear time,

      int k, l, K = p->card[0],L = p->card[1];
      if(dim == 1){
	int id0 = scope[0] == p->var_id[0]?0:1;
	int id1 = 1-id0;
	assert(K==L);
	double alpha = p->val[1];
	double gamma = p->val[K-1];
	//double* buf = (double*)malloc(sizeof(double)*K);
	double* buf = (double*)(parameters->buffer);
	buf[0] = p->msg[id1][0];
	double vMin = p->msg[id1][0];
	for(l=1; l<L; ++l){
	  buf[l] = p->msg[id1][l];
	  if(buf[l] > p->msg[id1][l-1] + alpha)
	    buf[l] = p->msg[id1][l-1] + alpha;
	  if(vMin > p->msg[id1][l] /*buf[l]*/)
	    vMin = p->msg[id1][l]; //buf[l];
	}
	l--;
	if (buf[l] > vMin + gamma)
	  buf[l] = vMin + gamma;
	for(l--; l>=0; l--){
	  if(buf[l] > buf[l+1] + alpha)
	    buf[l] = buf[l+1] + alpha;
	  if (buf[l] > vMin + gamma)
	    buf[l] = vMin + gamma;
	}
	for(k=0; k<K; ++k){
	  double m = buf[k] + p->msg[id0][k];
	  p->msg[id0][k] -= factor0*m;
	  target[k] += factor1*m;
	}
	//free(buf);
	return 0;
      }else{ // dim == 2
	assert(factor0==0);
	int transpose = (scope[0] == p->var_id[0])?0:1;
	int e=0;
	for(l=0; l<L; ++l){
	  for(k=0; k<K; ++k){
	    double m = p->msg[0][k] + p->msg[1][l] 
	      + p->val[abs(k-l)];
	    //	    + MIN(p->val[1], p->val[0]*abs(k-l));
	    target[transpose?(l+k*L):e] += factor1*m;
	    ++e;
	  }
	}
	return 0;
      }
    }

  }else if(p->type == FUNC_TYPE_TL2){
    int k,l,K = p->card[0],L = p->card[1];

    if(p->msg[2]){ // has binary message, quadratic time

      if(dim==1){

	int argmin=0;
	if(scope[0] == p->var_id[0]){
	  for(k=0; k<K; ++k){
	    int e = k; // k+l*K
	    double m = p->msg[1][0] + p->msg[2][e] + p->val[K+K+k];
	    e += K;
	    for(l=1; l<L; ++l){
	      m = MIN(m, p->msg[1][l] + p->msg[2][e]  + p->val[K+K+abs(k-l)]);
	      e += K;
	    }
	    m += p->msg[0][k];


	    p->msg[0][k] -= factor0*m;
	    target[k] += factor1*m;
	  }
	  return argmin;
	}else{

	  int e=0; // k+l*K
	  for(l=0; l<L; ++l){
	    double m = p->msg[0][0] + p->msg[2][e] + p->val[K+K+l];
	    ++e;
	    for(k=1; k<K; ++k){
	      m = MIN(m, p->msg[0][k] + p->msg[2][e] + p->val[K+K+abs(k-l)]);
	      ++e;
	    }
	    m += p->msg[1][l];

	    p->msg[1][l] -= factor0*m;
	    target[l] += factor1*m;
	  }
	  return argmin;
	}
      }else{

	int e=0;

	int argmin=0;
	int transpose = (scope[0] == p->var_id[0])?0:1;
	for(l=0; l<L; ++l){
	  for(k=0; k<K; ++k){
	    double m = p->msg[0][k] + p->msg[1][l] + 
	      p->msg[2][e] + p->val[K+K+abs(k-l)];


	    p->msg[2][e] -= factor0*m;
	    target[transpose?(l+k*L):e] += factor1*m;
	    ++e; // k+l*K
	  }
	}
	return argmin;
      }
    }else{ // no binary message, linear time,

      int k, l, K = p->card[0],L = p->card[1];
      if(dim == 1){
	int id0 = scope[0] == p->var_id[0]?0:1;
	int id1 = 1-id0;
	assert(K==L);
	double alpha = p->val[K+1];
	double gamma = p->val[0];
	double* buf = (double*)(parameters->buffer);
	int* parabolas = (int*)(buf + K);
	int* intersections = parabolas + K;
	double vMin = p->msg[id1][0];
	for(l=1; l<L; ++l)
	  if(vMin > p->msg[id1][l] /*buf[l]*/)
	    vMin = p->msg[id1][l]; //buf[l];
	if(alpha == 0){ 
	  for(k=0; k<K; ++k){
	    double m = vMin + p->msg[id0][k];
	    p->msg[id0][k] -= factor0*m;
	    target[k] += factor1*m;
	  }
	}else{
	  DistanceTransformL2(K, p->msg[id1], buf, NULL,
			      parabolas, intersections, p->val);
	  for(l=0; l<L; ++l)
	    if (buf[l] > vMin + gamma)
	      buf[l] = vMin + gamma;
	  for(k=0; k<K; ++k){
	    double m = buf[k] + p->msg[id0][k];
	    p->msg[id0][k] -= factor0*m;
	    target[k] += factor1*m;
	  }
	}
	/*
	  free(buf);
	  free(parabolas);
	  free(intersections);
	*/
	return 0;

      }else{ // dim == 2
	assert(factor0==0);
	int transpose = (scope[0] == p->var_id[0])?0:1;
	int e=0;
	for(l=0; l<L; ++l){
	  for(k=0; k<K; ++k){
	    double m = p->msg[0][k] + p->msg[1][l] +
	      p->val[K+K+abs(k-l)];
	  
	    //	    + MIN(p->val[1], p->val[0]*(k-l)*(k-l));

	    target[transpose?(l+k*L):e] += factor1*m;
	    ++e;
	  }
	}
	return 0;
      }
    }

  }else{
    printf("ERR: extractMinMarginal to be implemented (new type)\n");
    assert(p->type == FUNC_TYPE_CYCLE);
    assert(0);
  }


}

int getMessageHandle(SubProblem* p, int dim, int* scope, int* transpose)
{
  p->min_energy_valid = 0;
  *transpose = 0;

  if(p->num_variable==2){
    if(dim==1){
      if(scope[0] == p->var_id[0]) return 0; //p->msg[0];
      else{

	return 1;//p->msg[1];
      }
    }else{
      assert(scope[0] == p->var_id[0] && scope[1] == p->var_id[1]);
      return 2; //p->msg[2];
    }
  }else if(p->type == FUNC_TYPE_CYCLE){
    int N = p->num_variable;
    if(dim==1){
      int i=0;
      while(p->var_id[i] != scope[0] && i<N) ++i;

      //      p->unary_msg[i] = 1;
      return i; //p->msg[i];
    }else{
      int i=0;
      while(p->var_id[i] != scope[0] && i<N) ++i;

      if(p->var_id[i+1] == scope[1]){
	return N+i; //p->msg[N+i];
      }else{
	*transpose = 1;

	if(i==0) i=N;
	return N+i-1; //p->msg[N+i-1];
      }
    }

  }else if(p->type == FUNC_TYPE_TABLE){

    int N = p->num_variable;
    if(dim==1){
      int i=0;
      while(p->var_id[i] != scope[0] && i<N) ++i;
      assert(i<N);
      return i;
    }else{
      int i0=0, i1=0;
      while(p->var_id[i0] != scope[0] && i0<N) ++i0;
      while(p->var_id[i1] != scope[1] && i1<N) ++i1;
      if(i0<i1){
	*transpose = 0;
	return N + i0*N + i1;
      }else{
	*transpose = 1;
	return N + i1*N + i0;
      }
    }

  }else{
    printf("ERR: getMessageHandle to be implemented (new type)\n");
    assert(p->type == FUNC_TYPE_CYCLE);
    assert(0);
  }
}



void addMessage(SubProblem* p, int dim,
		int* scope, double factor,
		double* message)
{
  p->min_energy_valid = 0;
  int i,k,l;

  if(p->num_variable==2){
    if(dim==1){

      i = scope[0] == p->var_id[0]?0:1;
      for(k=0; k<p->card[i]; ++k)
	p->msg[i][k] += factor*message[k];
    }else{

      //     assert(dim==2);
      assert(scope[0] == p->var_id[0] && scope[1] == p->var_id[1]);
      for(k=0; k<p->card[0]*p->card[1]; ++k)
	p->msg[2][k] += factor*message[k];
    }
  }else if(p->type == FUNC_TYPE_CYCLE){
    int N = p->num_variable;
    if(dim==1){

      i=0;
      while(p->var_id[i] != scope[0] && i<N) ++i;
      //      p->unary_msg[i] = 1;
      for(k=0; k<p->card[i]; ++k)
	p->msg[i][k] += factor*message[k];

    }else{ // dim==2
      int i=0;
      while(p->var_id[i] != scope[0] && i<N) ++i;

      int e=0,K,L;
      if(p->var_id[i+1] == scope[1]){
	K = p->card[i]; L = p->card[i+1];
	for(k=0; k<K*L; ++k)
	  p->msg[N+i][k] += factor*message[k];
      }else{
	L = p->card[i]; K = p->card[(N+i-1)%N];
	if(i==0) i=N;
	for(k=0; k<K; ++k)
	  for(l=0; l<L; ++l) // e = l+k*L
	    p->msg[N+i-1][k+l*K] += factor*message[e++];
      }
    }
  }else if(p->type == FUNC_TYPE_TABLE){

    int N = p->num_variable;
    if(dim==1){
      i=0;
      while(p->var_id[i] != scope[0] && i<N) ++i;
      assert(i<N);
      for(k=0; k<p->card[i]; ++k)
	p->msg[i][k] += factor*message[k];
    }else{
      int msg_id;
      int i0=0, i1=0;
      while(p->var_id[i0] != scope[0] && i0<N) ++i0;
      while(p->var_id[i1] != scope[1] && i1<N) ++i1;
      if(i0<i1){
	msg_id = N + i0*N + i1;
	int K = (p->card[i0])*(p->card[i1]);
	for(k=0; k<K; ++k)
	  p->msg[msg_id][k] += factor * message[k];
	
      }else{ // i1 < i0
	msg_id = N + i1*N + i0;
	int K = p->card[i1]; 
	int L = p->card[i0];
	int e=0;
	for(k=0; k<K; ++k){
	  for(l=0; l<L; ++l){ // e = l+k*L
	    p->msg[msg_id][k+l*K] += factor*message[e];
	    ++e;
	  }
	}
      }
    }
      
  }else{
    printf("ERR: addMessage to be implemented (new type)\n");
    assert(p->type == FUNC_TYPE_CYCLE);
    assert(0);
  }

}


void buildTree(MapProblem* problem, int method_id, MapParameters* parameters)
{
  assert(method_id >= TREE_SP1 && method_id <= TREE_SP4);
  int i,j,M = problem->num_subproblem;
  int** sp_id = problem->sptree->sp_id;
  int* phi_id = problem->sptree->phi_id;
  SubProblemTree* t = problem->sptree;
  //int* nb_sp = problem->sptree->nb_sp;
  //int* nb_phi = problem->sptree->nb_phi;
  int* C = problem->sptree->candidates;
  int* color = problem->sptree->color;
  int* parent_sp = problem->sptree->parent_sp;
  int* parent_phi = problem->sptree->parent_phi;

  
  setIntVal(color, 0, M); // 0==white, 1==gray, 2==black
  setIntVal(parent_sp, -1, M);
  setIntVal(parent_phi, -1, M);

  int dfs,fixed_parent;
  if(method_id == TREE_SP1){
    dfs = 1; fixed_parent = 1;
  }else if(method_id == TREE_SP2){
    dfs = 1; fixed_parent = 0;
  }else if(method_id == TREE_SP3){
    dfs = 0; fixed_parent = 1;
  }else if(method_id == TREE_SP4){
    dfs = 0; fixed_parent = 0;
  }else{ assert(0); }

  int current, n, c0=0,c1=0;
  int root_subproblem;
  do{
    root_subproblem = rand()%M;
  }while(problem->sp[root_subproblem]->num_variable==1);

  C[c1++] = root_subproblem;
  color[root_subproblem] = 2;

  int edge_idx = problem->sptree->max_num_edges-1;

  while(c0!=c1){ // the stack/queue is not empty

    //printf("c1 = %d\n", c1);
    //assert(c1!=2);
    if(dfs) current = C[--c1]; // use C as stack
    else current = C[c0++];    // use C as queue

    // collect all neighbors for current subproblem
    if(parameters->bipartite_graph == GRAPH_DC){
      n = 0; 
      //int start = (problem->num_phi_of_sp[current] > problem->sp[current]->num_variable+1)?
      //	(problem->sp[current]->num_variable):0;
      for(i=0; i<problem->num_phi_of_sp[current]; ++i){
	int phi = problem->phi_of_sp[current][i];
	for(j=0; j<problem->num_sp_of_phi[phi]; ++j){
	  int sp1 = problem->sp_of_phi[phi][j];
	  //	  assert(sp1 < problem->num_subproblem);
	  if(sp1 != current){
	    if(n == t->nb_size){
	      t->nb_size *= 2;
	      t->nb_sp = (int*)realloc(t->nb_sp, sizeof(int)*(t->nb_size));
	      t->nb_phi = (int*)realloc(t->nb_phi, sizeof(int)*(t->nb_size));
	    }
	    t->nb_sp[n] = sp1; 
	    t->nb_phi[n++] = phi;
	  }
	}
      }
    }else{ // GRAPH_BETHE
      n = 0;
      if(problem->sp[current]->num_variable == 1){
	int phi = problem->sp[current]->var_id[0];
	for(j=0; j<problem->num_sp_of_phi[phi]; ++j){
	  int sp1 = problem->sp_of_phi[phi][j];
	  //	  assert(sp1 < problem->num_subproblem);
	  if(sp1 != current){
	    if(n == t->nb_size){
	      t->nb_size *= 2;
	      t->nb_sp = (int*)realloc(t->nb_sp, sizeof(int)*(t->nb_size));
	      t->nb_phi = (int*)realloc(t->nb_phi, sizeof(int)*(t->nb_size));
	    }
	    t->nb_sp[n] = sp1; 
	    t->nb_phi[n++] = phi;
	  }
	}
      }else{

	for(i=0; i<problem->sp[current]->num_variable; ++i){
	  int phi = problem->sp[current]->var_id[i];
	  int sp1 = problem->unary_subproblem_id[phi];
	  if(n == t->nb_size){
	    t->nb_size *= 2;
	    t->nb_sp = (int*)realloc(t->nb_sp, sizeof(int)*(t->nb_size));
	    t->nb_phi = (int*)realloc(t->nb_phi, sizeof(int)*(t->nb_size));
	  }
	  t->nb_sp[n] = sp1; 
	  t->nb_phi[n++] = phi;
	}
      }
    }
    /*
      printf("\n");
      for(i=0; i<n; ++i)
      printf("(sp %d, phi %d) ", nb_sp[i], nb_phi[i]);
      printf("\n");
    */

    if (n==0) break;

    permutateLists(t->nb_sp, t->nb_phi, n);
    //    promotePairwiseScopes(t->nb_sp, t->nb_phi, n, problem->num_variable);
    /*
      for(i=0; i<n; ++i)
      printf("(sp %d, phi %d) ", nb_sp[i], nb_phi[i]);
      printf("\n");
    */

    // add current to tree
    if(current != root_subproblem){
      sp_id[edge_idx][0] = current;
      if(fixed_parent){ 
	// choose neighbor to be the one thru which 'current' was discovered
	sp_id[edge_idx][1] = parent_sp[current];
	phi_id[edge_idx] = parent_phi[current];
      }else{ 	// randomly choose a neighor that is already in the tree
	i=0;
	while(color[t->nb_sp[i]] != 2 && i<n) i++;
	//	assert(i<n);
	sp_id[edge_idx][1] = t->nb_sp[i];
	phi_id[edge_idx] = t->nb_phi[i];
      }
      --edge_idx;
      color[current]=2;
    }
    //printf("n = %d\n", n);
    
    // put all white neighbors into candidate set
    for(i=0; i<n; ++i){
      //printf("try to add %d\n", nb_sp[i]);
      if(color[t->nb_sp[i]] == 0){
	//printf("add %d\n", nb_sp[i]);
	color[t->nb_sp[i]] = 1;
	parent_sp[t->nb_sp[i]] = current;
	parent_phi[t->nb_sp[i]] = t->nb_phi[i];
	C[c1++] = t->nb_sp[i];
	//	assert(c1 <= problem->num_subproblem);
      }
    }
  }
  // assert(edge_idx==-1);

  problem->sptree->start_idx = edge_idx+1;


  if(parameters->allocation_method != ALLOCATION_ROOT){
    double* subtree_weight = NULL;
    if(parameters->allocation_method == ALLOCATION_UNIFORM){

      subtree_weight = problem->sptree->subtree_weight;
      setDblVal(subtree_weight, 1.0, problem->num_subproblem);

    }else if(parameters->allocation_method == ALLOCATION_LEAF){

      subtree_weight = problem->sptree->subtree_weight;
      setDblVal(subtree_weight, 1.0, problem->num_subproblem);
      for(i=problem->sptree->start_idx; i<problem->sptree->max_num_edges; i++){
	int p = sp_id[i][1];
	subtree_weight[p] = 0;
      }

    }else if(parameters->allocation_method == ALLOCATION_NEXT_LEAF 
	     && problem->next_sptree != NULL){

      SubProblemTree *tmp = problem->sptree;
      problem->sptree = problem->next_sptree;
      problem->next_sptree = tmp;
      subtree_weight = problem->sptree->subtree_weight;
      assert(sp_id == problem->next_sptree->sp_id);
      setDblVal(subtree_weight, 0.0, problem->num_subproblem);
      for(i=problem->next_sptree->start_idx; 
	  i<problem->next_sptree->max_num_edges; i++){
	int c = sp_id[i][0];
	subtree_weight[c] = 1.0;
      }
      for(i=problem->next_sptree->start_idx; 
	  i<problem->next_sptree->max_num_edges; i++){
	int p = sp_id[i][1];
	subtree_weight[p] = 0.0;
      }
      sp_id = problem->sptree->sp_id;

    }else if(parameters->allocation_method == ALLOCATION_DEGREE){
      subtree_weight = problem->sptree->subtree_weight;
      copyDblVec(problem->subproblem_degree, subtree_weight, problem->num_subproblem);
    }else if(parameters->allocation_method == ALLOCATION_UNARY){
      assert(parameters->bipartite_graph == GRAPH_BETHE);
      subtree_weight = problem->sptree->subtree_weight;
      copyDblVec(problem->is_unary_subproblem, subtree_weight, problem->num_subproblem);
    }

    /*
    if(subtree_weight != NULL){
      double total=0.0;
      for(i=problem->sptree->start_idx; i<problem->sptree->max_num_edges; i++){
	int c = sp_id[i][0];
	int p = sp_id[i][1];
	assert(subtree_weight[c] >= 0);
	assert(subtree_weight[p] >= 0);
	if(parameters->allocation_method == ALLOCATION_UNARY){
	  if(problem->sp[c]->num_variable==1)
	    assert(subtree_weight[c] == 1.0);
	  else
	    assert(subtree_weight[c] == 0.0);
	}
	total += subtree_weight[c];
	if (i== problem->sptree->max_num_edges-1)
	  total += subtree_weight[p];
      }
      assert(total > 0);
    }
    */
    //////////////////////////////////////

    if(subtree_weight != NULL){
      assert(sp_id == problem->sptree->sp_id);
      assert(subtree_weight == problem->sptree->subtree_weight);
      for(i=problem->sptree->start_idx; i<problem->sptree->max_num_edges; i++){
	int c = sp_id[i][0];
	int p = sp_id[i][1];
	subtree_weight[p] += subtree_weight[c];
      }
    }
  }
}

void scaleVec(double* vec, int n, double factor)
{
  int i;
  for(i=0; i<n; ++i)
    vec[i] *= factor;
}


void calibrateTree(MapProblem* problem, MapParameters* parameters)
{

  /*
  double old_dual = evaluateDual(problem, parameters);
  int* solution = (int*)malloc(sizeof(int)*(problem->num_variable));
  getRandomState(problem, solution);
  */

  SubProblemTree* t = problem->sptree;

  // problem->msg_cnt += (t->max_num_edges - t->start_idx);

  Cycle* cy = problem->cy_solver;
  int i,transpose;
  for(i=t->start_idx; i<t->max_num_edges; ++i){ // forward pass
    SubProblem* child = problem->sp[t->sp_id[i][0]];
    SubProblem* parent = problem->sp[t->sp_id[i][1]];
    int sep_size = t->phi_id[i]<problem->num_variable?1:2;
    if(sep_size==1)
      problem->msg_cnt += (child->num_variable - 1);
    else
      problem->msg_cnt += (child->num_variable - 1) *(child->num_variable - 1);

    /*
    double c0,p0,c1,p1;
    c0 = evaluateSubProblemEnergy(child, solution, problem->initialized);
    p0 = evaluateSubProblemEnergy(parent, solution, problem->initialized);
    */

    int* sep_scope = problem->scope_of_phi[t->phi_id[i]];
    double* msg = 
      parent->msg[getMessageHandle(parent, sep_size, sep_scope, &transpose)];

    extractMinMarginal(parameters,child, cy, sep_size, sep_scope+transpose, 1.0, 1.0, msg);

    /*
    c1 = evaluateSubProblemEnergy(child, solution, problem->initialized);
    p1 = evaluateSubProblemEnergy(parent, solution, problem->initialized);
    assert(fabs(c1 + p1 - c0 - p0)<1e-6);
    */
  }

  if(parameters->allocation_method==ALLOCATION_ROOT) return;

  copyDblVec(t->subtree_weight,t->subtree_weight_dynamic,problem->num_subproblem);

  for(i=t->max_num_edges-1; i>=t->start_idx; --i){ // backward pass

    if(t->subtree_weight[t->sp_id[i][0]] > 0){

      SubProblem* child = problem->sp[t->sp_id[i][0]];
      SubProblem* parent = problem->sp[t->sp_id[i][1]];
      int sep_size = t->phi_id[i]<problem->num_variable?1:2;

      if(sep_size==1)
	problem->msg_cnt += (parent->num_variable - 1);
      else
	problem->msg_cnt += (parent->num_variable - 1)*(parent->num_variable - 1);

      /*
      double c0,p0,c1,p1;
      c0 = evaluateSubProblemEnergy(child, solution, problem->initialized);
      p0 = evaluateSubProblemEnergy(parent, solution, problem->initialized);
      */


      int* sep_scope = problem->scope_of_phi[t->phi_id[i]];

      double factor = t->subtree_weight[t->sp_id[i][0]] / 
	t->subtree_weight_dynamic[t->sp_id[i][1]];

      assert(!isnan(factor));
      assert(t->subtree_weight[t->sp_id[i][0]] >= 0);
      assert(t->subtree_weight_dynamic[t->sp_id[i][1]] > 0);
      assert(factor >= 0);

      double* msg = 
	child->msg[getMessageHandle(child, sep_size, sep_scope, &transpose)];

      extractMinMarginal(parameters,parent,cy,sep_size,sep_scope+transpose,factor,factor,msg);


      t->subtree_weight_dynamic[t->sp_id[i][1]] -= 
	t->subtree_weight[t->sp_id[i][0]];

      /*
      c1 = evaluateSubProblemEnergy(child, solution, problem->initialized);
      p1 = evaluateSubProblemEnergy(parent, solution, problem->initialized);
      assert(fabs(c1 + p1 - c0 - p0)<1e-6);

      new_dual2 = evaluateDual(problem, parameters);
      assert(fabs(new_dual2 - new_dual1)/fabs(new_dual2)<1e-6);
      */


    }
  }

  /*
  free(solution);
  new_dual2 = evaluateDual(problem, parameters);
  assert(fabs(new_dual2 - new_dual1)/fabs(new_dual2)<1e-6);
  */
}


void starAverage(MapProblem* problem, MapParameters* parameters)
{
  //  clock_t tic = clock();
  int i,j;
  int* phi = (int*)malloc(sizeof(int)*(problem->num_phi));

  randPermutation(phi, problem->num_phi);

  for(i=0; i<problem->num_phi; ++i){

    /*
    for(int idx=0; idx<problem->num_subproblem; idx++)
      problem->sp[idx]->min_energy_valid = 0;
    double obj0 =  evaluateDual(problem, parameters);
    */
    
    if(problem->msg_cnt >= parameters->stc_inner_iterations) break;

    int id = phi[i];
    int num_sp = problem->num_sp_of_phi[id];
    if(num_sp <= 1) continue;
    double denominator = (double)num_sp;
    int* sp_id = problem->sp_of_phi[id];

    for(j=0; j<num_sp; ++j){
      if(id < problem->num_variable)
	problem->msg_cnt += (problem->sp[sp_id[j]]->num_variable - 1);
      else
	problem->msg_cnt += (problem->sp[sp_id[j]]->num_variable - 1) * 
	  (problem->sp[sp_id[j]]->num_variable - 1);
    }
    if(denominator <= 1.0) continue;

    int* scope = problem->scope_of_phi[id];
    // problem->msg_cnt += (int)denominator;
    int K = id<problem->num_variable?problem->card[id]:
      problem->card[scope[0]] * problem->card[scope[1]];
    double* average = (double*)malloc(sizeof(double)*K);
    setDblVal(average, 0.0, K);
    for(j=0; j<num_sp; ++j){ // collect min-marginals

      extractMinMarginal(parameters,problem->sp[sp_id[j]], problem->cy_solver,
			 id<problem->num_variable?1:2,
			 scope, 1.0, 1.0, average);     

      /*
      problem->sp[sp_id[j]]->min_energy_valid = 0;
      assert(fabs(getSubProblemMin(problem->sp[sp_id[j]], problem->cy_solver, parameters))<DEBUG_THRESHOLD);
      */

    }
    scaleVec(average, K, 1.0/denominator); // take average

    int state_average;
    double min_energy = minimize(average, K, &state_average);
    //min_energy /= (double)num_sp;



    for(j=0; j<num_sp; ++j){ // add averaged min-marginals
      addMessage(problem->sp[sp_id[j]], 
		 id<problem->num_variable?1:2,
		 scope, 1.0, average);

      problem->sp[sp_id[j]]->min_energy_valid=1;
      problem->sp[sp_id[j]]->min_energy = min_energy;

    }
    free(average);

    /*
    for(int idx=0; idx<problem->num_subproblem; idx++)
      problem->sp[idx]->min_energy_valid = 0;
    if(!(obj0 <= evaluateDual(problem, parameters) + DEBUG_THRESHOLD)){
      printf("id = %d\n", id);
      printf("obj0 = %f\n", obj0);
      printf("new obj = %f\n", evaluateDual(problem, parameters));
      assert(0);
    }
    */

  }
  free(phi);
  //  printf("star Average: %.2f secs\n", ((double)clock()-ic)/CLOCKS_PER_SEC);
}

void calibrateBethePairs(MapProblem* problem, MapParameters* parameters)
{
  assert(parameters->allocation_method == ALLOCATION_UNIFORM || 
	 parameters->allocation_method == ALLOCATION_UNARY || 
	 parameters->allocation_method == ALLOCATION_DEGREE);
  int* sp_id = (int*)malloc(sizeof(int)*(problem->num_nonunary_subproblem));
  double factor;
  randPermutation(sp_id, problem->num_nonunary_subproblem);
  for(int i=0; i<problem->num_nonunary_subproblem; ++i){
    SubProblem *p = problem->sp[sp_id[i]];
    assert(p->num_variable>1);
    for(int j=0; j<p->num_variable; ++j){
      problem->msg_cnt++;
      int var = p->var_id[j];
      SubProblem *u = problem->sp[problem->unary_subproblem_id[var]];
      int msg_id = j; // findInVec(var, p->var_id, p->num_variable);
      // msg from i to f
      extractMinMarginal(parameters, u, problem->cy_solver,
			 1, u->var_id, 1.0, 1.0, p->msg[msg_id]);
      p->min_energy_valid = 0;
      if(parameters->allocation_method == ALLOCATION_UNIFORM){
	factor = 0.5;
      }else if(parameters->allocation_method == ALLOCATION_UNARY){
	factor = 1.0;
      }else{ // ALLOCATION_DEGREE
	double degree0 = problem->subproblem_degree[sp_id[i]];
	double degree1 = 
	  problem->subproblem_degree[problem->unary_subproblem_id[var]];
	factor = degree1/(degree0+degree1);
      }
      // msg from f to i      
      extractMinMarginal(parameters, p, problem->cy_solver,
			 1, u->var_id, factor, factor, u->msg[0]);
      u->min_energy_valid = 0;
    }
  }
  free(sp_id);
}

void calibrateBetheStars(MapProblem* problem, MapParameters* parameters)
{
  assert(parameters->allocation_method == ALLOCATION_UNIFORM || 
	 parameters->allocation_method == ALLOCATION_UNARY || 
	 parameters->allocation_method == ALLOCATION_DEGREE);
  int* sp_id = (int*)malloc(sizeof(int)*(problem->num_nonunary_subproblem));
  double factor;
  randPermutation(sp_id, problem->num_nonunary_subproblem);
  for(int i=0; i<problem->num_nonunary_subproblem; ++i){
    SubProblem *p = problem->sp[sp_id[i]];
    assert(p->num_variable>1);
    problem->msg_cnt += p->num_variable;

    double sum_sp_weight=0.0;
    if(parameters->allocation_method == ALLOCATION_DEGREE)    
      sum_sp_weight += problem->subproblem_degree[sp_id[i]];
    else if(parameters->allocation_method == ALLOCATION_UNIFORM)
      sum_sp_weight += 1.0;
    std::vector<double> sp_weight(p->num_variable,0);

    // add-in messages from unary subproblems
    for(int j=0; j<p->num_variable; ++j){

      int var = p->var_id[j];

      /*
      assert(problem->unary_subproblem_id[var] >= 
	     problem->num_nonunary_subproblem &&
	     problem->unary_subproblem_id[var] < 
	     problem->num_subproblem);
      */
      
      if(parameters->allocation_method == ALLOCATION_DEGREE){
	sp_weight[j] = 
	  problem->subproblem_degree[problem->unary_subproblem_id[var]];
      }else{
	sp_weight[j] = 1.0;
      }
      sum_sp_weight += sp_weight[j];


      SubProblem *u = problem->sp[problem->unary_subproblem_id[var]];
      int msg_id = j; // findInVec(var, p->var_id, p->num_variable);
      extractMinMarginal(parameters, u, problem->cy_solver,
			 1, u->var_id, 1.0, 1.0, p->msg[msg_id]);
      p->min_energy_valid = 0;
    }

    // pass messages to unary subproblems
    for(int j=0; j<p->num_variable; ++j){
      int var = p->var_id[j];
      SubProblem *u = problem->sp[problem->unary_subproblem_id[var]];
      factor = sp_weight[j]/sum_sp_weight;
      sum_sp_weight -= sp_weight[j];

      extractMinMarginal(parameters, p, problem->cy_solver,
			 1, u->var_id, factor, factor, u->msg[0]);
      u->min_energy_valid = 0;
    }
  }
  free(sp_id);
}

void calibrateBetheStarPlus(MapProblem* problem, MapParameters* parameters)
{
  assert(parameters->allocation_method == ALLOCATION_UNIFORM || 
	 parameters->allocation_method == ALLOCATION_UNARY || 
	 parameters->allocation_method == ALLOCATION_DEGREE);
  int* variable_id = (int*)malloc(sizeof(int)*(problem->num_variable));
  double factor;
  randPermutation(variable_id, problem->num_variable);
  for(int i=0; i<problem->num_variable; ++i){
    int var = variable_id[i];
    if(problem->num_sp_of_phi[var]==1) continue;

    SubProblem* u0 = problem->sp[problem->unary_subproblem_id[var]];

    // for each non-unary subproblem, collect messages from {-u0}
    // and pass message to u0
    for(int j=0; j<problem->num_sp_of_phi[var]; ++j){
      SubProblem *p = problem->sp[problem->sp_of_phi[var][j]];
      if (p->num_variable>1){

	problem->msg_cnt += p->num_variable;

	// from -i to f
	for(int k=0; k<p->num_variable; ++k){
	  int var0 = p->var_id[k];
	  if(var0 != var){
	    SubProblem *u = problem->sp[problem->unary_subproblem_id[var0]];
	    extractMinMarginal(parameters, u, problem->cy_solver,
			       1, u->var_id, 1.0, 1.0, u->msg[0]);
	    u->min_energy_valid = 0;
	  }
	}
	// from f to i
	
	extractMinMarginal(parameters, p, problem->cy_solver,
			   1, problem->scope_of_phi[var], 
			   1.0, 1.0, u0->msg[0]);
	u0->min_energy_valid = 0;
      }
    }

    double total_weight = 0.0;
    if(parameters->allocation_method == ALLOCATION_DEGREE){
      total_weight += problem->subproblem_degree[problem->unary_subproblem_id[var]];
    }else{
      total_weight += 1.0;
    }
    std::vector<double> weight_f(problem->num_sp_of_phi[var], 0);
    for(int j=0; j<problem->num_sp_of_phi[var]; ++j){
      SubProblem *p = problem->sp[problem->sp_of_phi[var][j]];
      if(parameters->allocation_method == ALLOCATION_UNIFORM){

	weight_f[j] = p->num_variable;

      }else if(parameters->allocation_method == ALLOCATION_UNARY){

	weight_f[j] = p->num_variable-1;

      }else{ // parameters->allocation_method == ALLOCATION_DEGREE

	weight_f[j] += problem->subproblem_degree[problem->sp_of_phi[var][j]];
	for(int k=0; k<p->num_variable; ++k){
	  int var0 = p->var_id[k];
	  if(var0 != var)
	    weight_f[j] += problem->subproblem_degree[problem->unary_subproblem_id[var0]];
	}
      }
      total_weight += weight_f[j];
    }

    // for each non-unary subproblem, pass message from var, and pass message
    // to all -var
    for(int j=0; j<problem->num_sp_of_phi[var]; ++j){
      SubProblem *p = problem->sp[problem->sp_of_phi[var][j]];
      if (p->num_variable>1){
	// msg from i to f
	factor = weight_f[j] / total_weight;
	total_weight -= weight_f[j];

	int msg_id = findInVec(var, p->var_id, p->num_variable);
	extractMinMarginal(parameters, u0, problem->cy_solver,
			   1, problem->scope_of_phi[var], factor,
			   factor, p->msg[msg_id]);
	p->min_energy_valid = 0;
	
	double sum_sp_weight=0.0;
	if(parameters->allocation_method == ALLOCATION_DEGREE)    
	  sum_sp_weight += problem->subproblem_degree[problem->sp_of_phi[var][j]];
	else if(parameters->allocation_method == ALLOCATION_UNIFORM)
	  sum_sp_weight += 1.0;
	std::vector<double> sp_weight(p->num_variable,0);

	for(int k=0; k<p->num_variable; ++k){
	  int var0 = p->var_id[k];
	  if(var0 != var){
	    if(parameters->allocation_method == ALLOCATION_DEGREE){
	      sp_weight[k] =
		problem->subproblem_degree[problem->unary_subproblem_id[var0]];
	    }else{
	      sp_weight[k] = 1.0;
	    }
	    sum_sp_weight += sp_weight[k];
	  }
	}	

	// msg from f to -i
	for(int k=0; k<p->num_variable; ++k){
	  int var0 = p->var_id[k];
	  if(var0 != var){
	    SubProblem *u = problem->sp[problem->unary_subproblem_id[var0]];

	    factor = sp_weight[k]/sum_sp_weight;
	    sum_sp_weight -= sp_weight[k];

	    extractMinMarginal(parameters, p, problem->cy_solver,
			       1, problem->scope_of_phi[var0], factor,
			       factor, u->msg[0]);
	    u->min_energy_valid = 0;
	  }
	}
      }
    }
  }
  free(variable_id);
}

void minHeapPushUp(int i, double* val, int* key)
{
  int p = (i-1)/2;
  if(val[p] > val[i]){
    double td = val[i];
    val[i] = val[p];
    val[p] = td;
    int ti = key[i];
    key[i] = key[p]; 
    key[p] = ti;
    minHeapPushUp(p, val, key);
  }
}

void minHeapPushDown(int i, double* val, int* key, int n)
{
  int l = i*2+1;
  int r = i*2+2;
  int s = -1;
  if(l>=n) return;
  if(val[i] > val[l] && (r>=n || val[r] > val[l]))
    s = l;
  else if(r<n && val[r] < val[i])
    s = r;
  if(s>=0){
    double td = val[i];
    val[i] = val[s];
    val[s] = td;
    int ti = key[i];
    key[i] = key[s]; 
    key[s] = ti;
    minHeapPushDown(s, val, key, n);
  }
}


void  getTopKScore(double *score, int N, int* K, int* indices, double threshold)
{


  int i,n = (*K);  // max heap size
  *K = 0;        // current heap size;
  for(i=0; i<N; ++i){ 
    // visit each score, try adding to min heap (key=indices, val=score)
    if(score[i] > threshold){
      if(*K < n){
	indices[(*K)] = i;
	score[(*K)] = score[i];
	++(*K);
	minHeapPushUp((*K)-1, score, indices); 
      }else if(score[i] > score[0]){
	indices[0] = i;
	score[0] = score[i];
	minHeapPushDown(0, score, indices, n);
      }
    }
  }
  if((*K) == 1 && score[0] <= threshold) (*K) = 0;

}

void my_swap(double *val, int* aux0,
	     int s, int e){
  if(s==e) return;
  double t0 = val[s];
  val[s] = val[e];
  val[e] = t0;
  int t = aux0[s];
  aux0[s] = aux0[e];
  aux0[e] = t;
}

// sort from high to low according 'val', move 'aux' around accordingly
void my_sort(double *val, int* aux0,
	     int s, int e)
{
  if(s>=e) return;
  int i,i0=s-1;
  for(i=s; i<e; ++i)
    if(val[i] > val[e])	
      my_swap(val,aux0,i,++i0);
  my_swap(val,aux0,e,i0+1);
  my_sort(val,aux0,s,i0);
  my_sort(val,aux0,i0+2,e);
}


void buildVariableNeighborLists(MapProblem* problem, MapParameters* parameters)
{
  int i,N = problem->num_variable;
  problem->var_degree = (int*)malloc(sizeof(int)*N);
  problem->var_nb = (int**)malloc(sizeof(int*)*N);
  problem->var_nb_edge = (int**)malloc(sizeof(int*)*N);
  //  problem->max_cardinality = 0;
  for(i=0; i<N; ++i){
    //    problem->max_cardinality = MAX(problem->max_cardinality, problem->card[i]);
    problem->var_degree[i] = 0;
    problem->var_nb[i] = NULL;
    problem->var_nb_edge[i] = NULL;
  }
  int edge_cnt = 0;
  for(i=0; i<problem->num_subproblem; ++i){
    SubProblem *p = problem->sp[i];
    if(p->num_variable == 2){
      int v0 = p->var_id[0];
      int v1 = p->var_id[1];
      append2TwoIntVec(v1, i, &(problem->var_nb[v0]), 
		       &(problem->var_nb_edge[v0]), &(problem->var_degree[v0]));
      append2TwoIntVec(v0, i, &(problem->var_nb[v1]), 
		       &(problem->var_nb_edge[v1]), &(problem->var_degree[v1]));
      edge_cnt++;
    }
  }
  problem->num_edge_factor = edge_cnt;

}

int isN4Grid(MapProblem* problem, int width)
{
  int i, N = problem->num_variable; 
  if ((N/width)*width != N) return 0;
  int H = N/width;
  int W = width;
  if(problem->num_subproblem < (H-1)*W + H*(W-1)) return 0;
  int edge_cnt=0;
  for(i=0; i<problem->num_subproblem; ++i){
    SubProblem *p = problem->sp[i];
    if(p->num_variable == 2){
      edge_cnt++;
      int v0 = p->var_id[0];
      int v1 = p->var_id[1];
      if(!(abs(v0-v1)==1 || abs(v0-v1)==W)) return 0;
    }
  }
  problem->num_edge_factor = edge_cnt;
  return edge_cnt == (H-1)*W + H*(W-1);
}

int isN8Grid(MapProblem* problem, int width)
{
  int i, N = problem->num_variable; 
  if ((N/width)*width != N) return 0;
  int H = N/width;
  int W = width;
  if(problem->num_subproblem < (H-1)*W + H*(W-1) + 2*(H-1)*(W-1)) return 0;
  int edge_cnt=0;
  for(i=0; i<problem->num_subproblem; ++i){
    SubProblem *p = problem->sp[i];
    if(p->num_variable == 2){
      edge_cnt++;
      int v0 = p->var_id[0];
      int v1 = p->var_id[1];
      if(!(abs(v0-v1)==1 || abs(abs(v0-v1)-W)<=1)) return 0;
    }
  }
  problem->num_edge_factor = edge_cnt;
  return edge_cnt == (H-1)*W + H*(W-1) + 2*(H-1)*(W-1);
}

int getGridEdgeIndex(int nb, int v0, int v1, int W, int H)
{
  if(v0>v1){int tmp = v0; v0 = v1; v1 = tmp;}

  if(nb==4){
    if(v1==v0+1)
      return (v0/W)*(W-1) + v0%W;
    else
      return H*(W-1) + v0;
  }else{
    if(v1==v0+1) 
      return (v0/W)*(W-1) + v0%W;
    else if(v1==v0+W)
      return H*(W-1) + v0;
    else if(v1==v0+W+1)
      return H*(W-1) + W*(H-1) + (v0/W)*(W-1) + v0%W;
    else 
      return H*(W-1) + W*(H-1) + (W-1)*(H-1) + ((v0-1)/W)*(W-1) + (v0-1)%W;
  }
  return -1;
}

void checkGridStructure(MapProblem* problem, MapParameters* parameters)
{
/*
  clock_t tic = clock();
*/
  int i=0; 
  problem->is_grid=0;
  problem->grid_width=0;
  problem->grid_height=0;
  for(i=0; i<problem->num_subproblem; ++i){
    SubProblem *p = problem->sp[i];
    if(p->num_variable == 2){
      int v0 = p->var_id[0];
      int v1 = p->var_id[1];
      if(abs(v0-v1)>1){
	if(isN4Grid(problem, abs(v0-v1))){
	  problem->grid_width = abs(v0-v1);
	  problem->is_grid = 4;
	  break;
	}
	if(isN8Grid(problem, abs(v0-v1))){
	  problem->grid_width = abs(v0-v1);
	  problem->is_grid = 8;
	  break;
	}
	if(isN8Grid(problem, abs(v0-v1)+1)){
	  problem->grid_width = abs(v0-v1)+1;
	  problem->is_grid = 8;
	  break;
	}
	if(isN8Grid(problem, abs(v0-v1)-1)){
	  problem->grid_width = abs(v0-v1)-1;
	  problem->is_grid = 8;
	  break;
	}
	break;
      }
    }
  }
  if(problem->is_grid > 0){
    problem->grid_height = problem->num_variable/problem->grid_width;
    problem->sp2phi = (int*)malloc(sizeof(int)* (problem->num_subproblem));
    setIntVal(problem->sp2phi, -1, problem->num_subproblem);
    problem->grid_edge2sp = (int*)malloc(sizeof(int)* (problem->num_edge_factor));
    for(i=0; i<problem->num_subproblem; ++i){
      SubProblem *p = problem->sp[i];
      if(p->num_variable == 2){
	int edge_id = getGridEdgeIndex(problem->is_grid, 
				       p->var_id[0], p->var_id[1],
				       problem->grid_width, 
				       problem->grid_height);
	assert(edge_id >=0 && edge_id < problem->num_edge_factor);
	problem->grid_edge2sp[edge_id] = i;
      }
    }
  }
/*
  printf("is_grid = %d, width=%d, height = %d, %.2f seconds\n", 
	 problem->is_grid, problem->grid_width, problem->grid_height,
	 ((double)clock()-tic)/CLOCKS_PER_SEC); 
*/

}

void  getCyclesInN4Grid(MapProblem* problem, MapParameters* parameters)
{
/*
  clock_t tic = clock();
*/
  int i,j,k;

  int W = problem->grid_width;
  int H = problem->grid_height;
  problem->cycle_dict_size = (W-1)*(H-1);
  int M = problem->cycle_dict_size;
  assert(M>0);
  problem->cycle_size = (int*)malloc(sizeof(int)*M);
  problem->cycle_nodes = (int**)malloc(sizeof(int*)*M);
  problem->cycle_edges = (int**)malloc(sizeof(int*)*M);

  for(i=0; i<W-1; ++i){
    for(j=0; j<H-1; ++j){
      int v[] = {j*W + i, j*W + i+1, (j+1)*W + i+1, (j+1)*W + i, j*W + i};
      int c = j*(W-1) + i;
      assert(c>=0 && c<M);
      problem->cycle_size[c] = 4;
      problem->cycle_nodes[c] = (int*)malloc(sizeof(int)*5);
      problem->cycle_edges[c] = (int*)malloc(sizeof(int)*4);
      for(k=0; k<4; ++k){
	problem->cycle_nodes[c][k] = v[k];
	int edge_id = getGridEdgeIndex(4, v[k], v[k+1], W, H);
	assert(edge_id >= 0 && edge_id < problem->num_edge_factor);
	problem->cycle_edges[c][k] = problem->grid_edge2sp[edge_id];
      }
      problem->cycle_nodes[c][4] = v[4];

    }
  }
/*
  printf("find cycles in n-4 grid, %.2f seconds\n",
	 ((double)clock()-tic)/CLOCKS_PER_SEC);
*/
}

void getCyclesInN8Grid(MapProblem* problem, MapParameters* parameters)
{
/*
  clock_t tic = clock();
*/
  int i,j,k,l;

  problem->cycle_dict_size = 4*(problem->grid_width-1)*(problem->grid_height-1);
  int M = problem->cycle_dict_size;
  problem->cycle_size = (int*)malloc(sizeof(int)*M);
  problem->cycle_nodes = (int**)malloc(sizeof(int*)*M);
  problem->cycle_edges = (int**)malloc(sizeof(int*)*M);
  int W = problem->grid_width;
  int H = problem->grid_height;
  for(i=0; i<W-1; ++i){
    for(j=0; j<H-1; ++j){
      int v[] = {j*W + i, j*W + i+1, (j+1)*W + i+1, 
		 (j+1)*W + i, j*W + i, j*W + i+1, (j+1)*W + i+1};
      for(k=0; k<4; ++k){
	int c = 4* (j*(W-1) + i) +k;
	assert(c>=0 && c<M);
	problem->cycle_size[c] = 3;
	problem->cycle_nodes[c] = (int*)malloc(sizeof(int)*4);
	problem->cycle_edges[c] = (int*)malloc(sizeof(int)*3);
	for(l=0; l<3; l++){
	  problem->cycle_nodes[c][l] = v[k+l];
	}
	problem->cycle_nodes[c][3] = v[k];
	for(l=0; l<3; l++){
	  int edge_id = getGridEdgeIndex(8, problem->cycle_nodes[c][l],
					 problem->cycle_nodes[c][l+1], W, H);
	  assert(edge_id >= 0 && edge_id < problem->num_edge_factor);
	  problem->cycle_edges[c][l] = problem->grid_edge2sp[edge_id];
	}
      }
    }
  }
/*
  printf("find cycles in n-8 grid, %.2f seconds\n",
	 ((double)clock()-tic)/CLOCKS_PER_SEC);
*/
}

void  getLargeCycles(MapProblem* problem,
		     MapParameters* parameters)
{
  int cycle_dict_size0 = problem->cycle_dict_size;
  int i;
  int N= problem->num_variable;
  int root;
  int l0,l1,r0,r1;
  int* l_nodes = (int*)malloc(sizeof(int) * (N+5)); 
  int* l_edges = (int*)malloc(sizeof(int) * (N+5)); 
  int* prev = (int*)malloc(sizeof(int) * N); 
  int* depth = (int*)malloc(sizeof(int) * N); 
  int* prev_edge = (int*)malloc(sizeof(int) * N); 
  int* color  = (int*)malloc(sizeof(int) * N); //0:white, 1:grey, 2:black
  int* C = (int*)malloc(sizeof(int) * N);
  int current, c0, c1;
  setIntVal(color, 0, N);  
  setIntVal(prev, -1, N); 
  setIntVal(prev_edge, -1, N); 
  setIntVal(depth, -1, N);  
  int cnt=0;

  while(cnt<N){
    do{
      root = rand()%N;
    }while(color[root] > 0);
    c0=0; 
    c1=0;
    C[c1++] = root;
    color[root] = 1;
    prev[root] = -2;
    prev_edge[root] = -2;
    depth[root] = 0;
    while(c0!=c1){
      current = C[c0++];    // use C as queue
      cnt++;
      color[current] = 2;
      for(i=0; i<problem->var_degree[current]; ++i){
	int nb = problem->var_nb[current][i];
	int nb_edge = problem->var_nb_edge[current][i];
	
	if(color[nb]==0){

	  C[c1++] = nb;	  
	  color[nb] = 1;
	  prev[nb] = current;
	  prev_edge[nb] = nb_edge;
	  depth[nb] = depth[current]+1;

	}else if(color[nb] == 2 && nb != prev[current]){

	  int t0 = current;
	  int t1 = nb;
	  l0=(N+5)/2;
	  r0=l0+1;
	  l1=(N+5)/2;
	  r1=l1+1;
	  l_nodes[r0++] = t0;
	  l_nodes[l0--] = t1;
	  l_edges[r1++] = nb_edge;
	  while(depth[t0] > depth[t1]){
	    assert(t0>=0);
	    l_nodes[r0++] = prev[t0];
	    l_edges[r1++] = prev_edge[t0];
	    t0 = prev[t0];
	  }
	  while(depth[t1] > depth[t0]){
	    assert(t1>=0);
	    l_nodes[l0--] = prev[t1];
	    l_edges[l1--] = prev_edge[t1];
	    t1 = prev[t1];
	  }
	  while(t0!=t1){
	    assert(t0>=0);
	    assert(t1>=0);
	    l_nodes[r0++] = prev[t0];
	    l_edges[r1++] = prev_edge[t0];
	    t0 = prev[t0];
	    l_nodes[l0--] = prev[t1];
	    l_edges[l1--] = prev_edge[t1];
	    t1 = prev[t1];
	  }
	  assert(l_nodes[l0+1] == l_nodes[r0-1]);
	  assert(r0-l0-1 == r1-l1);
	  int c = problem->cycle_dict_size;
	  problem->cycle_dict_size++;
	  problem->cycle_size = (int*)realloc(problem->cycle_size,
					      sizeof(int)*(c+1));
	  problem->cycle_nodes = (int**)realloc(problem->cycle_nodes,
						sizeof(int*)*(c+1));
	  problem->cycle_edges = (int**)realloc(problem->cycle_edges,
						sizeof(int*)*(c+1));
	  problem->cycle_size[c] = r0-l0-2;
	  problem->cycle_nodes[c] = (int*)malloc(sizeof(int)*(r0-l0-1));
	  problem->cycle_edges[c] = (int*)malloc(sizeof(int)*(r1-l1));
	  for(int idx=l0+1; idx<r0; ++idx)
	    problem->cycle_nodes[c][idx-l0-1] = l_nodes[idx];
	  for(int idx=l1+1; idx<r1; ++idx)
	    problem->cycle_edges[c][idx-l1-1] = l_edges[idx];
	  problem->max_cycle_length = MAX(problem->cycle_size[c],
					  problem->max_cycle_length);
	}
      }
    }
  }
  for(i=0; i<N; ++i)
    assert(color[i] == 2);
  printf("found %d large cycles\n", problem->cycle_dict_size - cycle_dict_size0);
  free(l_nodes);
  free(l_edges);
  free(prev);
  free(prev_edge);
  free(depth);
  free(color);
  free(C);
}

void  getAllTriangle(MapProblem* problem, int** adj_mat, 
		     MapParameters* parameters)
{
/*
  clock_t tic = clock();
*/
  int cycle_dict_size0 = problem->cycle_dict_size;
  int i, jid, kid, k, N = problem->num_variable;

  for(i=0; i<N; i++){ //
    //    for(i=0; i<1; i++){ // 
    for(jid = 0; jid < problem->var_degree[i]; jid++){
      int j = problem->var_nb[i][jid];
      if(j > i){
	for(kid = 0; kid < problem->var_degree[i]; kid++){
	  k = problem->var_nb[i][kid];
	  if(k > j && adj_mat[k][j]>=0){

	    int c = problem->cycle_dict_size;
	    problem->cycle_dict_size++;
	    problem->cycle_size = (int*)realloc(problem->cycle_size,
						sizeof(int)*(c+1));
	    problem->cycle_nodes = (int**)realloc(problem->cycle_nodes,
						  sizeof(int*)*(c+1));
	    problem->cycle_edges = (int**)realloc(problem->cycle_edges,
						  sizeof(int*)*(c+1));
	    problem->cycle_size[c] = 3;
	    problem->cycle_nodes[c] = (int*)malloc(sizeof(int)*4);
	    problem->cycle_edges[c] = (int*)malloc(sizeof(int)*3);
	    problem->cycle_nodes[c][0] = i;
	    problem->cycle_nodes[c][1] = j;
	    problem->cycle_nodes[c][2] = k;
	    problem->cycle_nodes[c][3] = i;
	    problem->cycle_edges[c][0] = adj_mat[i][j];
	    problem->cycle_edges[c][1] = adj_mat[j][k];
	    problem->cycle_edges[c][2] = adj_mat[k][i];
	  }
	}
      }
    }
  }
  if(problem->cycle_dict_size - cycle_dict_size0 > 0)
    problem->max_cycle_length = MAX(3, problem->max_cycle_length);
/*
  printf("found %d triangles: %.2f seconds, ", 
	 problem->cycle_dict_size - cycle_dict_size0,
	 ((double)clock()-tic)/CLOCKS_PER_SEC);
*/
}

void  getAllQuadrangle(MapProblem* problem, int** adj_mat, 
		       MapParameters* parameters)
{
/*
  clock_t tic = clock();
*/
  int cycle_dict_size0 = problem->cycle_dict_size;
  int i, jid, kid, lid, N = problem->num_variable;
  for( i=0; i<N; i++){
  //      for( i=0; i<1; i++){ //
    for(jid = 0; jid < problem->var_degree[i]; jid++){
      int j = problem->var_nb[i][jid];
      if(j > i){
	for(kid = jid+1; kid < problem->var_degree[i]; kid++){
	  int k = problem->var_nb[i][kid];
	  //	  if(k > i && adj_mat[k][j] >= 0){
	  if(k > i && adj_mat[k][j] < 0){
	    for(lid = 0; lid < problem->var_degree[k]; lid++){
	      int l = problem->var_nb[k][lid];
	      if(l > i && adj_mat[j][l]>=0 && adj_mat[i][l]<0){

		int c = problem->cycle_dict_size;
		problem->cycle_dict_size++;
		problem->cycle_size = (int*)realloc(problem->cycle_size,
						    sizeof(int)*(c+1));
		problem->cycle_nodes = (int**)realloc(problem->cycle_nodes,
						      sizeof(int*)*(c+1));
		problem->cycle_edges = (int**)realloc(problem->cycle_edges,
						      sizeof(int*)*(c+1));
		problem->cycle_size[c] = 4;
		problem->cycle_nodes[c] = (int*)malloc(sizeof(int)*5);
		problem->cycle_edges[c] = (int*)malloc(sizeof(int)*4);
		problem->cycle_nodes[c][0] = i;
		problem->cycle_nodes[c][1] = j;
		problem->cycle_nodes[c][2] = l;
		problem->cycle_nodes[c][3] = k;
		problem->cycle_nodes[c][4] = i;
		problem->cycle_edges[c][0] = adj_mat[i][j];
		problem->cycle_edges[c][1] = adj_mat[j][l];
		problem->cycle_edges[c][2] = adj_mat[l][k];
		problem->cycle_edges[c][3] = adj_mat[k][i];
		
		if(problem->cycle_dict_size > cycle_dict_size0 + 1000*N){
		  if(problem->cycle_dict_size - cycle_dict_size0 > 0)
		    problem->max_cycle_length = MAX(4, problem->max_cycle_length);
/*
		  printf("found %d quadrangles: %.2f seconds\n", 
			 problem->cycle_dict_size - cycle_dict_size0,
			 ((double)clock()-tic)/CLOCKS_PER_SEC); 
*/
		  return;
		}
	      }
	    }
	  }
	}
      }
    }
  }
  if(problem->cycle_dict_size - cycle_dict_size0 > 0)
    problem->max_cycle_length = MAX(4, problem->max_cycle_length);
/*
  printf("found %d quadrangles: %.2f seconds\n", 
	 problem->cycle_dict_size - cycle_dict_size0,
	 ((double)clock()-tic)/CLOCKS_PER_SEC); 
*/
}

void buildCycleDictionary(MapProblem* problem, MapParameters* parameters)
{
  //  printf("building cycle dictionary\n");
  problem->cycle_dict_size=0;
  problem->cycle_size = NULL;
  problem->cycle_nodes = NULL;
  problem->cycle_edges = NULL;
  problem->max_cycle_length = 0;

  if(problem->is_grid == 4){

    getCyclesInN4Grid(problem, parameters);
    problem->max_cycle_length = 4;

  }else if(problem->is_grid == 8){

    getCyclesInN8Grid(problem, parameters);
    problem->max_cycle_length = 3;

  }else{
    int i,j,N = problem->num_variable;
    int** adj_mat = (int**)malloc(sizeof(int*) * N);
    for(i=0; i<N; i++){
      adj_mat[i] = (int*)malloc(sizeof(int)*N);
      for(j=0; j<N; ++j)
	adj_mat[i][j] = -1;
    }
    for(i=0; i<N; i++)
      for(j=0; j<problem->var_degree[i]; j++)
	adj_mat[i][problem->var_nb[i][j]] = problem->var_nb_edge[i][j];

    getAllTriangle(problem, adj_mat, parameters);
    getAllQuadrangle(problem, adj_mat, parameters);      

    /*
    if(problem->cycle_dict_size == 0){
      getLargeCycles(problem, parameters);
    }
    */

    for(i=0; i<N; i++)
      free(adj_mat[i]);
    free(adj_mat);
  }

  problem->cycle_score = (double*)malloc(sizeof(double)*(problem->cycle_dict_size));
  setDblVal(problem->cycle_score, 0.0, problem->cycle_dict_size);
}

void scoreCycles(MapProblem* problem, MapParameters* parameters)
{

/* 
  clock_t tic = clock();
  printf("score cycles ... "); fflush(stdout);
*/
  int cnt=0;
  int i,j;
  assert(sizeof(Cycle) > 0);
  Cycle *cy = (Cycle*)malloc(sizeof(Cycle));
  assert(problem->max_cycle_length>0);
  assert(problem->max_cardinality>0);
  cy->allocateMemory(problem->max_cycle_length, problem->max_cardinality);
  int* card1 = (int*)malloc(sizeof(int)*(problem->max_cycle_length+1));
  //  double** ts1 = (double**)malloc(sizeof(double*)*problem->max_cycle_length);
  double** tp1 = (double**)malloc(sizeof(double*)*(problem->max_cycle_length));
  //  cy->unary_msg = (int*)malloc(sizeof(int)*problem->max_cycle_length);
  int Kmax = problem->max_cardinality;
  for(i=0; i<problem->max_cycle_length; ++i){

    //    cy->unary_msg[i] = 0;
    //    ts1[i] = (double*)malloc(sizeof(double)*Kmax);  
    //    setDblVal(ts1[i], 0.0, Kmax);    
    tp1[i] = (double*)malloc(sizeof(double)*Kmax*Kmax); 
  }
  cy->card = card1;
  cy->ts = NULL;

  for(i=0; i<problem->cycle_dict_size; ++i){
    //    clock_t tic1 = clock();
    //    printf("%d/%d : ", i, problem->cycle_dict_size); fflush(stdout);
    //problem->cycle_score[i] = (double)rand()/RAND_MAX; continue;
    if(problem->cycle_score[i] == -1) continue; // cycle had been added

    double lower_bound = 0.0;
    cy->N = problem->cycle_size[i];
    cy->L = cy->N-2;
    for(j=0; j<problem->cycle_size[i]; ++j){
      
      int var_id = problem->cycle_nodes[i][j];
      int edge_id = problem->cycle_edges[i][j];

      setDblVal(tp1[j], 0.0, problem->sp[edge_id]->card[0]*
		problem->sp[edge_id]->card[1]);
      card1[j] = problem->card[var_id];
      //      printf( "%d ", card1[j]); fflush(stdout);
      cy->tp[j] = tp1[j];
      int transpose = (problem->sp[edge_id]->var_id[0] == var_id)?0:1;

      lower_bound += getSubProblemMin(problem->sp[edge_id], problem->cy_solver, parameters);
      extractMinMarginal(parameters,problem->sp[edge_id], problem->cy_solver, 2, 
			 problem->sp[edge_id]->var_id + transpose,
			 0, 1.0, tp1[j]);
    }
    card1[cy->N] = card1[0];
    for(j=0; j<cy->L-1; ++j)
      cy->tp[cy->N+j] = cy->cache[j];
    for (j = 0; j < cy->L; ++j){
      cy->TA_static[j] = -1;
      cy->TB_static[j] = -1;
    }
    cnt++;
    if(parameters->use_fast_solver)
      cy->runFastSolver(0);
    else
      cy->runFullMinSum(0, 1);
    problem->cycle_score[i] = cy->obj - lower_bound;

    //    printf("%.2f seconds\n", ((double)clock()-tic1)/CLOCKS_PER_SEC);

/*
    assert(problem->cycle_score[i] > -1e-6);
*/

  }
  for(i=0; i<problem->max_cycle_length; ++i){
    //    free(ts1[i]);
    free(tp1[i]);
  }
  //  free(ts1);
  free(tp1);
  free(card1);
  //  free(cy->unary_msg);
  cy->freeMemory();
  free(cy);
/*
  printf(" done (%d cycles), %.2f seconds, ", cnt, ((double)clock()-tic)/CLOCKS_PER_SEC);	 
*/
}

int dualGapConverged(MapProblem* problem)
{
  double dual_gap = problem->primal_obj - problem->dual_obj;
  problem->dual_gap_decrease = (problem->last_dual_gap - dual_gap);
  if(problem->last_dual_gap > 0)
    problem->dual_gap_decrease /= problem->last_dual_gap;
  problem->last_dual_gap = dual_gap;
  double threshold = problem->param->dual_gap_threshold;
  if(problem->used_cycles==0) threshold /= 10.0;
  if(problem->dual_gap_decrease < threshold)
    return 1;
  return 0;
}


void addCycles(MapProblem* problem, MapParameters* parameters)
{
/*  
  double dual_before_adding_cycles = evaluateDual(problem, parameters);
*/
/*
  clock_t tic = clock();
*/

  if(problem->cycle_initialized==0){
    problem->cycle_initialized = 1;
    if(problem->is_grid==-1)
      checkGridStructure(problem, parameters);
    if(problem->is_grid==0)
      buildVariableNeighborLists(problem, parameters);
    buildCycleDictionary(problem, parameters);
    problem->cy_solver = (Cycle*)malloc(sizeof(Cycle));
    problem->cy_solver->allocateMemory(problem->max_cycle_length, 
				       problem->max_cardinality);
    problem->sp2cycle = (int*)malloc(sizeof(int)* (problem->num_subproblem));
    setIntVal(problem->sp2cycle, -1, problem->num_subproblem);
  }

  if(problem->cycle_dict_size==0) return;

  int i,j;

  scoreCycles(problem, parameters);

  int K;
  int* cycles2add;
  K = MAX(5, (int)((parameters->add_cycle_batchsize)*(double)(problem->cycle_dict_size)));
  cycles2add = (int*)malloc(sizeof(int)*K);
  getTopKScore(problem->cycle_score, 
	       problem->cycle_dict_size,
	       &K, cycles2add, 1e-8);

/*  
  double dualobj = evaluateDual(problem, parameters);
  assert(fabs(dualobj - dual_before_adding_cycles)<1e-6);
*/
  if (K==0) return;

  problem->used_cycles += K;
  problem->num_subproblem += K;
  problem->sp=(SubProblem**)realloc(problem->sp, 
				    sizeof(SubProblem*)*problem->num_subproblem);
  problem->sp2cycle = (int*)realloc(problem->sp2cycle,
				    sizeof(int)*(problem->num_subproblem));
  problem->num_phi_of_sp = (int*)realloc(problem->num_phi_of_sp,
					 sizeof(int)*problem->num_subproblem);
  problem->phi_of_sp = (int**)realloc(problem->phi_of_sp,
				      sizeof(int*)*problem->num_subproblem);
  if(parameters->bipartite_graph == GRAPH_BETHE){
    problem->is_unary_subproblem = (double*)realloc(problem->is_unary_subproblem,
						    sizeof(double)*
						    (problem->num_subproblem));
  }
  int cycle_cnt;
  for(cycle_cnt=0; cycle_cnt<K; ++cycle_cnt){
    i = problem->num_subproblem-K+cycle_cnt; // index into problem->sp
    int c = cycles2add[cycle_cnt];           // index into problem->cycle_*
    problem->cycle_score[c] = -1;            // mark as used
    problem->sp[i] = (SubProblem*)malloc(sizeof(SubProblem));
    if(parameters->bipartite_graph == GRAPH_BETHE)
      problem->is_unary_subproblem[i] = 0.0;
    problem->sp2cycle[i] = c;
    SubProblem* p = problem->sp[i];
    p->num_variable = problem->cycle_size[c];
    int N = p->num_variable;
    p->func_id = -1;
    p->type = FUNC_TYPE_CYCLE;
    p->min_energy_valid = 0;
    p->val = NULL;
    p->num_msg = 2*N;
    p->msg = (double**)malloc(sizeof(double*)* (p->num_msg));
    p->card = (int*)malloc(sizeof(int)*(N+1));
    //    p->ld_card = (int*)malloc(sizeof(int)*(N+1));
    p->var_id = problem->cycle_nodes[c];
    for(j=0; j<N+1; ++j){
      p->card[j] = problem->card[p->var_id[j]];
    }
    p->sol = (int*)malloc(sizeof(int)*(N+1));
    problem->num_phi_of_sp[i] = 2*N;
    problem->phi_of_sp[i] = (int*)malloc(sizeof(int)*(problem->num_phi_of_sp[i]));

    for(j=0; j<N; ++j){

      // unary phi
      int phi_id =  p->var_id[j];
      problem->phi_of_sp[i][j] = phi_id;
      problem->sp_of_phi[phi_id] = 
	(int*)realloc(problem->sp_of_phi[phi_id],
		      sizeof(int)*(++(problem->num_sp_of_phi[phi_id])));
      problem->sp_of_phi[phi_id][problem->num_sp_of_phi[phi_id]-1] = i;

      // binary phi
      int edge_id = problem->cycle_edges[c][j];
      int* sp2phi = problem->sp2phi;
      if(sp2phi[edge_id] < 0){ // add new phi
	sp2phi[edge_id] = problem->num_phi;
	phi_id = sp2phi[edge_id];
	problem->phi_of_sp[i][j+N] = phi_id;
	append2IntVec(2, &(problem->num_sp_of_phi), &(problem->num_phi));

	problem->sp_of_phi = 
	  (int**)realloc(problem->sp_of_phi,sizeof(int*)*problem->num_phi);

	int edge_card = problem->sp[edge_id]->card[0]*
	  problem->sp[edge_id]->card[1];

	problem->sp_of_phi[phi_id] = (int*)malloc(sizeof(int)*2);
	problem->sp_of_phi[phi_id][0] = edge_id;
	problem->sp_of_phi[phi_id][1] = i;
	problem->scope_of_phi = 
	  (int**)realloc(problem->scope_of_phi,sizeof(int*)*problem->num_phi);
	problem->scope_of_phi[phi_id] = problem->sp[edge_id]->var_id;
	append2IntVec(phi_id, &(problem->phi_of_sp[edge_id]), 
		      &(problem->num_phi_of_sp[edge_id]));
	
	problem->sp[edge_id]->msg[2] = (double*)malloc(sizeof(double)*edge_card);
	setDblVal(problem->sp[edge_id]->msg[2], 0.0, edge_card);
 
      }else{ // add to existing phi
	phi_id = sp2phi[edge_id];
	problem->phi_of_sp[i][j+N] = phi_id;
	append2IntVec(i, &(problem->sp_of_phi[phi_id]), 
		      &(problem->num_sp_of_phi[phi_id]));
      }
      // init messages
      p->msg[j] = (double*)malloc(sizeof(double)*(p->card[j]));
      setDblVal(p->msg[j], 0.0, p->card[j]);
      p->msg[j+N] = 
	(double*)malloc(sizeof(double)*(p->card[j])*(p->card[j+1]));
      setDblVal(p->msg[j+N], 0.0, p->card[j]*p->card[j+1]);
    }
  }
  free(cycles2add);

/*
  assert(problem->used_cycles>0);
  printf("checking subproblem connectivities ... "); fflush(stdout);
  for(i=0; i<problem->num_phi; ++i){
    for(j=0; j<problem->num_sp_of_phi[i]; ++j){
      int sp_id = problem->sp_of_phi[i][j];
      assert(findInVec(i, problem->phi_of_sp[sp_id], 
		       problem->num_phi_of_sp[sp_id])  
	     < problem->num_phi_of_sp[sp_id]);
    }
  }

  for(i=0; i<problem->num_subproblem; ++i){
    for(j=0; j<problem->num_phi_of_sp[i]; ++j){
      int phi_id = problem->phi_of_sp[i][j];
      assert(findInVec(i, problem->sp_of_phi[phi_id], 
		       problem->num_sp_of_phi[phi_id])  
	     < problem->num_sp_of_phi[phi_id]);
    }
  }
  printf("done\n");
  double dual_after_adding_cycles = evaluateDual(problem, parameters);
  assert(fabs(dual_after_adding_cycles - dual_before_adding_cycles)<1e-6);
*/

/*
  printf("added %d cycles: %.2f seconds in total\n", K,
	 ((double)clock()-tic)/CLOCKS_PER_SEC);
*/

}

void       updateSubProblemDegrees(MapProblem* problem)
{
  problem->subproblem_degree = (double*)realloc(problem->subproblem_degree, 
						sizeof(double)*(problem->num_subproblem));
  for(int i=0; i<problem->num_subproblem; ++i){
    problem->subproblem_degree[i] = 0;
    if(problem->param->bipartite_graph == GRAPH_DC){
      for(int j=0; j<problem->num_phi_of_sp[i]; ++j){
	int phi = problem->phi_of_sp[i][j];
	problem->subproblem_degree[i] += problem->num_sp_of_phi[phi]-1;
      }
    }else{ // GRAPH_BETHE
      if(problem->sp[i]->num_variable==1){
	int phi = problem->sp[i]->var_id[0];
	problem->subproblem_degree[i] = problem->num_sp_of_phi[phi]-1;
      }else{
	problem->subproblem_degree[i] = problem->sp[i]->num_variable;
      }
    }
  }
}


int findPairId(int v0, int v1, int* list_size, int** lists, int N)
{
  if(v0 > v1)
    return findPairId(v1, v0, list_size, lists, N);
  // v0 < v1  
  if(list_size[v0] == 0) return -1;
  int i=0;
  while(lists[v0][i] % N != v1 && i < list_size[v0]) i++;
  if(i==list_size[v0]) return -1;
  //  assert(lists[v0][i] % N == v1);
  return lists[v0][i] / N;
  // find id * N + v1 in v0's list

  // var_id = element % N
  // id = element / N
  // element = id * N + var_id
}

void addPairwiseMsgScopes(MapProblem* problem, MapParameters* parameters)
{
  int num_var = problem->num_variable;
  int* var2pair_size = (int*)malloc(sizeof(int) * num_var);
  int* var2edge_size = (int*)malloc(sizeof(int) * num_var);
  int** var2pair = (int**)malloc(sizeof(int*) * num_var); // phi_id
  int** var2edge = (int**)malloc(sizeof(int*) * num_var); // sp_id

  for(int i=0; i<num_var; ++i){
    var2pair_size[i] = 0;
    var2edge_size[i] = 0;
    var2pair[i] = NULL;
    var2edge[i] = NULL;
  }
  
  for(int i=0; i<problem->num_subproblem; ++i){
    SubProblem* p = problem->sp[i];
    if(p->num_variable == 2){
      int v0 = MIN(p->var_id[0], p->var_id[1]);
      int v1 = MAX(p->var_id[0], p->var_id[1]);
      assert(v0 >= 0);
      assert(v1 < num_var);
      append2IntVec(i * num_var + v1, &(var2edge[v0]), &(var2edge_size[v0]));
    }    
  }

  for(int i=0; i<problem->num_subproblem; ++i){
    SubProblem* p = problem->sp[i];
    if(p->type == FUNC_TYPE_TABLE && p->num_variable > 2){
      int N = p->num_variable;
      //      int num_pairs = (N*(N-1))/2;
      //      problem->num_phi_of_sp[i] += num_pairs;
      //      problem->phi_of_sp[i] = (int*)realloc(problem->phi_of_sp[i], 
      //					    sizeof(int)*problem->num_phi_of_sp[i]);

      p->num_msg = N+N*N;
      p->msg = (double**)realloc(p->msg, sizeof(double*) * (p->num_msg));
      //      p->ld_msg = (double**)realloc(p->ld_msg, sizeof(double*) * (p->num_msg));
      for(int j=N; j<N+N*N; ++j){
      	p->msg[j] = NULL;
      //	p->ld_msg[j] = NULL;
      }

      for(int j=0; j<N; ++j){
	for(int k=j+1; k<N; ++k){
	  int v0 = p->var_id[j];
	  int v1 = p->var_id[k];
	  int K = (p->card[j]) *(p->card[k]);

	  int msg_idx = N + j*N + k; // j<k
	  p->msg[msg_idx] = (double*)malloc(sizeof(double) * K);
	  vecset(p->msg[msg_idx], 0.0, K);
	  
	  int phi_id = findPairId(v0, v1, var2pair_size, var2pair, num_var);
	  if(phi_id >= 0){
	    append2IntVec(i, &(problem->sp_of_phi[phi_id]), &(problem->num_sp_of_phi[phi_id]));
	    append2IntVec(phi_id, &(problem->phi_of_sp[i]), &(problem->num_phi_of_sp[i]));
	  }else{
	    int edge_id = findPairId(v0, v1, var2edge_size, var2edge, num_var);  
	    int new_phi_id = problem->num_phi;
	    append2IntVec(1, &(problem->num_sp_of_phi), &(problem->num_phi));
	    problem->sp_of_phi = 
	      (int**)realloc(problem->sp_of_phi,sizeof(int*)*(problem->num_phi));
	    problem->sp_of_phi[new_phi_id] = (int*)malloc(sizeof(int));
	    problem->sp_of_phi[new_phi_id][0] = i;
	    problem->scope_of_phi = 
	      (int**)realloc(problem->scope_of_phi,sizeof(int*)*(problem->num_phi));
	    problem->scope_of_phi[new_phi_id] = (int*)malloc(sizeof(int)*3);

	    if(edge_id >= 0){
	      problem->scope_of_phi[new_phi_id][0] = problem->sp[edge_id]->var_id[0];
	      problem->scope_of_phi[new_phi_id][1] = problem->sp[edge_id]->var_id[1];
	      problem->scope_of_phi[new_phi_id][2] = problem->sp[edge_id]->var_id[0];
	    }else{
	      problem->scope_of_phi[new_phi_id][0] = v0;
	      problem->scope_of_phi[new_phi_id][1] = v1;
	      problem->scope_of_phi[new_phi_id][2] = v0;
	    }
	    if(v0 < v1)
	      append2IntVec(num_var * new_phi_id + v1,
			    &(var2pair[v0]),
			    &(var2pair_size[v0]));
	    else
	      append2IntVec(num_var * new_phi_id + v0, 
			    &(var2pair[v1]),
			    &(var2pair_size[v1]));

	    append2IntVec(new_phi_id, 
		       &(problem->phi_of_sp[i]),
		       &(problem->num_phi_of_sp[i]));

	    if(edge_id >= 0){
	      //	      assert(edge_id < problem->num_subproblem);
	      problem->sp2phi[edge_id] = new_phi_id;
	      append2IntVec(new_phi_id, &(problem->phi_of_sp[edge_id]), 
			 &(problem->num_phi_of_sp[edge_id]));
	      assert(problem->num_phi_of_sp[edge_id] == 3);
	      append2IntVec(edge_id, &(problem->sp_of_phi[new_phi_id]),
			 &(problem->num_sp_of_phi[new_phi_id]));
	      problem->sp[edge_id]->msg[2] = (double*)malloc(sizeof(double)*K);
	      vecset(problem->sp[edge_id]->msg[2], 0.0, K);
	    }
	  }
	}
      }
    }
  }

  for(int i=0; i<num_var; ++i){
    if(var2pair_size[i])
      free(var2pair[i]);
    if(var2edge_size[i])
      free(var2edge[i]);
  }
  free(var2pair_size);
  free(var2edge_size);
  free(var2pair);
  free(var2edge);

}

void initMapProblem(MapProblem* problem, MapParameters* parameters)
{
  problem->param = parameters;
/*
  printf("initMapProblem\n");
*/

  problem->cycle_initialized = 0;  
  problem->is_grid = -1;
  int i,j;
  //  printf("check connectivity before init\n");
  //  checkConnectivity(problem); 

  problem->num_nonunary_subproblem = problem->num_subproblem;
  problem->unary_subproblem_id = (int*)malloc(sizeof(int)*(problem->num_variable));

  int* unary_subproblem_id = problem->unary_subproblem_id;
  // swap all unary subproblems to the end
  for(int i=0; i<problem->num_variable; ++i)
    unary_subproblem_id[i] = -1;
  i=0;
  while(i<problem->num_nonunary_subproblem){
    if(problem->sp[i]->num_variable == 1){
      --(problem->num_nonunary_subproblem);
      SubProblem* tmp = problem->sp[i];
      problem->sp[i] = problem->sp[problem->num_nonunary_subproblem];
      problem->sp[problem->num_nonunary_subproblem] = tmp;
      assert(problem->sp[problem->num_nonunary_subproblem]->var_id[0] >= 0);
      assert(problem->sp[problem->num_nonunary_subproblem]->var_id[0] < problem->num_variable);
      int var = problem->sp[problem->num_nonunary_subproblem]->var_id[0];
      
      if(unary_subproblem_id[var] == -1){
	unary_subproblem_id[var] = problem->num_nonunary_subproblem;
      }else{
	SubProblem *u0 = problem->sp[unary_subproblem_id[var]];
	assert(u0->var_id[0] == tmp->var_id[0]);
	assert(u0->card[0] == tmp->card[0]);
	vecadd(u0->val, tmp->val, 1.0, u0->card[0]);
	vecset(tmp->val, 0.0, u0->card[0]);
	assert(problem->sp[problem->num_subproblem-1]->num_variable==1);
	problem->sp[problem->num_nonunary_subproblem] = 
	  problem->sp[problem->num_subproblem-1];
	problem->sp[problem->num_subproblem-1] = tmp;
	int var1 = problem->sp[problem->num_nonunary_subproblem]->var_id[0];
	unary_subproblem_id[var1] = problem->num_nonunary_subproblem;
	problem->num_subproblem--;
      }
    }else{
      ++i;
    }
  }

  // add missing unaries
  if(parameters->bipartite_graph == GRAPH_BETHE){
    int num_new_unary = 0;
    int sp_id = problem->num_subproblem;
    for(int i=0; i<problem->num_variable; ++i){
      if(unary_subproblem_id[i] < 0){
	num_new_unary ++;
      }
    }
    problem->num_subproblem += num_new_unary;
    problem->sp=
      (SubProblem**)realloc(problem->sp, 
			    sizeof(SubProblem*)*problem->num_subproblem);
    for(int i=0; i<problem->num_variable; ++i){
      if(unary_subproblem_id[i] < 0){
	unary_subproblem_id[i] = sp_id;
	problem->sp[sp_id] = (SubProblem*)malloc(sizeof(SubProblem));
	SubProblem* p = problem->sp[sp_id];
	int K = problem->card[i];
	p->num_variable = 1;
	p->func_id = -1;
	p->type = FUNC_TYPE_TABLE;
	p->min_energy_valid = 0;
	p->val = NULL;
	p->card = (int*)malloc(sizeof(int));
	p->var_id = (int*)malloc(sizeof(int));
	p->var_id[0] = i;
	p->card[0] = K;
	p->sol = (int*)malloc(sizeof(int));
	sp_id++;
      }
    }
  }

  // fill-in sp_of_phi, init messages for each sp
  problem->num_phi = problem->num_variable;
  problem->num_sp_of_phi = (int*)malloc(sizeof(int)*(problem->num_phi));

  problem->scope_of_phi = (int**)malloc(sizeof(int*)*(problem->num_phi));
  problem->sp_of_phi = (int**)malloc(sizeof(int*)*(problem->num_phi));

  for(i=0; i<problem->num_phi; ++i){
    problem->num_sp_of_phi[i] = 0;
    problem->scope_of_phi[i] = (int*)malloc(sizeof(int));
    problem->scope_of_phi[i][0] = i;
    problem->sp_of_phi[i] = NULL;
  }

  if(parameters->bipartite_graph == GRAPH_BETHE){
    problem->is_unary_subproblem = 
      (double*)malloc(sizeof(double)*(problem->num_subproblem));
    vecset(problem->is_unary_subproblem, 0.0, problem->num_subproblem);
  }

  int N = (parameters->bipartite_graph == GRAPH_DC)?
    problem->num_nonunary_subproblem:
    (problem->num_subproblem);
    
  for(i=0; i<N; ++i){
    SubProblem* p = problem->sp[i];
    if(parameters->bipartite_graph == GRAPH_BETHE){
      if(p->num_variable==1)
	problem->is_unary_subproblem[i] = 1.0;
    }else{
      assert(p->num_variable>1);
    }
    for(j=0; j<p->num_variable; ++j){
      int v = p->var_id[j];
      append2IntVec(i, &(problem->sp_of_phi[v]), 
		    &(problem->num_sp_of_phi[v]));
    }
    p->num_msg = p->num_variable==1?1:(p->num_variable+1);
    p->msg = (double**)malloc(sizeof(double*)*(p->num_msg));
    for(j=0; j<p->num_msg; ++j)
      p->msg[j] = NULL;
    for(j=0; j<p->num_variable; ++j){
      p->msg[j] = (double*)malloc(sizeof(double)*(p->card[j]));
      setDblVal(p->msg[j], 0.0, p->card[j]);
    }
  }
  problem->max_cardinality = 0;
  for(i=0; i<problem->num_variable; ++i)
    problem->max_cardinality = MAX(problem->max_cardinality, problem->card[i]);

/*
  printf("max_card = %d\n", problem->max_cardinality);
*/

  parameters->buffer = (void*)malloc(sizeof(double)*5*(problem->max_cardinality));

  if(parameters->bipartite_graph == GRAPH_DC){
    // split each unary subproblem into non unary subproblems
    // and remove unary subproblems
    for(i=0; i<problem->num_variable; ++i){
      if(unary_subproblem_id[i] < 0) continue;
      SubProblem* unary = problem->sp[unary_subproblem_id[i]];
      assert(unary->var_id[0] == i);
      if(problem->num_sp_of_phi[i]==0){

	// swap this unary subproblem to the end of non unary subproblems
	// and keep it

	if(problem->sp[problem->num_nonunary_subproblem]){
	  assert(unary_subproblem_id[problem->sp[problem->num_nonunary_subproblem]->var_id[0]] == problem->num_nonunary_subproblem);
	}

	problem->sp[unary_subproblem_id[i]] = 
	  problem->sp[problem->num_nonunary_subproblem];
	problem->sp[problem->num_nonunary_subproblem] = unary;

	if(problem->sp[unary_subproblem_id[i]]){
	  int var = problem->sp[unary_subproblem_id[i]]->var_id[0];
	  assert(var >= i);
	  unary_subproblem_id[var] = unary_subproblem_id[i];
	}
	unary_subproblem_id[i] = problem->num_nonunary_subproblem;
	problem->num_nonunary_subproblem ++;

	unary->num_msg = 1;
	unary->msg = (double**)malloc(sizeof(double*));
	unary->msg[0] = (double*)malloc(sizeof(double)*(unary->card[0]));
	setDblVal(unary->msg[0], 0.0, unary->card[0]);

      }else{
	double factor = 1.0/problem->num_sp_of_phi[i];
	for(j=0; j<problem->num_sp_of_phi[i]; ++j){
	  SubProblem* p = problem->sp[problem->sp_of_phi[i][j]];
	  int id = findInVec(i, p->var_id, p->num_variable);
/*
	  assert(id < p->num_variable);
*/      
	  addVec2Vec(unary->val, p->msg[id], problem->card[i], factor);
	}
	// setDblVal(unary->val, 0.0, problem->card[i]);
	freeSubProblem(unary);
	free(unary);
	problem->sp[unary_subproblem_id[i]] = NULL;
	unary_subproblem_id[i] = -1;
      }
    }
    problem->num_subproblem = problem->num_nonunary_subproblem;
    problem->sp = (SubProblem**)realloc(problem->sp, 
					sizeof(SubProblem*)*
					problem->num_subproblem);

  }

  // pre-compute values for truncated distance functions
  // NOTE: this would change how the subproblems are evaluated 

  for(i=0; i<problem->num_function; ++i){
    Function* f = &(problem->func[i]);
    int K = problem->max_cardinality;
    if(f->type == FUNC_TYPE_TL1){
      double alpha = f->val[0];
      double gamma = f->val[1];
      f->val = (double*)realloc(f->val, sizeof(double)*K);
      for(j=0; j<K; ++j){
	f->val[j] = MIN(alpha * j, gamma);
      }
    }else if(f->type == FUNC_TYPE_TL2){
      double alpha = f->val[0];
      double gamma = f->val[1];
      f->val = (double*)realloc(f->val, sizeof(double)*3*K);
      f->val[0] = gamma;
      f->val[K] = 0;
      f->val[K+K] = 0;
      for(j=1; j<K; ++j){
	f->val[j] = 1.0/(alpha*j);
	f->val[K+j] = alpha*j*j;
	f->val[K+K+j] = MIN(alpha*j*j, gamma);
      }
    }
  }
  // merge unary subproblems' func value into incoming messages
  for(int i=0; i<problem->num_subproblem; ++i){
    if(problem->sp[i]->num_variable==1){
      SubProblem* u = problem->sp[i];
      if(u->val){
	vecadd(u->msg[0], u->val, 1.0, u->card[0]);
	u->val = NULL;
      }
    }
  }

  problem->initialized = 1;
  ////////////////////////////////////////////////////////////////////////

  for(i=0; i<problem->num_subproblem; ++i){
    assert(problem->sp[i]);
    int func_id = problem->sp[i]->func_id;
    if(func_id >= 0){
      if(problem->func[func_id].type == FUNC_TYPE_TL1 ||
	 problem->func[func_id].type == FUNC_TYPE_TL2){
	problem->sp[i]->val = problem->func[func_id].val;
	//	problem->sp[i]->size_val = problem->func[func_id].size_val;
	assert(problem->sp[i]->card[0] == problem->max_cardinality);
	assert(problem->sp[i]->card[1] == problem->max_cardinality);
      }
    }
  }

  // fill-in phi_of_sp
  problem->num_phi_of_sp = (int*)malloc(sizeof(int)*(problem->num_subproblem));
  problem->phi_of_sp = (int**)malloc(sizeof(int*)*(problem->num_subproblem));
  for(i=0; i<problem->num_subproblem; ++i){
    problem->num_phi_of_sp[i] = problem->sp[i]->num_variable;
    problem->phi_of_sp[i] = (int*)malloc(sizeof(int)* (problem->num_phi_of_sp[i]));
    for(j=0; j<problem->num_phi_of_sp[i]; ++j)
      problem->phi_of_sp[i][j] = problem->sp[i]->var_id[j];
  }

  problem->sp2phi = (int*)malloc(sizeof(int)* (problem->num_subproblem));  
  setIntVal(problem->sp2phi, -1, problem->num_subproblem);

  if(parameters->pair_msg_for_high_order_clique)
    addPairwiseMsgScopes(problem, parameters);

  renewMRFSolution(problem, parameters);
  problem->dual_obj = evaluateDual(problem, parameters);
  problem->init_dual_obj = problem->dual_obj;
  problem->init_primal_obj = problem->primal_obj;
  problem->last_dual_gap = problem->primal_obj - problem->dual_obj;
  problem->dual_gap_decrease = 1.0;  
  problem->next_action = ACTION_STC;
  problem->used_cycles = 0;
  allocateSubProblemTree(problem);

  //  printf("check connectivity after init\n");
  //  checkConnectivity(problem);

/*
  printf("init primal = %f, dual = %f\n", problem->init_primal_obj,
	 problem->init_dual_obj);
*/

  if(parameters->add_cycle_batchsize < 0){
    if(problem->num_variable > 1e4)
      parameters->add_cycle_batchsize = 0;
    else
      parameters->add_cycle_batchsize = 0.001;
  }

  if(parameters->start_cycle_batchsize > 0){
    double tmp = parameters->add_cycle_batchsize;
    parameters->add_cycle_batchsize = parameters->start_cycle_batchsize;
    addCycles(problem, parameters); 
    parameters->add_cycle_batchsize = tmp;
    freeSubProblemTree(problem);
    allocateSubProblemTree(problem);



  }

  if(parameters->allocation_method == ALLOCATION_DEGREE){
    problem->subproblem_degree = NULL;
    updateSubProblemDegrees(problem);
  }

  if(parameters->allocation_method == ALLOCATION_NEXT_LEAF){
    int method_id = parameters->tree_selection;
    //    assert(method_id != -1);
    problem->next_sptree = NULL;
    buildTree(problem, method_id, parameters);
    problem->next_sptree = problem->sptree;
    allocateSubProblemTree(problem);
    assert(problem->next_sptree);
    assert(problem->sptree);
  }
}

int runInference(MapProblem* problem, MapParameters* parameters)
{
  double dual_gap = problem->primal_obj - problem->dual_obj;
  if(problem->dual_obj != 0)
    assert(dual_gap/fabs(problem->dual_obj) > -1e-6);
  else
    assert(dual_gap > -1e-6);

  if(dual_gap < 1e-6) return 1;

  if(problem->next_action == ACTION_STC){
    
    int tree_select = parameters->tree_selection;

    problem->msg_cnt = 0;
    while(problem->msg_cnt < parameters->stc_inner_iterations){
      //      printf("."); fflush(stdout);
      if(tree_select >= TREE_SP1){
	//	clock_t tic = clock();
	buildTree(problem, tree_select, parameters);
	//	printf("(b: %.2f seconds)", ((double)clock()-tic)/CLOCKS_PER_SEC);
	//	tic = clock();
	calibrateTree(problem, parameters);
	//	printf("(c: %.2f seconds)", ((double)clock()-tic)/CLOCKS_PER_SEC);

      }else{
	if(parameters->bipartite_graph == GRAPH_DC){
	  if(tree_select == TREE_STARP){
	    starAverage(problem, parameters);
	  }else{assert(0);}
	}else{ // GRAPH_BETHE
	  if(tree_select == TREE_PAIR){
	    calibrateBethePairs(problem, parameters);
	  }else if(tree_select == TREE_STAR){
	    calibrateBetheStars(problem, parameters);
	  }else if(tree_select == TREE_STARP){
	    calibrateBetheStarPlus(problem, parameters);
	  }else{assert(0);}
	}
      }
    }// end while

    renewDualObj(problem, parameters);
    //clock_t tic = clock();
    renewMRFSolution(problem, parameters);
    //printf("(%.2f seconds)", ((double)clock()-tic)/CLOCKS_PER_SEC);

    if(dualGapConverged(problem)){
      if(parameters->add_cycle_batchsize > 0){
	problem->next_action = ACTION_ADD_CYCLE;
      }
      problem->last_action = ACTION_STC;
    }

  }else if(problem->next_action == ACTION_SUBSPACE_SEARCH){

    // pass;
    problem->next_action = ACTION_STC;
    problem->last_action = ACTION_SUBSPACE_SEARCH;


  }else if(problem->next_action == ACTION_ADD_CYCLE){

    addCycles(problem, parameters); 
    freeSubProblemTree(problem);
    allocateSubProblemTree(problem);
    if(parameters->allocation_method == ALLOCATION_DEGREE)
      updateSubProblemDegrees(problem);

    if(parameters->allocation_method == ALLOCATION_NEXT_LEAF){
      SubProblemTree * tmp = problem->sptree;
      problem->sptree = problem->next_sptree;
      freeSubProblemTree(problem);
      problem->sptree = tmp;
      int method_id = parameters->tree_selection;
      problem->next_sptree = NULL;
      buildTree(problem, method_id, parameters);
      problem->next_sptree = problem->sptree;
      allocateSubProblemTree(problem);
    }
    problem->next_action = ACTION_STC;
  }
  return 0;
}

void freeMapProblem(MapProblem** problem)
{
  int i;

  if((*problem)->initialized){

    if((*problem)->param->bipartite_graph == GRAPH_BETHE)
      free((*problem)->is_unary_subproblem);
    free((*problem)->unary_subproblem_id);
    free((*problem)->param->buffer);

    if((*problem)->sptree)
      freeSubProblemTree(*problem);
    if((*problem)->param->allocation_method == ALLOCATION_NEXT_LEAF){
      (*problem)->sptree = (*problem)->next_sptree;
      freeSubProblemTree(*problem);
    }
    if((*problem)->param->allocation_method == ALLOCATION_DEGREE)
      free((*problem)->subproblem_degree);
    for(i=0; i<(*problem)->num_phi; ++i){
      free((*problem)->sp_of_phi[i]);
    }
    
    for(i=0; i<(*problem)->num_variable; ++i)
      free((*problem)->scope_of_phi[i]);
    for(i=0; i<(*problem)->num_subproblem; ++i)
      free((*problem)->phi_of_sp[i]);
    if((*problem)->num_sp_of_phi)
      free((*problem)->num_sp_of_phi);
    free((*problem)->sp_of_phi);
    free((*problem)->num_phi_of_sp);
    free((*problem)->phi_of_sp);
    free((*problem)->scope_of_phi);
    free((*problem)->sp2phi);

    if((*problem)->used_cycles){
      (*problem)->cy_solver->freeMemory();
      free((*problem)->cy_solver);
      free((*problem)->sp2cycle);
      if((*problem)->is_grid==0){
	free((*problem)->var_degree);
	for(i=0; i< (*problem)->num_variable; ++i){
	  free((*problem)->var_nb[i]);
	  free((*problem)->var_nb_edge[i]);
	}
	free((*problem)->var_nb);
	free((*problem)->var_nb_edge);
      }else{
	free((*problem)->grid_edge2sp);
      }
      for(i=0; i< (*problem)->cycle_dict_size; ++i){
	free((*problem)->cycle_nodes[i]);
	free((*problem)->cycle_edges[i]);
      }
      free((*problem)->cycle_size);
      free((*problem)->cycle_score);
      free((*problem)->cycle_nodes);
      free((*problem)->cycle_edges);
    }

  }
  for(i=0; i<(*problem)->num_function; ++i)
    free((*problem)->func[i].val);
  for(i=0; i<(*problem)->num_subproblem; ++i){
    freeSubProblem((*problem)->sp[i]);
    free((*problem)->sp[i]);
  }
  free((*problem)->card);
  free((*problem)->sp);
  free((*problem)->func);
  free((*problem)->sol);
  free(*problem);
}

#endif
